/*
Copyright (c) 2011, Sylvain Raybaud <sylvain.raybaud@gmail.com>
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
Neither the name of the PAROLE team, LORIA, INRIA nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdio.h>
#include <glib.h>
#include <getopt.h>
#include <string.h>
#include <stdlib.h>

#include <cutils.h>

#include "ctm2plain.h"

#define DEFAULT_HMM MODELDIR "/hmm/en_US/hub4wsj_sc_8k"

int main(int argc, char ** argv) {
    
    parameters_t params = parse_cl(argc,argv);
    
    init(&params);
    
    if(params.filenames==NULL) { /* reading from stdin */
        ctm2plain(&params,NULL);
    }
    else {
        int i = 0;
        while(params.filenames[i]!=NULL) {
            ctm2plain(&params,params.filenames[i]);
            i += 1;
        }
    }
    
    parameters_free(&params);
    
    return 0;
}

parameters_t parse_cl(int argc, char ** argv) {
    debug("starting");
    
    static const char * usage = "Usage: %s [options] [ctm file(s)] [> plain text]\nOptions are:\n\
    --lm or -l: language model in DMP format\n\
    --dict or -d: source language dictionary (for recognition)\n\
    --fdict or -f: filler/noise dictionary (for recognition)\n\
    --hmm or -H: hmm directory (only for peace of mind, not really used except for silence definition)\n\
    --segmentation-method or -t (emit|prefer|best|best-prefer|silence|length|mix|none): choose segmentation method. (default=mix)\n\
    --max-length or -m: maximum length of a segment (default=40, 0 for no limit)\n\
    --dont-clean or -C: don't clean asr hypothesis (remove '_', retokenise, ...)\n\
    --keep-fillers or -k: don't remove fillers and silences\n\
    --verbose or -v: print out some details on stderr\n\
    --help or -h: print this message and exit\n";
    
    static struct option long_options[] = {
        {"verbose",             no_argument,       NULL, 'v'}, /* verbosity */
        {"lm",                  required_argument, NULL, 'l'},
        {"dict",                required_argument, NULL, 'd'},
        {"fdict",               required_argument, NULL, 'f'},
        {"hmm",                 required_argument, NULL, 'H'},
        {"max-length",          required_argument, NULL, 'm'},
        {"segmentation-method", required_argument, NULL, 't'},
        {"dont-clean",          no_argument,       NULL, 'C'},
        {"keep-fillers",        no_argument,       NULL, 'k'},
        {"help",                no_argument,       NULL, 'h'},
        {0, 0, 0, 0}
    };
    
    parameters_t params = {0};
    params.seg_mix = 1;
    params.max_length = 40;
    params.clean = 1;
    
    int help = 0;
    
    int option_code = 1;
    
    while(option_code!=-1) {
        
        int option_index = 0;
        option_code = getopt_long(argc, argv, "m:l:f:d:t:H:hv",long_options, &option_index);
        switch(option_code) {
            case 0:  /* long option setting a flag */
                break;
            case 'v':
                params.verbose = 1;
                break;
            case 'k':
                params.keep_fillers = 1;
                break;
            case 'C':
                params.clean = 0;
                break;
            case 'h':
                help = 1;
                break;
            case 'l':
                params.lmfile = optarg;
                break;
            case 'H':
                params.hmm = optarg;
                break;
            case 'f':
                params.fdict = optarg;
                break;
            case 'd':
                params.dict = optarg;
                break;
            case 't':
                if      (!strcmp(optarg,"emit")) params.seg_emit = 1;
                else if (!strcmp(optarg,"prefer")) params.seg_prefer = 1;
                else if (!strcmp(optarg,"best")) params.seg_best = 1;
                else if (!strcmp(optarg,"best-prefer")) params.seg_best_prefer = 1;
                else if (!strcmp(optarg,"silence")) params.seg_silence = 1;
                else if (!strcmp(optarg,"length")) params.seg_length = 1;
                else if (!strcmp(optarg,"mix")) params.seg_mix = 1;
                else if (!strcmp(optarg,"none")) params.seg_none = 1;
                else myfail(1,"invalid segmentation method '%s'",optarg);
                break;
            case 'm':
                params.max_length = atoi(optarg);
                break;
            case -1:  /* end of options */
                break;
            case '?':  /* unknown option */
                help = 1;
                break;
            default:  /* serious error */
                myerror("you are not reading this.");
                abort();
        }
        
    }
    
    if(help) {
        printf(usage,argv[0]);
        exit(0);
    }
    
    if(params.seg_emit || params.seg_prefer || params.seg_best || params.seg_best_prefer || params.seg_silence || params.seg_length || params.seg_none)
        params.seg_mix = 0;
    
    /* file names */
    int i = optind;
    params.filenames = (optind<argc)?malloc((argc-optind+1)*sizeof(char*)):NULL;
    while(i<argc) {
        params.filenames[i-optind] = argv[i];
        i += 1;
    }
    params.filenames[i-optind] = NULL;
    
    myverbose(params.verbose,"lm=%s",params.lmfile);
    myverbose(params.verbose,"dict=%s",params.dict);
    myverbose(params.verbose,"fdict=%s",params.fdict);
    myverbose(params.verbose,"hmm=%s",params.hmm);
    myverbose(params.verbose,"seg=%s",params.seg_emit?"emit":(params.seg_prefer?"prefer":(params.seg_best?"best":(params.seg_silence?"silence":(params.seg_length?"length":(params.seg_mix?"mix":(params.seg_none?"none":(params.seg_best_prefer?"best-prefer":"undefined"))))))));
    myverbose(params.verbose,"max length=%i",params.max_length);
    myverbose(params.verbose,"clean=%s",params.clean?"yes":"no");
    myverbose(params.verbose,"keep fillers=%s",params.keep_fillers?"yes":"no");
    
    return params;
    
}


void init(parameters_t * params) {
    debug("starting");
    
    params->ps_cmd_ln = cmd_ln_init(params->ps_cmd_ln,ps_args(),TRUE,
                                    "-fdict", params->fdict,
                                    "-lm", params->lmfile,
                                    "-lmname", "default",
                                    "-dict", params->dict,
                                    "-hmm", params->hmm,
                                    NULL);
                                           
    params->ps = ps_init(params->ps_cmd_ln);
    
    params->lm = ngram_model_set_select(ps_get_lmset(params->ps),"default");
    params->sentence_start_id = ngram_wid(params->lm,"<s>");
    params->unk_id = ngram_unknown_wid(params->lm);
    params->lm_zero = ngram_zero(params->lm);
    
    int32 id;
    int id_ptr = 0;
    params->mt_cut_symbols_id = malloc(8*sizeof(int32));
    if((id = ngram_wid(params->lm,","))!=params->unk_id)
        params->mt_cut_symbols_id[id_ptr++] = id;
    if((id = ngram_wid(params->lm,";"))!=params->unk_id)
        params->mt_cut_symbols_id[id_ptr++] = id;
    if((id = ngram_wid(params->lm,":"))!=params->unk_id)
        params->mt_cut_symbols_id[id_ptr++] = id;
    if((id = ngram_wid(params->lm,"?"))!=params->unk_id)
        params->mt_cut_symbols_id[id_ptr++] = id;
    if((id = ngram_wid(params->lm,"!"))!=params->unk_id)
        params->mt_cut_symbols_id[id_ptr++] = id;
    if((id = ngram_wid(params->lm,"."))!=params->unk_id)
        params->mt_cut_symbols_id[id_ptr++] = id;
    if((id = ngram_wid(params->lm,"</s>"))!=params->unk_id)
        params->mt_cut_symbols_id[id_ptr++] = id;
    params->mt_cut_symbols_id[id_ptr] = params->unk_id;
}

int cut_on_wid(parameters_t * params, int32 wid) {
    debug("starting");
    int i = 0;
    while(params->mt_cut_symbols_id[i]!=params->unk_id) {
        if(wid==params->mt_cut_symbols_id[i]) {
            return TRUE;
        }
        i++;
    }
    return FALSE;
}


void parameters_free(parameters_t * p) {
    debug("starting");
    free(p->filenames);
    ps_free(p->ps);
}


void ctm2plain(parameters_t * params, char * filename) {
    debug("starting");
    
    FILE * in = (filename!=NULL)?fopen(filename,"r"):stdin;
    if(in==NULL) {
        myerror("could not open file '%s' for writing.",filename);
    }
    
    FILE * out = NULL;
    char * outfilename = (filename!=NULL)?malloc((strlen(filename)+5)*sizeof(char)):NULL;
    if(outfilename!=NULL) {
        outfilename = strcpy(outfilename,filename);
        outfilename = strcat(outfilename,".txt");
        out = fopen(outfilename,"w");
    }
    else {
        out = stdout;
    }
    if(out==NULL) {
        myerror("could not open file '%s' for writing.",outfilename);
    }
    
    word_list_t * stream = NULL;
    size_t l = 0;
    char * line = NULL;
    while(getline(&line,&l,in)>0) {
        gchar ** tokens = g_strsplit(line," ",0);
        stream = add_word(stream,tokens[4]);
        g_strfreev(tokens);
        stream = segment(params,stream,out); /* extract all segments you can get */
    }
    
    write_segment(params,stream,-1,out); /* flush whatever remains */
    
    if(outfilename!=NULL) fclose(out);
    
}

word_list_t * add_word(word_list_t * list, gchar * word) {
    debug("starting");
    word_list_t * new = new_word(word);
    if(list==NULL) {
        return new;
    }
    list_end(list)->next = new;
    return list;
}

word_list_t * new_word(gchar * word) {
    debug("starting");
    word_list_t * new = malloc(sizeof(word_list_t));
    new->next = NULL;
    new->word = malloc((strlen(word)+1)*sizeof(word));
    strcpy(new->word,word);
    return new;
}

word_list_t * list_end(word_list_t * list) {
    debug("starting");
    word_list_t * item = list;
    while(item!=NULL && item->next!=NULL) {
        item = item->next;
    }
    return item;
}

word_list_t * write_segment(parameters_t * params, word_list_t * from, int chunk_length, FILE * out) {
    debug("starting");
    word_list_t * token = from;
    int len = 1;
    while(token!=NULL && (chunk_length<0 || len<=chunk_length)) {
        if(params->keep_fillers || !is_non_word(token->word)) {
            if(params->clean) {
                char * clean_word = clean(token->word);
                fprintf(out,"%s ",clean_word);
                free(clean_word);
            }
            else {
                fprintf(out,"%s ",token->word);
            }
        }
        word_list_t * next = token->next;
        free(token->word);
        free(token);
        token = next;
        len += 1;
    }
    if(chunk_length>0) fputs("\n",out);
    
    if(len<chunk_length) {
        mywarning("requested chunk longer than word stream... really strange.");
    }
    
    return token;
}

word_list_t * segment(parameters_t * params, word_list_t * stream, FILE * out) {
    debug("starting");
    
    word_list_t * list = stream;
    int chunk_length = 0;
    do {
        if(params->seg_emit)        chunk_length = segment_emit(params,list);
        if(params->seg_prefer)      chunk_length = segment_prefer(params,list);
        if(params->seg_best)        chunk_length = segment_at_best(params,list);
        if(params->seg_best_prefer) chunk_length = segment_best_prefer(params,list);
        if(params->seg_silence)     chunk_length = segment_silence(params,list);
        if(params->seg_length)      chunk_length = segment_length(params,list);
        if(params->seg_mix)         chunk_length = segment_mix(params,list);
        /* if(params->seg_none)    cut = NULL; */
        list = write_segment(params,list,chunk_length,out);
    }
    while(chunk_length>0) ;
    
    return list;
}

/* acoustic based method */
/* cut on ++silence++ */
int segment_silence(parameters_t * params, word_list_t * list) {
    debug("starting");
    word_list_t * stream = list;
    int i = 1;
    while((params->max_length<=0 || i<=params->max_length) && stream!=NULL) {
        if(is_silence(stream->word)) {
            return i;
        }
        stream = stream->next;
        i += 1;
    }
    
    return 0;
}

/* dummy method: cut after n words max */
int segment_length(parameters_t * params, word_list_t * list) {
    debug("starting");
    if(params->max_length<=0)
        return 0;
    
    int i = 1;
    word_list_t * stream = list;
    while(i<params->max_length && stream!=NULL) {
        stream = stream->next;
        i++;
    }
    
    if(i==params->max_length) {
        return i;
    }
    else {
        return 0;
    }
}

/* pure N-gram method */
/* cut when token </s> (or any cutable item) has the highest probability in 4-gram LM*/
/* return last word (not </s>) */
int segment_emit(parameters_t * params, word_list_t * list) {
    debug("starting");
    
    int len = 1;
    /* reverse context: trigram[0]=w_-1, trigram[1]=w_-2, trigram[2]=w_-3 */
    int32 bigram[3] = {0,0};
    bigram[0] = params->sentence_start_id;
    
    word_list_t * stream = word_list_get_first_word(list);
    while((params->max_length<=0 || len<=params->max_length) && stream!=NULL) {
        
        bigram[1] = bigram[0];
        bigram[0] = ngram_wid(params->lm,stream->word);
        ngram_iter_t * bigram_iter = ngram_ng_iter(params->lm,bigram[0],&(bigram[1]),(bigram[1]==0)?0:1);
        
        int32 best_score = params->lm_zero;
        int32 eos_score = params->lm_zero;
        ngram_iter_t * trigram_iter = (bigram_iter!=NULL)?ngram_iter_successors(bigram_iter):NULL;
        
        while(trigram_iter!=NULL) {
            int32 score = 0;
            int32 backoff = 0;
            const int32 * wids = ngram_iter_get(trigram_iter,&score,&backoff);
            
            if(score>eos_score && cut_on_wid(params,wids[0])) {
                eos_score = score;
            }
            if(score>best_score) {
                best_score = score;
            }
            trigram_iter = ngram_iter_next(trigram_iter);
        }
        if(eos_score>=best_score) {
            return len;
        }
        stream = word_list_get_first_word(stream->next);
        len += 1;
    }
    
    return 0;
}

/* pure N-gram method */
/* cut when token </s> (or any cutable item) has a higher probability in 4-gram LM than actual following word */
/* return last word (not </s>) */
/* it tends to make short segments (2 to 6 words, avg=2.5) */
int segment_prefer(parameters_t * params, word_list_t * list) {
    debug("starting");
    
    int len = 1;
    word_list_t * stream = word_list_get_first_word(list);
    
    /* reverse context: trigram[0]=w_i, trigram[1]=w_i-1, trigram[2]=w_i-2 */
    int32 trigram[3] = {0,0,0};
    trigram[1] = params->sentence_start_id;
    if(stream!=NULL) trigram[0] = ngram_wid(params->lm,stream->word);
    
    while((params->max_length<=0 || len<=params->max_length) && stream!=NULL) {
        
        word_list_t * next_word = word_list_get_first_word(stream->next);
        if(next_word==NULL) {
            stream = NULL;
            break;
        }
        
        trigram[2] = trigram[1];
        trigram[1] = trigram[0];
        trigram[0] = ngram_wid(params->lm,next_word->word);
        
        /* get score of actual item */
        int32 score = params->lm_zero;
        int32 n_used;
        if(trigram[1]==0)
            score = ngram_ng_score(params->lm,trigram[0],NULL,0,&n_used);
        else if(trigram[2]==0)
            score = ngram_ng_score(params->lm,trigram[0],&(trigram[1]),1,&n_used);
        else
            score = ngram_ng_score(params->lm,trigram[0],&(trigram[1]),2,&n_used);
        
        /* get score of breaker */
        int32 eos_score = params->lm_zero;
        int i = 0;
        while(params->mt_cut_symbols_id[i]!=params->unk_id) {
            if(trigram[1]==0)
                eos_score = max(eos_score,ngram_ng_score(params->lm,params->mt_cut_symbols_id[i],NULL,0,&n_used));
            else if(trigram[2]==0)
                eos_score = max(eos_score,ngram_ng_score(params->lm,params->mt_cut_symbols_id[i],&(trigram[1]),1,&n_used));
            else
                eos_score = max(eos_score,ngram_ng_score(params->lm,params->mt_cut_symbols_id[i],&(trigram[1]),2,&n_used));
            i++;
        }
        
        /* decide */
        if(eos_score>=score) {
            return len;
        }
        stream = next_word;
        len += 1;
    }
    
    if(len>=params->max_length && params->max_length>0) {
        return len;
    }
    
    return 0;
}

/* pure N-gram method */
/* if there exist a position within the window where */
/* token </s> (or any cutable item) has a higher probability in 3-gram LM than actual following word */
/* then cut after such a position where the probability of </s> is the highest */
int segment_best_prefer(parameters_t * params, word_list_t * list) {
    debug("starting");
    
    int len = 1;
    word_list_t * stream = word_list_get_first_word(list);
    
    /* reverse context: trigram[0]=w_i, trigram[1]=w_i-1, trigram[2]=w_i-2 */
    int32 trigram[3] = {0,0,0};
    trigram[1] = params->sentence_start_id;
    if(stream!=NULL) trigram[0] = ngram_wid(params->lm,stream->word);
    
    int best_cut = 0;
    int32 best_cut_score = params->lm_zero;
    
    while((params->max_length<=0 || len<=params->max_length) && stream!=NULL) {
        
        word_list_t * next_word = word_list_get_first_word(stream->next);
        if(next_word==NULL) {
            break;
        }
        
        trigram[2] = trigram[1];
        trigram[1] = trigram[0];
        trigram[0] = ngram_wid(params->lm,next_word->word);
        
        /* get score of actual item */
        int32 score = params->lm_zero;
        int32 n_used;
        if(trigram[1]==0)
            score = ngram_ng_score(params->lm,trigram[0],NULL,0,&n_used);
        else if(trigram[2]==0)
            score = ngram_ng_score(params->lm,trigram[0],&(trigram[1]),1,&n_used);
        else
            score = ngram_ng_score(params->lm,trigram[0],&(trigram[1]),2,&n_used);
        
        /* get score of breaker */
        int32 eos_score = params->lm_zero;
        int i = 0;
        while(params->mt_cut_symbols_id[i]!=params->unk_id) {
            if(trigram[1]==0)
                eos_score = max(eos_score,ngram_ng_score(params->lm,params->mt_cut_symbols_id[i],NULL,0,&n_used));
            else if(trigram[2]==0)
                eos_score = max(eos_score,ngram_ng_score(params->lm,params->mt_cut_symbols_id[i],&(trigram[1]),1,&n_used));
            else
                eos_score = max(eos_score,ngram_ng_score(params->lm,params->mt_cut_symbols_id[i],&(trigram[1]),2,&n_used));
            i++;
        }
        
        /* decide */
        if(eos_score>=score && eos_score>=best_cut_score) {
            best_cut = len;
            best_cut_score = eos_score;
        }
        stream = next_word;
        len += 1;
    }
    if(len>=params->max_length && best_cut!=0) {
        return best_cut;
    }
    
    return 0;
}

int segment_at_best(parameters_t * params, word_list_t * list) {
    debug("starting");
    
    int len = 1;
    word_list_t * stream = word_list_get_first_word(list);
    int best_cut = 0;
    int32 best_cut_score = params->lm_zero;
    
    /* reverse context: bigram[0]=w_i, bigram[1]=w_i-1*/
    int32 bigram[2] = {0,0};
    bigram[0] = params->sentence_start_id;
    
    while((params->max_length<=0 || len<=params->max_length) && stream!=NULL) {
        
        bigram[1] = bigram[0];
        bigram[0] = ngram_wid(params->lm,stream->word);
        
        /* get score of breaker */
        int n_used = 0;
        int32 score = params->lm_zero;
        int i = 0;
        while(params->mt_cut_symbols_id[i]!=params->unk_id) {
            if(bigram[0]==0)
                score = score,ngram_ng_score(params->lm,params->mt_cut_symbols_id[i],NULL,0,&n_used);
            else if(bigram[1]==0)
                score = score,ngram_ng_score(params->lm,params->mt_cut_symbols_id[i],bigram,1,&n_used);
            else
                score = score,ngram_ng_score(params->lm,params->mt_cut_symbols_id[i],bigram,2,&n_used);
            if(score>=best_cut_score) {
                best_cut_score = score;
                best_cut = len;
            }
            i++;
        }
        
        stream = word_list_get_first_word(stream->next);
        len += 1;
    }
    
    if(len>=params->max_length && params->max_length>0 && best_cut!=0) {
        return best_cut;
    }
    
    return 0;
}

/* mix acoustic and N-gram method, with maximum chunk length */
/* cut if cutable symbol more likely than following word */
/* if none found, cut on first silence */
/* if that fails, cut where cutable symbol is most likely in the range [0,max_chunk] */
/* return last word (not </s>) */
int segment_mix(parameters_t * params, word_list_t * list) {
    debug("starting");
    
    int chunk = segment_best_prefer(params,list);
    if(chunk>0 || word_list_length(list)<params->max_length) return chunk; /* don't feel compelled to cut if max length not reached */
    
    chunk = segment_silence(params,list);
    if(chunk>0) return chunk;
    
    chunk = segment_at_best(params,list);
    if(chunk>0) return chunk;
    
    return 0;
}


int word_list_length(word_list_t * stream) {
    int len = 0;
    word_list_t * current = stream;
    while(current!=NULL) {
        current = current->next;
        len += 1;
    }
    return len;
}

word_list_t * word_list_get_first_word(word_list_t * stream) {
    word_list_t * current = stream;
    while(current!=NULL && is_non_word(current->word))
        current = current->next;
    return current;
}

int is_non_word(char * word) {
    if(word==NULL)
        return 1;
    size_t len = strlen(word);
    if((len>=4 && word[0]=='+' && word[1]=='+' && word[len-2]=='+' && word[len-1]=='+')
        || (!strcmp(word,"<sil>") || !strcmp(word,"SIL"))) {
        return 1;
    }
    else {
        return 0;
    }
}

int is_silence(char * word) {
    if(word==NULL)
        return 0;
    if(!strcmp(word,"<sil>") || !strcmp(word,"SIL")) {
        return 1;
    }
    else {
        return 0;
    }
}


char * clean(char * word) {
    /* tokenize */
    /* replace "'" with " ' " */
    /* and "_" with " " */
    
    /* compute length of tokenized string */
    size_t len = strlen(word);
    size_t new_len = 1; /* init with space for \0 */
    size_t i;
    for(i=0;i<len;i++) {
        if(word[i]=='\'' && i>0) {
            new_len += (i<len && word[i+1]!=' ')?3:2;
        }
        else if(word[i]=='(') {
            break;
        }
        else {
            new_len += 1;
        }
    }
    /* allocate new string */
    char * tok = malloc(new_len*sizeof(char));
    /* tokenize  string */
    int tok_ptr = 0;
    for(i=0;i<len;i++) {
        if(word[i]=='\'' && i>0) {
            tok[tok_ptr++] = ' ';
            tok[tok_ptr++] = '\'';
            if(i<len && word[i+1]!=' ')
                tok[tok_ptr++] = ' ';
        }
        else if(word[i]=='_') {
            tok[tok_ptr++] = ' ';
        }
        else if(word[i]=='(') {
            tok[tok_ptr] = '\0';
            break;
        }
        else {
            tok[tok_ptr++] = word[i];
        }
    }
    tok[new_len-1] = '\0';
    
    return tok;
}
