/*
*
* Copyright (C) 2009, Sylvain Raybaud <sylvain.raybaud@crans.org>
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of Sylvain Raybaud nor the names of its contributors may
*       be used to endorse or promote products derived from this software
*       without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY Sylvain Raybaud <sylvain.raybaud@crans.org> ``AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Sylvain Raybaud BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
*/

#ifndef _mycutils_h_
#define _mycutils_h_

#ifdef __cplusplus
extern "C" {
#endif

#include <stdio.h>
#include <stdarg.h>
#include <zlib.h>
#if defined(MSDOS) || defined(OS2) || defined(WIN32) || defined(__CYGWIN__)
#  include <fcntl.h>
#  include <io.h>
#  define SET_BINARY_MODE(file) setmode(fileno(file), O_BINARY)
#else
#  define SET_BINARY_MODE(file)
#endif
#define CHUNK 262144

#ifndef MAXGZLINELENGTH
    #define MAXGZLINELENGTH 65000
#endif

#define max(a,b) (((a)>(b))?(a):(b))
#define min(a,b) (((a)<(b))?(a):(b))

typedef struct {

    /* flag: is this file open? */
    int open;

    /* flag: is the file a gziped or a plain file ? */
    int gziped;

    /* depending on the flag at least one of them shall be NULL */
    FILE * file;
    gzFile gzfile;

    /* if something went wrong */
    int error;

    /* reached EOF */
    int eof;

    /* retain file name*/
    char * filename;

} myFile;

/* if filename ends with ".gz" then open it through zlib o/w through stdio */
myFile myfopen(const char * filename, const char * mode);

/* similar to fclose except it transparently handles gziped and regular files */
#define myfclose(file_) do { free(file_.filename); if(file_.open && file_.gziped) { gzclose(file_.gzfile); } else if(file_.open) { fclose(file_.file); } } while(0)

/* similar to fread except it transparently handles gziped and regular files */
//#define myfread(file_,ptr,size,n) ((file_.gziped)?gzread(file_.gzfile,ptr,(size)*(n)):fread(ptr,size,n,file_.file))
size_t myfread(myFile f,void * ptr,size_t size, size_t n);

int myfflush(myFile f);

/* similar to fprintf except it transparently handles gziped and regular files */
#define myfprintf(file_,format,...) (file_.gziped?gzprintf(file_.gzfile,format,__VA_ARGS__):fprintf(file_.file,format,__VA_ARGS__))

/* similar to fscanf except it transparently handles gziped and regular files */
#define myfscanf(file_,format,...) (file_.gziped?mygzscanf(file_.gzfile,format,__VA_ARGS__):fscanf(file_.file,format,__VA_ARGS__))
int mygzscanf(gzFile file, const char * format, ...);

/* similar to getline except it transparently handles gziped or regular files */
ssize_t mygetline(char ** lineptr, size_t * length, myFile file);
/* read a line and dismiss blank or hash-prefixed lines */
ssize_t myclevergetline(char ** lineptr, size_t * length, myFile file);

// #ifdef _PTHREAD_H
/* implement thread safe verbose and error messages */
//
// pthread_mutex_t cutils_stderr_mutex = PTHREAD_MUTEX_INITIALIZER;
//
// #define myfail(code,...) do { pthread_mutex_lock(&cutils_stderr_mutex);
// fprintf(stderr,"*** ERROR *** [%s] ", __func__);
// fprintf(stderr,__VA_ARGS__);
// fprintf(stderr,"\n");
// pthread_mutex_unlock(&cutils_stderr_mutex);
// exit(code); }  while(0)
//
// #define myerror(...) do {
// pthread_mutex_lock(&cutils_stderr_mutex);
// fprintf(stderr,"*** ERROR *** [%s] ", __func__);
// fprintf(stderr,__VA_ARGS__);
// fprintf(stderr,"\n");
// pthread_mutex_unlock(&cutils_stderr_mutex); }  while(0)
//
// #define mywarning(...) do { pthread_mutex_lock(&cutils_stderr_mutex);
// fprintf(stderr,"*** WARNING *** [%s] ", __func__);
// fprintf(stderr,__VA_ARGS__);
// fprintf(stderr,"\n");
// pthread_mutex_unlock(&cutils_stderr_mutex); }  while(0)
//
// #define myinfo(...) do { pthread_mutex_lock(&cutils_stderr_mutex);
// fprintf(stderr,"*** INFO *** ");
// fprintf(stderr,__VA_ARGS__);
// fprintf(stderr,"\n");
// pthread_mutex_unlock(&cutils_stderr_mutex); }  while(0)
//
// #define myverbose(verb,...) do {
// if(verb) { pthread_mutex_lock(&cutils_stderr_mutex);
//     fprintf(stderr,">>> ");
//     fprintf(stderr,__VA_ARGS__); fprintf(stderr,"\n");
//     pthread_mutex_unlock(&cutils_stderr_mutex); }
// } while(0)
//
// #ifdef DEBUG
// #define debug(...) do { pthread_mutex_lock(&cutils_stderr_mutex);
// fprintf(stderr, "[%s(%d)] %s: ", __FILE__, __LINE__, __func__);
// fprintf(stderr,  __VA_ARGS__);
// fputs("\n", stderr);
// pthread_mutex_unlock(&cutils_stderr_mutex); } while(0)
// #else
/* no debug */
// #define debug(...)
// #endif
//
/* _PTHREAD_H */
// #else

#define myfail(code,...) do { fprintf(stderr,"*** ERROR *** [%s] ", __func__); fprintf(stderr,__VA_ARGS__); fprintf(stderr,"\n"); exit(code); }  while(0)
#define myerror(...) do { fprintf(stderr,"*** ERROR *** [%s] ", __func__); fprintf(stderr,__VA_ARGS__); fprintf(stderr,"\n"); }  while(0)
#define mywarning(...) do { fprintf(stderr,"*** WARNING *** [%s] ", __func__); fprintf(stderr,__VA_ARGS__); fprintf(stderr,"\n"); }  while(0)
#define myinfo(...) do { fprintf(stderr,"*** INFO *** "); fprintf(stderr,__VA_ARGS__); fprintf(stderr,"\n"); }  while(0)
#define myverbose(verb,...) do { if(verb) { fprintf(stderr,__VA_ARGS__); fprintf(stderr,"\n"); } } while(0)

#ifdef DEBUG
#define debug(...) do { \
fprintf(stderr, "[%s(%d)] %s: ", __FILE__, __LINE__, __func__); \
fprintf(stderr,  __VA_ARGS__); \
fputs("\n", stderr); \
} while(0)
#else
#define debug(...) /* no debug */
#endif

//
/* _PTHREAD_H */
// #endif
//

#ifdef __cplusplus
}
#endif

#endif /* _mycutils_h_ */
