
#include "language_model.h"
#include "common.h"

#include <logmath.h>

#include <cutils.h>

#include <pthread.h>
#include <stdio.h>
#include <math.h>

/*
 * Functions for querying Sphinx's implementation of language models.
 */

float sphinx_score_unigram(lmi_data_t data, wid_t a) {
  sphinx_lm_t *lm = (sphinx_lm_t*) data;
  int32 n_used;
  return logmath_exp(lm->lmath, ngram_ng_prob(lm->lm, a.idx, NULL, 0, &n_used));
}

float sphinx_score_bigram(lmi_data_t data, wid_t a, wid_t b) {
  sphinx_lm_t *lm = (sphinx_lm_t*) data;
  int32 hist[1];
  hist[0] = a.idx;
  int32 n_used;
  return logmath_exp(lm->lmath,ngram_ng_prob(lm->lm, b.idx, hist, 1, &n_used));
}

float sphinx_score_trigram(lmi_data_t data, wid_t a, wid_t b, wid_t c) {
  sphinx_lm_t *lm = (sphinx_lm_t*) data;
  int32 hist[2];
  hist[0] = b.idx;
  hist[1] = a.idx;
  int32 n_used;
  return logmath_exp(lm->lmath,ngram_ng_prob(lm->lm, c.idx, hist, 2, &n_used));
}

float sphinx_score_ngram(lmi_data_t data, int n, wid_t *t) {
  sphinx_lm_t *lm = (sphinx_lm_t*) data;
  int i;
  int32 *hist = malloc((n-1)*sizeof(int));
  for (i=0 ; i<n-1 ; i++)
    hist[n-i-1] = t[i].idx;
  int32 n_used;
  float p = logmath_exp(lm->lmath,ngram_ng_prob(lm->lm, t[n-1].idx, hist, n-1, &n_used));
  free(hist);
  return p;
}

int sphinx_unigram_exists(lmi_data_t data, wid_t a) {
  sphinx_lm_t *lm = (sphinx_lm_t*) data;
  int32 n_used = 0; /* must be initialised (may not be set) */
  ngram_ng_prob(lm->lm, a.idx, NULL, 0, &n_used);
  return n_used==1;
}

int sphinx_bigram_exists(lmi_data_t data, wid_t a, wid_t b) {
  sphinx_lm_t *lm = (sphinx_lm_t*) data;
  int32 hist[1];
  hist[0] = a.idx;
  int32 n_used = 0;
  ngram_ng_prob(lm->lm, b.idx, hist, 1, &n_used);
  return n_used==2;
}

int sphinx_trigram_exists(lmi_data_t data, wid_t a, wid_t b, wid_t c) {
  sphinx_lm_t *lm = (sphinx_lm_t*) data;
  int32 hist[2];
  hist[0] = b.idx;
  hist[1] = a.idx;
  int32 n_used = 0;
  ngram_ng_prob(lm->lm, c.idx, hist, 2, &n_used);
  return n_used==3;
}

wid_t sphinx_wid(lmi_data_t data, char *word) {
  wid_t wid;
  wid.idx = ngram_wid(((sphinx_lm_t*)data)->lm, word);
  return wid;
}

void sphinx_free_data(lmi_data_t data) {
  sphinx_lm_t * _data = (sphinx_lm_t*)data;
  logmath_free(_data->lmath);
  ngram_model_free(_data->lm);
}

lmi_t sphinx_lmi_new(sphinx_lm_t* data) {
  lmi_t sphinx = {
    sphinx_score_unigram,
    sphinx_score_bigram,
    sphinx_score_trigram,
    sphinx_score_ngram,
    sphinx_unigram_exists,
    sphinx_bigram_exists,
    sphinx_trigram_exists,
    sphinx_wid,
    sphinx_free_data,
    (lmi_data_t) data,
    sphinx_wid(data, "<s>"),
    sphinx_wid(data, "</s>")
  };
  return sphinx;
}

sphinx_lm_t *sphinx_open_lm(char *filename, logmath_t *lmath) {
  /* alloc */
  sphinx_lm_t *lmi = malloc(sizeof(sphinx_lm_t));
  /* init / loading */
  lmi->lmath = logmath_retain(lmath);
  lmi->lm = 
    ngram_model_read(NULL /*options*/,
		     filename,
		     NGRAM_AUTO,
		     lmi->lmath);
  pthread_mutex_init(&lmi->mutex, NULL);
  /* error check */
  if (lmi->lm == NULL)
    myfail(EXIT_PS_ERROR, "error loading Sphinx ngram_model_t from file '%s'", filename);
  return lmi;
}
