/*
Copyright (c) 2011, Sylvain Raybaud <sylvain.raybaud@gmail.com>
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
Neither the name of the PAROLE team, LORIA, INRIA nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdio.h>
#include <glib.h>
#include <pthread.h>
#include <sys/time.h>
#include <time.h>

#include <ngram_model.h>
#include <pocketsphinx.h>

#include <cutils.h>

#include "engine.h"
#include "common.h"
#include "sentence.h"
#include "maparole.h"
#include "interface_ps.h"
#include "common.h"
#include "language_model.h"


/* Initialize a Sphinx decoder with cmd_ln and store it at '*ps' */
static void init_ps_decoder(ps_decoder_t ** ps, ps_config_t * ps_config, cmd_ln_t * cmd_ln) {
    debug("start");
    
    *ps = ps_init(cmd_ln);
    if (*ps == NULL) {
      myerror("Could not initialize PocketSphinx decoder");
      exit(EXIT_PS_ERROR);
    }
    ps_config->lm = ps_update_lmset(*ps,ngram_model_retain(ps_config->lm));
}

/* Initialize a Sphinx pronunciation dictionary */
//TODO: there must be a way to get dict from decoder object instead of reloading it!
static void init_ps_dict(dict_t ** dict, bin_mdef_t ** bin_mdef, decoder_config_t cfg) {
    debug("start");
    
    *bin_mdef = bin_mdef_read_text(cfg.ps_cmd_ln,cfg.mdef_filename); //should this be retained until dict is freed?
    *dict = dict_init(cfg.ps_cmd_ln,*bin_mdef);
    
}

static char * stream_to_string(asr_stream_t * from, int length) {
    /* debug("start"); */
    
    if(from==NULL) {
        return NULL;
    }
    
    int nb_words = 1;
    size_t len = 1;
    char * str = malloc(len*sizeof(char));
    str[len-1] = '\0';
    asr_stream_t * stream = from;
    while((length<0 || nb_words<=length) && stream!=NULL) {
        if(stream->word!=NULL && !stream->non_word && strcmp(stream->word,"euh")) {  /* do this properly! */
            len += strlen(stream->word)+1;
            str = realloc(str,len*sizeof(char));
            strcat(str,stream->word); str[len-2] = ' '; str[len-1] = '\0';
        }
        stream = stream->next;
        nb_words += 1;
    }
    return str;
}


mp_engine_t mp_start_engine(mp_config_t * cfg) {
    debug("starting");

    mp_engine_t engine = {0};

    /* start engine */

    init_ps_lm(cfg);

    if(cfg->ps_config.lm==NULL) {
        myfail(EXIT_PS_ERROR,"could not initialise LM set");
    }

    int i;
    engine.decoder_count = cfg->ps_config.decoders.count;
    engine.decoder_list = malloc(engine.decoder_count * sizeof(ps_decoder_t*));
    engine.dict_list = malloc(engine.decoder_count * sizeof(dict_t*));
    engine.bin_mdef_list = malloc(engine.decoder_count * sizeof(bin_mdef_t*));
    for (i=0 ; i<engine.decoder_count ; i++) {
        
        init_ps_decoder(&engine.decoder_list[i],
                        &cfg->ps_config,
                        cfg->ps_config.decoders.cfg[i].ps_cmd_ln);
        init_ps_dict(&engine.dict_list[i],
                     &engine.bin_mdef_list[i],
                     cfg->ps_config.decoders.cfg[i]);
    }
    
    time(&(engine.time_started));
    struct tm * date = localtime(&(engine.time_started));
    engine.timestamp = malloc(20*sizeof(char));
    sprintf(engine.timestamp,"%i.%.2i.%.2i-%.2i.%.2i.%.2i",date->tm_year,date->tm_mon,date->tm_mday,date->tm_hour,date->tm_min,date->tm_sec);

    return engine;
}

void mp_terminate(mp_config_t * config, mp_engine_t * engine) {
    debug("starting");

    free(config->inputfilenames);
    free(config->ps_config.lms);
    free(engine->timestamp);
    ps_terminate(&(config->ps_config),engine);
}

void ps_terminate(ps_config_t * cfg,mp_engine_t * engine) {
    debug("starting");
    int i;
    /* debug("ps LMs"); */
    /* int i; */
    /* ngram_model_free(cfg->lm); */
    /* for(i=0;i<cfg->n_lm;i++) { */
    /*     ngram_model_free(cfg->lms[i]); */
    /* } */
    for (i=0 ; i<cfg->n_lm ; i++) {
      free(cfg->lmset[i]);
      free(cfg->lmnames[i]);
    }
    free(cfg->lmset);
    free(cfg->lmnames);
    free(cfg->lmweights);
    logmath_free(cfg->lmath);
    for (i=0 ; i<engine->decoder_count ; i++) {
        ps_free(engine->decoder_list[i]);
        dict_free(engine->dict_list[i]);
        bin_mdef_free(engine->bin_mdef_list[i]);
    }
    free(engine->decoder_list);
    free(engine->dict_list);
    free(engine->bin_mdef_list);
    
    for (i=0 ; i<cfg->decoders.count ; i++) {
        free(cfg->decoders.cfg[i].hmm_filename);
        free(cfg->decoders.cfg[i].dict_filename);
        free(cfg->decoders.cfg[i].fdict_filename);
        free(cfg->decoders.cfg[i].mdef_filename);
    }
    free(cfg->decoders.cfg);
    ngram_model_free(cfg->lm);
}

void mp_start_translation(mp_config_t * cfg, mp_engine_t * engine) {
    debug("starting");
    myverbose(cfg->verbose,"starting translation manager");
    
    int n = 0;
    while(n<cfg->n_input) {
        
        manager_data_t pool = manager_data_init(cfg,engine,cfg->inputfilenames==NULL?NULL:cfg->inputfilenames[n]);

        pthread_t asr_buffer_manager_thread;
        pthread_t asr_manager_thread;
        pthread_t asr_stream_manager_thread;
        pthread_t segment_manager_thread;
        
        int asr_buffer_manager_created  = pthread_create(&asr_buffer_manager_thread, NULL,  asr_buffer_manager, (void*)&pool);
        int asr_manager_created         = pthread_create(&asr_manager_thread,        NULL,  asr_manager,        (void*)&pool);
        int asr_stream_manager_created  = pthread_create(&asr_stream_manager_thread, NULL,  asr_stream_manager, (void*)&pool);
        int segment_manager_created     = pthread_create(&segment_manager_thread,    NULL,  segment_manager,    (void*)&pool);
        
        if(asr_buffer_manager_created || asr_manager_created || asr_stream_manager_created || segment_manager_created) {
            manager_data_free(&pool);
            myfail(EXIT_UNKNOWN_ERROR,"Could not start threads.");
        }
        
        
        /* end */
        myverbose(cfg->verbose,"[engine] all threads launched, waiting for threads to finish.");
        
        pthread_join(asr_buffer_manager_thread,NULL);
        pthread_join(asr_manager_thread,NULL);
        pthread_join(asr_stream_manager_thread,NULL);
        pthread_join(segment_manager_thread,NULL);
        
        myverbose(cfg->verbose,"freeing pool");
        manager_data_free(&pool);
        
        n += 1;
    }
    
}

manager_data_t manager_data_init(mp_config_t * cfg, mp_engine_t * engine, const char * input) {
    debug("starting");
    myverbose(cfg->verbose,"populating pool");
    manager_data_t pool = {0};
    decoders_set_t * deco = &pool.decoders;
    
    pool.in_filename = input;

    pool.mp_config = cfg;
    pool.mp_engine = engine;

    pool.lm = cfg->ps_config.lm;
    pool.sentence_start_id = ngram_wid(pool.lm,"<s>");
    pool.lm_zero = ngram_zero(pool.lm);
    pthread_mutex_init(&pool.lm_mutex,NULL);
    

    deco->count = pool.mp_engine->decoder_count;
    deco->list = malloc(deco->count * sizeof (deco->list[0]));
    
    int i;
    deco->time_to_die = 0;
    for (i=0 ; i<deco->count ; i++) {
      deco->list[i].instance = engine->decoder_list[i];
      /* semaphores for parallel decoding synchronisation */
      sem_init(&deco->list[i].sem_start, 0 /*pshared*/, 0 /*value*/);
      sem_init(&deco->list[i].sem_finish, 0 /*pshared*/, 0 /*value*/);
    }

    /* data for audio */
    
    pool.n_channel = 1; /* no choice yet */

    pool.input_buffer_size = cfg->asr_buffer_size;

    //TODO: check all models have compatible parameters
    fe_t * fe = ps_get_fe(deco->list[0].instance);
    fe_get_input_size(fe,&pool.samples_frame_shift,&pool.samples_per_frame);
    
    pool.sampling_rate = cmd_ln_float_r(cfg->ps_config.decoders.cfg[0].ps_cmd_ln,"-samprate");
    /* convert durations to sizes (number of samples (int16)). make sure the frame count is an integer */
    pool.asr_segment_size = pool.sampling_rate*cfg->asr_segment_length;
    pool.asr_segment_size += pool.samples_frame_shift-((pool.asr_segment_size-pool.samples_per_frame)%pool.samples_frame_shift); /* round number of frames */
    pool.asr_segment_overlap = pool.sampling_rate*cfg->asr_segment_overlap;
    if(pool.asr_segment_overlap>(pool.asr_segment_size/2)) pool.asr_segment_overlap = pool.asr_segment_size/2;
    pool.segment_time_shift = (float32)(pool.asr_segment_size-pool.asr_segment_overlap)/pool.sampling_rate;
    pool.segment_time_offset = 0;

    pool.input_buffer = g_queue_new();
    pthread_mutex_init(&pool.input_buffer_mutex,NULL);
    pthread_cond_init(&pool.input_buffer_one_pushed,NULL);
    pthread_cond_init(&pool.asr_hyp_queue_one_hyp_pushed,NULL);

    pool.asr_hyp_queue = g_queue_new();
    pthread_mutex_init(&pool.asr_hyp_queue_mutex,NULL);
    pthread_cond_init(&pool.input_buffer_one_poped,NULL);

    pool.asr_stream = NULL;
    pthread_mutex_init(&pool.asr_stream_mutex,NULL);
    pthread_cond_init(&pool.asr_stream_updated,NULL);
    
    myverbose(cfg->verbose,"pool info:");
    myverbose(cfg->verbose,"  sampling rate = %fHz",pool.sampling_rate);
    myverbose(cfg->verbose,"  frame shift = %i samples",pool.samples_frame_shift); 
    myverbose(cfg->verbose,"  asr buffer = %i segments",cfg->asr_buffer_size);
    myverbose(cfg->verbose,"  asr segment = %is (%li samples)",cfg->asr_segment_length,pool.asr_segment_size);
    myverbose(cfg->verbose,"  asr segment overlap = %is (%li samples)",cfg->asr_segment_overlap,pool.asr_segment_overlap);
    myverbose(cfg->verbose,"  frame size = %i samples",pool.samples_per_frame);
    myverbose(cfg->verbose,"  inter segments time shift = %fs",pool.segment_time_shift);
        
    
    /* verbosity */
    
    if(input==NULL) {
        myverbose(cfg->verbose,"translating speech from stdin (mic?)");
        pool.in.file = stdin;
        pool.in.filename = NULL;
    }
    else {
        myverbose(cfg->verbose,"translating speech from '%s'",input);
        pool.in = myfopen(input,"rb");
    }
    char * outfilename = gen_filename(input,cfg->prefix,(cfg->compress_out)?".log.gz":".log",(input==NULL)?engine->timestamp:NULL);
    myverbose(cfg->verbose,"writing output to '%s'",outfilename);
    pool.out = myfopen(outfilename,"a");
    free(outfilename);
    
    pool.save_ctm_filename = gen_filename(input,cfg->prefix,(cfg->compress_out)?".ctm.gz":".ctm",(input==NULL)?engine->timestamp:NULL);
    myverbose(cfg->verbose,"saving CTM to '%s'",pool.save_ctm_filename);
    
    if(pool.mp_config->phonemes) {
        pool.save_phonemes_filename = gen_filename(input,cfg->prefix,(cfg->compress_out)?".phonemes.ctm.gz":".phonemes.ctm",(input==NULL)?engine->timestamp:NULL);
        myverbose(cfg->verbose,"saving phonemes CTM to '%s'",pool.save_phonemes_filename);
    }

    return pool;
}


void manager_data_free(manager_data_t * pool) {
    debug("starting");
    decoders_set_t * deco = &pool->decoders;

    myfclose(pool->out);
    if(pool->mp_config->inputfilenames!=NULL)
        myfclose(pool->in);

    free(pool->save_ctm_filename);
    free(pool->save_phonemes_filename);
    
    if(pool->asr_stream!=NULL && pool->asr_stream->word!=NULL) {
        mywarning("there are tokens left in asr_stream");
        asr_stream_free(pool->asr_stream);
    }
    
	while (! g_queue_is_empty(pool->input_buffer)) {
        audio_chunk_free((audio_chunk_t*)g_queue_pop_head(pool->input_buffer));
    }
    g_queue_free(pool->input_buffer);
    
    int i;
    for (i=0 ; i<deco->count ; i++) {
        sem_destroy(&deco->list[i].sem_start);
        sem_destroy(&deco->list[i].sem_finish);
    }
    free(deco->list);

}


void find_best_decoder(manager_data_t * pool, audio_chunk_t * chunk) {
    debug("starting");
    decoders_set_t * deco = &pool->decoders;
    int i;

    /* Extract a smaller chunk */
    float extract_duration = 3.0;	  /* s */
    float input_samprate = pool->sampling_rate; /* Hz */
    pool->audio_extract.size = extract_duration*input_samprate;
    pool->audio_extract.audio = realloc(pool->audio_extract.audio,
                                        pool->audio_extract.size * sizeof(pool->audio_extract.audio[0]));
    memcpy(pool->audio_extract.audio, chunk->audio,
           pool->audio_extract.size*sizeof(pool->audio_extract.audio[0]));

    /* Launch all decoders */
    for (i=0 ; i<deco->count ; i++)
        sem_post(&deco->list[i].sem_start);
    /* Wait for all results */
    for (i=0 ; i<deco->count ; i++)
        sem_wait(&deco->list[i].sem_finish);

    /* Select the decoder which gives the highest score. */
    int argmax=-1;
    float max=-1.;
    for (i=0 ; i<deco->count ; i++) {
        debug("decoder %d got score %f", i, deco->list[i].score);
        if (deco->list[i].score>max) {
            argmax = i;
            max = deco->list[i].score;
        }
    }
    deco->current = argmax;
    debug("current decoder is now #%d", deco->current);
}

/* Create one thread per decoder from function 'asr_decoder' */
static void start_decoder_threads(manager_data_t * pool) {
    debug("starting");
    decoders_set_t * deco = &pool->decoders;
    int i, err;
    for (i=0 ; i<deco->count ; i++) {
        deco->list[i].thread_data.pool = pool;
        deco->list[i].thread_data.thread_number = i;
        err = pthread_create(&deco->list[i].thread, NULL,
                asr_decoder, (void*)&deco->list[i].thread_data);
        if (err)
            myfail(EXIT_UNKNOWN_ERROR,"Could not start decoder thread number %d", i);
    }
}

/* Stop each thread created by 'start_decoder_threads' */
static void terminate_decoder_threads(manager_data_t * pool) {
    debug("starting");
    decoders_set_t * deco = &pool->decoders;
    int i;
    deco->time_to_die = 1;
    /* Wake up every thread */
    for (i=0 ; i<deco->count ; i++)
        sem_post(&deco->list[i].sem_start);
    /* Wait for every thread to die */
    for (i=0 ; i<deco->count ; i++)
        pthread_join(deco->list[i].thread, NULL);
}


/* Thread function.
   Each decoder has a thread, and each thread runs the decoder on the next audio,
   then stores the score in deco->scores[thread_number].
   Those threads are triggered by find_best_decoder() via sem_start,
   and find_best_decoder waits for them to complete via sem_finish. */
void * asr_decoder(void * v_data) {
    decoder_thread_data_t * data = (decoder_thread_data_t*) v_data;
    decoders_set_t * deco = &data->pool->decoders;
    ps_decoder_t * decoder = deco->list[data->thread_number].instance;

    debug("starting decoder number %d", data->thread_number);
    
    while (1) {
        sem_wait(&deco->list[data->thread_number].sem_start);

        if (deco->time_to_die)
            break;
        
        audio_chunk_t extract = data->pool->audio_extract;

        pthread_mutex_lock(&data->pool->lm_mutex);
        int start_utt = ps_start_utt(decoder,NULL);
        int nb_frames = ps_process_raw(decoder,extract.audio,extract.size,FALSE,TRUE);
        int end_utt = ps_end_utt(decoder);
        if (start_utt < 0 || end_utt < 0 || nb_frames < 0)
            myfail(EXIT_PS_ERROR, "Error processing raw audio extract in decoder %d", data->thread_number);
        
        int32 score;
        ps_get_hyp(decoder, &score, NULL);
        deco->list[data->thread_number].score = logmath_exp(data->pool->mp_config->ps_config.lmath, score);
        debug("Decoder number %d got score %f (for utterance: %s)",
        data->thread_number, deco->list[data->thread_number].score,
        ps_get_hyp(decoder, &score, NULL));
        pthread_mutex_unlock(&data->pool->lm_mutex);
        
        sem_post(&deco->list[data->thread_number].sem_finish);
    }
    debug("Decoder thread number %d: dying", data->thread_number);
    pthread_exit(NULL);
}

/* read chunks of audio and queue them */
void *asr_buffer_manager(void * data) {
    debug("starting");

    manager_data_t * pool = data;
    
    int16 * overlap = malloc(pool->asr_segment_overlap*sizeof(int16));

    int first_round = 1;
    int input_depleted = 0;
    while(!input_depleted) {

        int16 * audio = malloc(pool->asr_segment_size*sizeof(int16));
        size_t n_samples;
        if(first_round) {
            /* if this is the first round, there is no overlaping segment to copy */
            n_samples = myfread(pool->in,audio,sizeof(int16),pool->asr_segment_size);
        }
        else {
            /* first copy overlaping segment, then read input */
            memcpy(audio,overlap,sizeof(int16)*pool->asr_segment_overlap);
            n_samples = myfread(pool->in,&(audio[pool->asr_segment_overlap]),sizeof(int16),pool->asr_segment_size-pool->asr_segment_overlap);
            n_samples += pool->asr_segment_overlap;
        }
        
        if(n_samples<pool->asr_segment_size) {
            /* less data was read than requested. it means input is depleted. this was the last round */
            myverbose(pool->mp_config->verbose,"[buffer manager] input is depleted.");
            input_depleted = 1;
        }

        pthread_mutex_lock(&pool->input_buffer_mutex);
        if(g_queue_get_length(pool->input_buffer)>=pool->input_buffer_size) {
            myverbose(pool->mp_config->verbose,"[buffer manager] waiting for some room in the buffer");
            pthread_cond_wait(&pool->input_buffer_one_poped,&pool->input_buffer_mutex);
            myverbose(pool->mp_config->verbose,"[buffer manager] received signal to go");
        }
        
        if(n_samples>pool->asr_segment_overlap || (first_round && n_samples>0)) {
            /* new data was read. create chunk */
            audio_chunk_t * chunk = audio_chunk_new(audio,n_samples);
            g_queue_push_head(pool->input_buffer,chunk);
        }
        else {
            /* if chunk is created, it will be freed by asr manager when consumed. otherwise, it needs to be freed here. */
            free(audio);
        }

        if(input_depleted) {
            pool->input_buffer_manager_is_dead = 1;
        }
        /* signal the new chunk and release mutex */
        pthread_cond_signal(&pool->input_buffer_one_pushed);
        pthread_mutex_unlock(&pool->input_buffer_mutex);

        if(!input_depleted) {
            /* if this wasn't last round, save overlaping segment for next round */
            memcpy(overlap,&(audio[pool->asr_segment_size-pool->asr_segment_overlap]),sizeof(int16)*pool->asr_segment_overlap);
        }
    
	first_round = 0;
    }
    
    myverbose(pool->mp_config->verbose,"[buffer manager] bye.");
    free(overlap);
    pthread_exit(NULL);
}

/* for find_best_decoder() calling "strategy" */
static int asr_score_dropped(manager_data_t * pool) {
  float score_drop_threshold = 0.2; /* TODO tune the threshold */
  /* pool's fields are set by process_raw() */
  return (pool->penultimate_reco_score - pool->last_reco_score) > score_drop_threshold;
}

/* generate ASR utterances and queue them */
void *asr_manager(void * data) {
    debug("starting");

    manager_data_t * pool = data;
    int first_round = 1;
    start_decoder_threads(pool);
    
    int input_depleted = 0;
    while(!input_depleted) {
    
        pthread_mutex_lock(&pool->input_buffer_mutex);
        
        if(g_queue_is_empty(pool->input_buffer) && !pool->input_buffer_manager_is_dead) {
            myverbose(pool->mp_config->verbose,"[asr manager] waiting for a new audio chunk");
            pthread_cond_wait(&pool->input_buffer_one_pushed,&pool->input_buffer_mutex);
            myverbose(pool->mp_config->verbose,"[asr manager] received signal to go");
        }

        audio_chunk_t * chunk = (audio_chunk_t*)g_queue_pop_tail(pool->input_buffer);

        pthread_cond_signal(&pool->input_buffer_one_poped);
        pthread_mutex_unlock(&pool->input_buffer_mutex);

        if(chunk!=NULL) {
            /* process raw */

            /* Possibly choose another decoder */
            if (pool->decoders.count > 1) {
            /* TODO: insert a smart strategy here */
                if (first_round ||
                    asr_score_dropped(pool) ||
                    pool->last_utt_contains_noise || /* set by process_raw() */
                    pool->last_utt_contains_silence) { /* set by process_raw() */
                    
                        find_best_decoder(pool, chunk);
                }
            }

            translation_pair_t * t_item = translation_pair_new();
            translation_pair_set_audio(t_item,chunk);
            
            pthread_mutex_lock(&pool->lm_mutex); /* sphinx functions aren't thread-safe */
            
            int nb_frames = process_raw(pool,chunk);
            debug("processing raw");
            if(nb_frames<0) {
                audio_chunk_free(chunk);
                pthread_mutex_unlock(&pool->lm_mutex);
                asr_manager_terminate(pool);
                myerror("something went wrong while decoding audio");
                pthread_exit(NULL);
            }
            debug("getting asr result");
            translation_pair_get_asr_result(t_item,pool);

            pool->last_utt_contains_noise = 0;
            pool->last_utt_contains_silence = 0;
            asr_stream_t *s = translation_pair_get_src_stream(t_item);
            
            if (s != NULL) {
                
                float t_start, t_end;
                t_start = s->t_start;
                myfprintf(pool->out, "%sRecognized:\n", "");
                while (s != NULL) {
                    myfprintf(pool->out, "%s ", s->word);
                    if (is_silence(s))
                        pool->last_utt_contains_silence = 1;
                    if (dict_filler_word(pool->mp_engine->dict_list[pool->decoders.current],s->wid))
                        pool->last_utt_contains_noise = 1;
                    t_end = s->t_end;
                    s = s->next;
                }
                
                myfprintf(pool->out, "\n(score: %f; time: from t=%.3fs to t=%.3fs)\n\n",
                pool->last_reco_score, t_start, t_end);
                
            }
            myfflush(pool->out);

            char * string = stream_to_string(translation_pair_get_src_stream(t_item),-1);
            myverbose(pool->mp_config->verbose,"[asr manager] got utterance '%s' (score: %f)",string, pool->last_reco_score);
            free(string);
            
            pool->segment_time_offset += pool->segment_time_shift;

            pthread_mutex_unlock(&pool->lm_mutex);
            
            pthread_mutex_lock(&pool->asr_hyp_queue_mutex);
            g_queue_push_head(pool->asr_hyp_queue,t_item);
            pthread_cond_signal(&pool->asr_hyp_queue_one_hyp_pushed);
            pthread_mutex_unlock(&pool->asr_hyp_queue_mutex);
        
        }
        else {
            /* buffer manager could not fill the buffer, it means input is depleted */
            myverbose(pool->mp_config->verbose,"[asr manager] input is depleted.");
            asr_manager_terminate(pool);
            input_depleted = 1;
        }
    first_round = 0;
    }
    terminate_decoder_threads(pool);
    myverbose(pool->mp_config->verbose,"[asr manager] bye.");
    pthread_exit(NULL);
}


/* get ASR utterances and merge them into stream */
void *asr_stream_manager(void * data) {
    debug("starting");

    manager_data_t * pool = data;
    
    int input_depleted = 0;
    while(!input_depleted) {
        pthread_mutex_lock(&pool->asr_hyp_queue_mutex);
        if(g_queue_get_length(pool->asr_hyp_queue)<=1 && !pool->asr_manager_is_dead) {
            myverbose(pool->mp_config->verbose,"[stream manager] waiting for a new utterance");
            pthread_cond_wait(&pool->asr_hyp_queue_one_hyp_pushed,&pool->asr_hyp_queue_mutex);
            myverbose(pool->mp_config->verbose,"[stream manager] received signal to go");
        }
        
        translation_pair_t * t_item = (translation_pair_t*)g_queue_pop_tail(pool->asr_hyp_queue);
        pthread_mutex_unlock(&pool->asr_hyp_queue_mutex);
        pthread_mutex_lock(&pool->asr_stream_mutex);
        if(t_item==NULL) {
            /* asr manager could not fill the queue, it means input is depleted */
            myverbose(pool->mp_config->verbose,"[stream manager] input is depleted.");
            pool->asr_stream_manager_is_dead = 1;
            input_depleted = 1;
        }
        else {
            merge_utt_into_stream(pool,t_item->src_stream);
            char * str = stream_to_string(pool->asr_stream,-1);
            myverbose(pool->mp_config->verbose,"[stream manager] recognised so far and not yet translated:\n '%s'",str);
            free(str);
        }

        pthread_cond_signal(&(pool->asr_stream_updated));
        pthread_mutex_unlock(&pool->asr_stream_mutex);
    }
    myverbose(pool->mp_config->verbose,"[stream manager] bye.");
    pthread_exit(NULL);
}

/* extract mt segments from stream */
void *segment_manager(void * data) {
    debug("starting");

    manager_data_t * pool = data;
    
    int input_depleted = 0;
    while(!input_depleted) {
        pthread_mutex_lock(&pool->asr_stream_mutex);
        int chunk_length = 0;
        while((chunk_length = get_mt_segment_from_stream(pool))==0 && !pool->asr_stream_manager_is_dead) {
            myverbose(pool->mp_config->verbose,"[segment manager] waiting for a stream update");
            pthread_cond_wait(&pool->asr_stream_updated,&pool->asr_stream_mutex);
            myverbose(pool->mp_config->verbose,"[segment manager] received signal to go");
        }
        pthread_mutex_unlock(&pool->asr_stream_mutex);

        if(chunk_length==0)
            chunk_length = -1; /* if no segment found, but we are here, it means input is depleted. just take whatever remains */
        
        debug("writing ctm");
        write_ctm(pool,chunk_length);
        if(pool->mp_config->phonemes) write_phonemes_ctm(pool,chunk_length);
        
        pool->asr_stream = asr_stream_free_n(pool->asr_stream,chunk_length);
        if(pool->asr_stream!=NULL) {
            pool->asr_stream->prev = NULL;
        }

        if(chunk_length<=0) {
            /* input is depleted and stream is empty. die. */
            myverbose(pool->mp_config->verbose,"[segment manager] input is depleted.");
            input_depleted = 1;
        }
    }
    myverbose(pool->mp_config->verbose,"[segment manager] bye.");
    pthread_exit(NULL);
}
