/*
Copyright (c) 2011, Sylvain Raybaud <sylvain.raybaud@gmail.com>
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
Neither the name of the PAROLE team, LORIA, INRIA nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <math.h>
#include <glib.h>

#include <Manager.h>
#include <StaticData.h>
#include <Parameter.h>
#include <Sentence.h>
#include <WordLattice.h>
#include <TranslationSystem.h>
#include <Factor.h>
#include <Util.h>
using namespace Moses;

#include <vector>
#include <string>
using namespace std;

#include "interface_moses.h"
#include "interface_moses_cplusplus.h"
#include "sentence.h"
#include "maparole.h"
#include "engine.h"
#include "common.h"

/* shall return a pointer to a Param object */
int moses_read_config_file(const char * cfgfilename, moses_config_t * cfg) {
    Parameter * params = new Parameter();
    bool status = params->LoadParam(string(cfgfilename));
    cfg->moses_parameters = params;
    return status?1:0;
}

void moses_terminate(moses_config_t * config) {
  Parameter * params = (Parameter*) config->moses_parameters;
  delete params;
}

int moses_static_data(moses_config_t * config) {
    return StaticData::LoadDataStatic(((Parameter*)config->moses_parameters))?1:0;
}

void moses_translate(manager_data_t * pool, translation_pair_t * t) {
    if(!pool->mp_config->one_best) {
        myerror("can't translate lattices yet");
        return;
    }
    else {
      moses_translate_string(pool, t);
    }
}

void moses_translate_string(manager_data_t * pool, translation_pair_t * pair) {

    const TranslationSystem& system = StaticData::Instance().GetTranslationSystem(TranslationSystem::DEFAULT); /* there is only one translation system for now */
    const StaticData &staticData = StaticData::Instance();
    const vector<FactorType> &inputFactorOrder = staticData.GetInputFactorOrder();
    
    string source(pair->source->str);
    istringstream * in = new istringstream(source+"\n",istringstream::in);
    InputType * input = new Sentence(Input);
    input->Read(*in,inputFactorOrder);
    
    Manager * manager = new Manager(*input,staticData.GetSearchAlgorithm(), &system);
    manager->ProcessSentence();
    const Hypothesis* hypo = manager->GetBestHypothesis();

    string hyp = moses_get_hyp(hypo);
    char * hyp_ret = (char*)malloc((strlen(hyp.c_str())+1)*sizeof(char));
    strcpy(hyp_ret,hyp.c_str());

    pair->translation_score = UntransformScore(hypo->GetScore());
    translation_pair_set_target(pair, hyp.c_str());

    delete in;
    delete manager;
    delete input;
}

string moses_get_hyp(const Hypothesis* hypo) {
    
    string current("");
    TargetPhrase p = hypo->GetTargetPhrase();
    for (size_t pos = 0 ; pos<p.GetSize() ; pos++) {
        const Factor *factor = p.GetFactor(pos, 0);
        current += factor->GetString()+string(" ");
    }
    
    const Hypothesis * prev = hypo->GetPrevHypo();
    if(prev != NULL)
        return moses_get_hyp(prev)+string(" ")+current;
    return current;
}


PCN::CN slf2cn(slf_t *lat) {
    debug("start");
    
    slf_node_t ** ordered_nodes = slf_topological_sort(lat);
    int * pos_list = get_distances(lat,ordered_nodes);
    PCN::CN cn =  topo2pcn(lat,ordered_nodes,pos_list);
    
    delete [] pos_list;
    return cn;
    
}

PCN::CN topo2pcn(slf_t * lat, slf_node_t ** nodes, int * pos) {
    debug("start");
    
    GRegex * noise_re = mk_noise_re();
    
    PCN::CN cn;
    
    int i;
    for(i=0;i<(lat->nodes_count-1);i++) { /* remove last node because it has no exit */
        PCN::CNCol col;
        slf_links_list_t * alt = nodes[i]->exits;
        while(alt!=NULL && alt->link!=NULL) {
            vector<float> scores;
            scores.push_back(float(exp(alt->link->acoustic+lat->lmscale*alt->link->lm)));
            
            string word(match_noise(noise_re,alt->link->word)?"*EPS*":alt->link->word);
            pair<string,vector<float> > linkdata(word,scores);
            
            int distance = pos[alt->link->to]-i;
            if(distance<=0)
                mywarning("Non positive distance in topo2pcn. Shouldn't happen.");
            
            PCN::CNAlt cnalt(linkdata,size_t(distance));
            col.push_back(cnalt);
            alt = alt->tail;
        }
        cn.push_back(col);
    }
    
    g_regex_unref(noise_re);
    
    return cn;
}
