/*
Copyright (c) 2011, Sylvain Raybaud <sylvain.raybaud@gmail.com>
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
Neither the name of the PAROLE team, LORIA, INRIA nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef _maparole_h_
#define _maparole_h_

#include <pocketsphinx.h>

#include <cutils.h>

/** \file
 * Public header for maparole.
 */

#define SEG_NONE 0
#define SEG_EMIT 1
#define SEG_PREFER 2
#define SEG_BEST 3
#define SEG_BEST_PREFER 4
#define SEG_SILENCE 5
#define SEG_LENGTH 6
#define SEG_MIX 7
#define SEG_AUDIO 8

#ifdef __cplusplus
extern "C" {
#endif

/** Moses' configuration */
typedef struct moses_config_s {
    void * moses_parameters;
    char * lm_filename;
} moses_config_t;

/** Configuration options for a single Sphinx decoder. */
typedef struct decoder_config_s {
    cmd_ln_t * ps_cmd_ln; 	/**< pocketsphinx object representing options. It is stored here because it may be used in the code */
    char * hmm_filename; 	/**< The acoustic model directory  */
    char * mdef_filename;	/**< The mdef part of the acoustic model */
    char * dict_filename;	/**< Source language dictionary */
    char * fdict_filename;	/**< Filler dictionary */
} decoder_config_t;

/** Sphinx's configuration */
typedef struct ps_config_s {
    char ** lmset;
    char ** lmnames;
    float32 * lmweights;
    float64 lmbase;
    int n_lm;
    ngram_model_t * lm;
    ngram_model_t ** lms;
    logmath_t *lmath; /**< Sphinx's logmath, stored for use with several LMs */
    float32 lw;
    const char * lw_str;

    struct {
      decoder_config_t * cfg; 	/** Array of configurations */
      int count;		/** Size of the array */
    } decoders;	/**< Configuration for all the instanciated decoders */
} ps_config_t;

/** Configuration options for confidence measures.
    Used as a substructure of mp_config_t. */
typedef struct cm_config_s {
    /** Backward source language model */
    char * asr_backward_lm_filename;
    /** Backward target language model */
    char * mt_backward_lm_filename;

    /** Neural network to import for MT word level confidence measures. */
    char * mt_wlf_fann_filename;
    /** Neural network to import for MT sentence level confidence measures. */
    char * mt_slf_fann_filename;
    /** Neural network to import for ASR word level confidence measures. */
    char * asr_wlf_fann_filename;
    /** Neural network to import for ASR sentence level confidence measures. */
    char * asr_slf_fann_filename;
} cm_config_t;

/** Full configuration for maparole s2tt. */
typedef struct mp_config_s {
    
    const char ** inputfilenames;
    int n_input;
    int out_to_stdout;
    int compress_out;
    int lattice_input;
    int one_best;
    const char * prefix;
    
    ps_config_t ps_config;
    moses_config_t moses_config;
    cm_config_t cm_config;
    
    const char * cfg_s2tt_filename;
    const char * cfg_ps_filename;
    const char * cfg_moses_filename;
    
    int asr_segment_length;
    int asr_buffer_size;
    int asr_segment_overlap;
    
    int seg_method;
    int max_mt_chunk;
    
    int save_asr;
    int save_ctm;
    int save_raw_mt_seg;
    int save_clean_mt_seg;
    
    int verbose;
    
} mp_config_t;


/** engine */
typedef struct {
    ps_decoder_t ** decoder_list;	/**< list of the initialised decoders */
    int decoder_count;			/**< size of the list */

    time_t time_started;
    char * timestamp;
    
} mp_engine_t;

void mp_parse_cfg_s2tt(mp_config_t *, const char * cfgfilename);
void mp_parse_cfg_ps(mp_config_t *, const char * cfgfilename);
void mp_parse_cfg_moses(mp_config_t *, const char * cfgfilename);

int mp_check_ps_config(mp_config_t * cfg);
int mp_check_moses_config(mp_config_t * cfg);
int mp_check_s2tt_config(mp_config_t * cfg);

void mp_terminate(mp_config_t * config, mp_engine_t *);

mp_engine_t mp_start_engine(mp_config_t *);
void ps_terminate(ps_config_t * cfg,mp_engine_t * engine);

/* engine */
char const * mp_translate_audio_to_text(mp_config_t * config,mp_engine_t *);
char const * mp_translate_text_to_text(mp_config_t * config,mp_engine_t *);

void mp_start_translation(mp_config_t * cfg, mp_engine_t * engine);
void init_ps_lm(mp_config_t * config);

#ifdef __cplusplus
}
#endif

#endif /* _maparole_h_ */
