/*
Copyright (c) 2011, Sylvain Raybaud <sylvain.raybaud@gmail.com>
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
Neither the name of the PAROLE team, LORIA, INRIA nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <glib.h>

#include <pocketsphinx.h>

#include <cutils.h>

#include "common.h"
#include "parse_cl.h"
#include "maparole.h"

void parse_cl(mp_config_t * config, int argc, char ** argv) {
    
    static const char * usage = "Usage: %s <options> [input files] [< input audio] [> translated text]\nOptions are:\n\
  general options:\n\
    --compress or -z: compress output (no effect if -o is used)\n\
    --stdout or -o: output translated text to stdout instead of <file>.s2tt\n\
    --prefix or -P STRING: prefix for output files\n\
    --s2tt-config or -s FILE: speech to text translation configuration file (defaults to s2tt.cfg)\n\
    --verbose or -v: print out some details on stderr\n\
    --help or -h: print this message and exit\n\
  options specific to the ASR system (audio/lattice):\n\
    --pocketsphinx-config or -p PATH: pocketsphinx configuration file (defaults to ps.cfg)\n\
    --source-lm or -l file[:name=STRING][:[weight=]FLOAT]: source language model (for recognition). You can use this option many times to specify many LM.\n\
    --source-lm-weight or -w FLOAT: source language model weight, after all specified models have been interpolated (for recognition)\n\
    --lm-base or -b FILE: logmath base for lm component of sphinxbase (please only use this option if you know what you are doing) [default=1.0001]\n\
    --target-lm or -L FILE: target language model (for translation)\n\
    --source-backward-lm or -i FILE:  source backward language model (for confidence measures)\n\
    --target-backward-lm or -I FILE:  target backward language model (for confidence measures)\n\
    --source-sentlevel-nn or -x FILE:  neural network file for word level confidence measures\n\
    --target-sentlevel-nn or -X FILE:  neural network file for word level confidence measures\n\
    --source-wordlevel-nn or -y FILE:  neural network file for word level confidence measures\n\
    --target-wordlevel-nn or -Y FILE:  neural network file for word level confidence measures\n\
    --acoustic-model of -M hmm[:dict[:fdict[:mdef]]] : acoustic model files for a decoder\n\
    --save-asr or -A: save raw ASR hypothesis (including noise and pronunciation variants) to <file>.asr or <timestamp>.asr if input is mic or stdin.\n\
    --save-ctm or -C: save raw ASR hypothesis (including noise and pronunciation variants) to <file>.ctm or <timestamp>.ctm if input is mic or stdin. Useful for evaluating ASR accuracy when reference transcriptions are available.\n\
  options specific to audio input:\n\
    --hmm or -H DIR: source language accoustic model (for recognition)\n\
    --asr-segment or -S INT: Duration of speech chunks to recognise (in seconds). If too low, ASR accuracy will be poor. if too high, it will not work at all. I suggest you stick to the default (10s)\n\
    --asr-buffer or -B INT: number of ASR segment to bufferize (default=10)\n\
    --asr-segment-overlap or -O INT: how much asr segments overlap. bigger = slower but better accuracy. Must be bigger than one frame and smaller than one segment. (in seconds, default=3).\n\
    --one-best or -1: Only pass 1-best ASR hypothesis to the translation system, instead of lattice. Much faster, less accurate.\n\
  options specific to lattice input:\n\
    --lattice-input or -a: input is SLF lattice (default: false)\n\
  options specific to the translation system:\n\
    --moses-config or -m PATH: moses configuration file (defaults to moses.ini)\n\
    --target-lm or -L FILE: target language model (for translation)\n\
    --mt-segmentation or -t (emit|prefer|best|best-prefer|silence|length|mix|none): choose ASR hypothesis resegmetnation method. (default=mix)\n\
    --max-mt-chunk or -c INT: maximum length of segments passed to MT system (default=40)\n\
    --save-raw-mt-seg or -R: save raw segments passed to MT system (including noise and pronunciation variants) to <file>.raw_mt_seg or <timestamp>.raw_mt_seg if input is mic or stdin.\n\
    --save-clean-mt-seg or -T: save clean segments passed to MT system (retokenised, no acoustic indications) to <file>.clean_mt_seg or <timestamp>.clean_mt_seg if input is mic or stdin.\n";
    
    
    static struct option long_options[] = {
        {"verbose",             no_argument,       NULL, 'v'}, /* verbosity */
        {"compress",            no_argument,       NULL, 'z'},
        {"stdout",              no_argument,       NULL, 'o'},
        {"lattice-input",       no_argument,       NULL, 'a'},
        {"prefix",              required_argument, NULL, 'P'},
        {"moses-config",        required_argument, NULL, 'm'},
        {"pocketsphinx-config", required_argument, NULL, 'p'},
        {"s2tt-config",         required_argument, NULL, 's'},
        {"source-lm",           required_argument, NULL, 'l'},
        {"source-lm-weight",    required_argument, NULL, 'w'},
        {"target-lm",           required_argument, NULL, 'L'},
        {"source-backward-lm",  required_argument, NULL, 'i'},
        {"target-backward-lm",  required_argument, NULL, 'I'},
        {"source-sentlevel-nn", required_argument, NULL, 'x'},
        {"target-sentlevel-nn", required_argument, NULL, 'X'},
        {"source-wordlevel-nn", required_argument, NULL, 'y'},
        {"target-wordlevel-nn", required_argument, NULL, 'Y'},
        {"acoustic-model",      required_argument, NULL, 'M'},
        {"one-best",            no_argument,       NULL, '1'},
        {"asr-segment",         required_argument, NULL, 'S'},
        {"asr-segment-overlap", required_argument, NULL, 'O'},
        {"asr-buffer",          required_argument, NULL, 'B'},
        {"max-mt-chunk",        required_argument, NULL, 'c'},
        {"save-asr",            no_argument,       NULL, 'A'},
        {"save-raw-mt-seg",     no_argument,       NULL, 'R'},
        {"save-clean-mt-seg",   no_argument,       NULL, 'C'},
        {"help",                no_argument,       NULL, 'h'},
        {0, 0, 0, 0}
    };
    
    int help = 0;
    
    int option_code = 1;
    
    while(option_code!=-1) {
        
        int option_index = 0;
        option_code = getopt_long(argc, argv, "s:p:m:l:b:L:i:I:x:X:y:Y:M:S:O:B:t:c:P:w:ACRTozvh",long_options, &option_index);
        switch(option_code) {
            case 0:  /* long option setting a flag */
                break;
            case 'v':
                config->verbose = 1;
                break;
            case 'z':
                config->compress_out = 1;
                break;
            case 'a':
                config->lattice_input = 1;
                break;
            case '1':
                config->one_best = 1;
                break;
            case 'h':
                help = 1;
                break;
            case 'P':
                config->prefix = optarg;
                break;
            case 'm':
                config->cfg_moses_filename = optarg;
                break;
            case 'p':
                config->cfg_ps_filename = optarg;
                break;
            case 's':
                config->cfg_s2tt_filename = optarg;
                break;
            case 'l':
                add_lm(config,optarg);
                break;
            case 'w':
                config->ps_config.lw = (float32)atof(optarg);
                config->ps_config.lw_str = optarg;
                break;
            case 'b':
                config->ps_config.lmbase = (float64)atof(optarg);
                break;
            case 'L':
                config->moses_config.lm_filename = optarg;
                break;
            case 'i':
                config->cm_config.asr_backward_lm_filename = optarg;
                break;
            case 'I':
                config->cm_config.mt_backward_lm_filename = optarg;
                break;
            case 'x':
                config->cm_config.asr_slf_fann_filename = optarg;
                break;
            case 'X':
                config->cm_config.mt_slf_fann_filename = optarg;
                break;
            case 'y':
                config->cm_config.asr_wlf_fann_filename = optarg;
                break;
            case 'Y':
                config->cm_config.mt_wlf_fann_filename = optarg;
                break;
            case 'M':
	        add_decoder(config, optarg);
                break;
            case 'S':
                config->asr_segment_length = atoi(optarg);
                break;
            case 'O':
                config->asr_segment_overlap = atoi(optarg);
                break;
            case 'B':
                config->asr_buffer_size = atoi(optarg);
                break;
            case 't':
                if      (!strcmp(optarg,"emit")) config->seg_method = SEG_EMIT;
                else if (!strcmp(optarg,"prefer")) config->seg_method = SEG_PREFER;
                else if (!strcmp(optarg,"best")) config->seg_method = SEG_BEST;
                else if (!strcmp(optarg,"best-prefer")) config->seg_method = SEG_BEST_PREFER;
                else if (!strcmp(optarg,"silence")) config->seg_method = SEG_SILENCE;
                else if (!strcmp(optarg,"length")) config->seg_method = SEG_LENGTH;
                else if (!strcmp(optarg,"mix")) config->seg_method = SEG_MIX;
                else if (!strcmp(optarg,"none")) config->seg_method = SEG_NONE;
                else myfail(EXIT_CONFIG_ERROR,"invalid segmentation method '%s'",optarg);
                break;
            case 'c':
                config->max_mt_chunk = atoi(optarg);
                break;
            case 'A':
                config->save_asr = 1;
                break;
            case 'R':
                config->save_raw_mt_seg = 1;
                break;
            case 'T':
                config->save_clean_mt_seg = 1;
                break;
            case 'C':
                config->save_ctm = 1;
                break;
            case 'o':
                config->out_to_stdout = 1;
                break;
            case -1:  /* end of options */
                break;
            case '?':  /* unknown option */
                help = 1;
                break;
            default:  /* serious error */
                myerror("you are not reading this.");
                abort();
        }
        
    }
    
    if(help) {
        printf(usage,argv[0]);
        exit(EXIT_SUCCESS);
    }
    
    /* input files names */
    int i = optind;
    config->inputfilenames = (optind<argc)?malloc((argc-optind)*sizeof(char*)):NULL;
    while(i<argc) {
        config->inputfilenames[i-optind] = argv[i];
        i += 1;
    }
    config->n_input = (optind<argc)?(argc-optind):1; /* if no file provided, input is mic/stdin */
    
    
    return;
    
}

void add_decoder(mp_config_t * config, char * amspecs) {
  gchar ** specs = g_strsplit(amspecs, ":", 0);

  config->ps_config.decoders.count++;
  config->ps_config.decoders.cfg = realloc(config->ps_config.decoders.cfg,
					   config->ps_config.decoders.count * 
					   sizeof(config->ps_config.decoders.cfg[0]));

  int last = config->ps_config.decoders.count - 1;
  decoder_config_t *deco_cfg = &config->ps_config.decoders.cfg[last];

  deco_cfg->hmm_filename = malloc((strlen(specs[0])+1) * sizeof(char));
  strcpy(deco_cfg->hmm_filename, specs[0]);

  if (specs[1] == NULL)
    goto free_and_return;

  if (*specs[1] != '\0') {
    deco_cfg->dict_filename = malloc((strlen(specs[1])+1) * sizeof(char));
    strcpy(deco_cfg->dict_filename, specs[1]);
  }

  if (specs[2] == NULL)
    goto free_and_return;

  if (*specs[2] != '\0') {
    deco_cfg->fdict_filename = malloc((strlen(specs[2])+1) * sizeof(char));
    strcpy(deco_cfg->fdict_filename, specs[2]);
  }

  if (specs[3] == NULL)
    goto free_and_return;

  if (*specs[3] != '\0') {
    deco_cfg->mdef_filename = malloc((strlen(specs[3])+1) * sizeof(char));
    strcpy(deco_cfg->mdef_filename, specs[3]);
  }

 free_and_return:
  g_strfreev(specs);
  return;
}

void add_lm(mp_config_t * config, char * lmspecs) {
    
    char * file = NULL;
    char * name = NULL;
    int weight_found = 0;
    float32 weight = 1;

    gchar ** specs = g_strsplit(lmspecs,":",0);

    if(specs[0]==NULL) {
        myfail(EXIT_CONFIG_ERROR,"invalid lm specification");
    }
    file = malloc((strlen(specs[0])+1)*sizeof(char));
    strcpy(file, specs[0]);
    
    int i = 1;
    while(specs[i]!=NULL) {
        gchar ** keyval = g_strsplit(specs[i],"=",2);
        if(keyval[0]==NULL) {
            myfail(EXIT_CONFIG_ERROR,"invalid lm specification");
        }
	/* FIXME Looks unusual, if done purposely, should be explained */
        if(keyval[1]==NULL || !strcmp(keyval[0],"weight") || !strcmp(keyval[0],"w")) {
	  if(weight_found)
	    myfail(EXIT_CONFIG_ERROR,"lm weight redefined.");
	  char * s = (keyval[1]!=NULL)?keyval[1]:keyval[0];
	  weight = (float32)atof(s);
	  weight_found = 1;
        }
        if(!strcmp(keyval[0],"name")) {
	  if(name!=NULL)
	    myfail(EXIT_CONFIG_ERROR,"lm name redefined.");
	  name = malloc((strlen(keyval[1])+1)*sizeof(char));
	  strcpy(name, keyval[1]);
        }
	g_strfreev(keyval);
        i += 1;
    }

    g_strfreev(specs);
    
    if(name==NULL) {
      name = malloc((strlen(file)+1)*sizeof(char));
      strcpy(name, file);
    }
    
    config->ps_config.n_lm += 1;
    config->ps_config.lmset = realloc(config->ps_config.lmset,(config->ps_config.n_lm)*sizeof(char*));
    config->ps_config.lmnames = realloc(config->ps_config.lmnames,(config->ps_config.n_lm)*sizeof(char*));
    config->ps_config.lmweights = realloc(config->ps_config.lmweights,(config->ps_config.n_lm)*sizeof(float32));
    
    config->ps_config.lmset[config->ps_config.n_lm-1] = file;
    config->ps_config.lmnames[config->ps_config.n_lm-1] = name;
    config->ps_config.lmweights[config->ps_config.n_lm-1] = weight;

}

int mp_check_cl(mp_config_t * cfg) {
    
    int status = 0;
    
    if(cfg->asr_segment_length<1) {
        myerror("Invalid ASR segment length '%i'",cfg->asr_segment_length);
        status = EXIT_CONFIG_ERROR;
    }
    if(cfg->asr_segment_overlap<0 || cfg->asr_segment_overlap>=cfg->asr_segment_length) {
        myerror("Invalid ASR segment overlap '%i'",cfg->asr_segment_overlap);
        status = EXIT_CONFIG_ERROR;
    }
    if(cfg->asr_buffer_size<1) {
        myerror("Invalid ASR buffer size '%i'",cfg->asr_buffer_size);
        status = EXIT_CONFIG_ERROR;
    }
    if(cfg->max_mt_chunk<=0) {
        myerror("Invalid maximum mt chunk size '%i'",cfg->max_mt_chunk);
        status = EXIT_CONFIG_ERROR;
    }
    
    return status;
}
