/*
Copyright (c) 2011, Sylvain Raybaud <sylvain.raybaud@gmail.com>
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
Neither the name of the PAROLE team, LORIA, INRIA nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <pthread.h>
#include <glib.h>

#include <ngram_model.h>
#include <logmath.h>
#include <pocketsphinx.h>

#include "engine.h"
#include "maparole.h"
#include "common.h"

asr_stream_t * asr_stream_new_item(char * word, asr_stream_t * prev, asr_stream_t * next, float32 start, float32 end, int non_word) {
    debug("starting");
    asr_stream_t * new_item = malloc(sizeof(asr_stream_t));
    new_item->word = word;
    new_item->next = next;
    new_item->prev = prev;
    new_item->t_start = start;
    new_item->t_end = end;
    new_item->cm = 0;
    new_item->non_word = non_word;
    return new_item;
}


/*                         ***     CAUTION     ***                               */
/* asr low level manipulation functions assume the mutex on the stream is locked! */
void asr_stream_free(asr_stream_t * stream) {
    debug("starting");
    while(stream!=NULL) {
        asr_stream_t * current = stream;
        stream = current->next;
        free(current->word);
        free(current);
    }
}

asr_stream_t * asr_stream_free_n(asr_stream_t * stream, int n) {
    debug("starting");
    int i = 1;
    while(stream!=NULL && (n<0 || i<=n)) {
        asr_stream_t * current = stream;
        stream = current->next;
        free(current->word);
        free(current);
        i += 1;
    }
    return stream;
}


void asr_stream_free_backward(asr_stream_t * stream) {
    debug("starting");
    while(stream!=NULL) {
        asr_stream_t * current = stream;
        stream = current->prev;
        free(current->word);
        free(current);
    }
}

/* get last item of a stream */
/* NULL if stream is NULL (empty) */
asr_stream_t * asr_stream_get_end(asr_stream_t * stream) {
    debug("starting");
    asr_stream_t * item = stream;
    while(item!=NULL && item->next!=NULL) {
        item = item->next;
    }
    return item;
}
/* get first item of a stream */
/* NULL if stream is NULL (empty) */
asr_stream_t * asr_stream_rewind(asr_stream_t * stream) {
    debug("starting");
    asr_stream_t * item = stream;
    while(item!=NULL && item->prev!=NULL) {
        item = item->prev;
    }
    return item;
}


char * stream_to_string(asr_stream_t * from, int length) {
    debug("starting");
    if(from==NULL) {
        return NULL;
    }
    
    int nb_words = 1;
    size_t len = 1;
    char * str = malloc(len*sizeof(char));
    str[len-1] = '\0';
    asr_stream_t * stream = from;
    while((length<0 || nb_words<=length) && stream!=NULL) {
        if(stream->word!=NULL && !stream->non_word && strcmp(stream->word,"euh")) {  /* do this properly! */
            len += strlen(stream->word)+1;
            str = realloc(str,len*sizeof(char));
            strcat(str,stream->word);
            strcat(str," ");
        }
        stream = stream->next;
        nb_words += 1;
    }
    return str;
}

char * stream_to_string_raw(asr_stream_t * from, int length) {
    debug("starting");
    if(from==NULL) {
        return NULL;
    }
    
    int nb_words = 1;
    size_t len = 1;
    char * str = malloc(len*sizeof(char));
    str[len-1] = '\0';
    asr_stream_t * stream = from;
    while((length<0 || nb_words<=length) && stream!=NULL) {
        if(stream->word!=NULL) {
            len += strlen(stream->word)+1;
            str = realloc(str,len*sizeof(char));
            strcat(str,stream->word);
            strcat(str," ");
        }
        stream = stream->next;
        nb_words += 1;
    }
    return str;
}

void write_ctm(manager_data_t * pool, int length) {
    debug("start");
    if(pool->asr_stream==NULL || length==0) {
        return;
    }
    
    int nb_words = 1;
    asr_stream_t * stream = pool->asr_stream;
    while((length<0 || nb_words<=length) && stream!=NULL) {
        if(stream->word!=NULL) {
            myfprintf(pool->save_ctm_file,"%s %i %f %f %s %f\n",pool->in.filename,pool->n_channel,stream->t_start,stream->t_end-stream->t_start,stream->word,stream->cm);
        }
        stream = stream->next;
        nb_words += 1;
    }
}


/* preprocess string for mt: */
/* tokenize and stuff. */
char * clean_for_mt(char * word) {
    debug("start");
    /* tokenize */
    /* replace "'" with " ' " */
    /* and "_" with " " */
    
    /* compute length of tokenized string */
    size_t len = strlen(word);
    size_t new_len = 1; /* init with space for \0 */
    size_t i;
    for(i=0;i<len;i++) {
        if(word[i]=='\'' && i>0) {
            new_len += (i<len && word[i+1]!=' ')?3:2;
        }
        else {
            new_len += 1;
        }
    }
    /* allocate new string */
    char * tok = malloc(new_len*sizeof(char));
    /* tokenize  string */
    int tok_ptr = 0;
    for(i=0;i<len;i++) {
        if(word[i]=='\'' && i>0) {
            tok[tok_ptr++] = ' ';
            tok[tok_ptr++] = '\'';
            if(i<len && word[i+1]!=' ')
                tok[tok_ptr++] = ' ';
        }
        else if(word[i]=='_') {
            tok[tok_ptr++] = ' ';
        }
        else {
            tok[tok_ptr++] = word[i];
        }
    }
    tok[new_len-1] = '\0';
    
    return tok;
}

char * remove_variant(const char * word) {
    debug("starting");
    size_t len = strlen(word);
    char * w = malloc((len+1)*sizeof(char));
    size_t i;
    for(i=0;i<len;i++) {
        if(word[i]=='(') {
            w[i] = '\0';
            break;
        }
        else {
            w[i] = word[i];
        }
    }
    w[len] = '\0';
    return w;
}

int is_non_word(const char * word) {
    debug("starting");
    if(word==NULL)
        return 1;
    size_t len = strlen(word);
    if((len>=4 && word[0]=='+' && word[1]=='+' && word[len-2]=='+' && word[len-1]=='+')
        || (!strcmp(word,"<sil>") || !strcmp(word,"SIL"))) {
        return 1;
    }
    else {
        return 0;
    }
}

asr_stream_t * asr_stream_get_first_word(asr_stream_t * stream) {
    debug("starting");
    asr_stream_t * current = stream;
    while(current!=NULL && (current->word==NULL || current->non_word))
        current = current->next;
    return current;
}

int is_silence(asr_stream_t * item) {
    if(item->word!=NULL
        && (!strcmp(item->word,"++silence++") || !strcmp(item->word,"<sil>") || !strcmp(item->word,"SIL")))
        return 1;
    else
        return 0;
}

int asr_stream_length(asr_stream_t * stream) {
    debug("starting");
    int len = 0;
    asr_stream_t * current = stream;
    while(current!=NULL) {
        current = current->next;
        len += 1;
    }
    return len;
}

char * gen_filename(const char * base, const char * prefix, char * suffix, char * timestamp)  {
    debug("starting");
    
    if(!base && !prefix && !timestamp)
        myfail(EXIT_UNKNOWN_ERROR,"cannot conjure a filename out of thin air.");
        
    if(base==NULL) {
        size_t len = 1;
        if(suffix!=NULL) len += strlen(suffix);
        if(prefix!=NULL) len += strlen(prefix);
        if(timestamp!=NULL) len += strlen(timestamp);
        char * outfname = malloc(len*sizeof(char));
        sprintf(outfname,"%s%s%s",prefix!=NULL?prefix:"",timestamp!=NULL?timestamp:"",suffix!=NULL?suffix:"");
        return outfname;
    }
    
    if(prefix!=NULL) {
        gchar ** base_split = g_strsplit_set(base,"/\\",0);
        int i=-1; while(base_split[i+1]!=NULL) i++;
        size_t len = 1;
        if(i>=0) len += strlen(base_split[i]);
        if(suffix!=NULL) len += strlen(suffix);
        if(prefix!=NULL) len += strlen(prefix);
        if(timestamp!=NULL) len += strlen(timestamp);
        char * outfname = malloc(len*sizeof(char));
        sprintf(outfname,"%s%s%s%s",prefix,(i>=0)?base_split[i]:"",timestamp!=NULL?timestamp:"",suffix!=NULL?suffix:"");
        g_strfreev(base_split);
        return outfname;
    }
    
    size_t len = 1+strlen(base);
    if(suffix!=NULL) len += strlen(suffix);
    if(timestamp!=NULL) len += strlen(timestamp);
    char * outfname = malloc(len*sizeof(char));
    sprintf(outfname,"%s%s%s",base,timestamp!=NULL?timestamp:"",suffix!=NULL?suffix:"");
    return outfname;
    
}

void init_ps_lm(mp_config_t * config) {
    
    if(config->ps_config.n_lm==0)
        myfail(EXIT_CONFIG_ERROR,"no LM specified");
    
    logmath_t * lm = logmath_init(config->ps_config.lmbase,0,1);
    
    config->ps_config.lms = malloc(config->ps_config.n_lm*sizeof(ngram_model_t*));
    int i;
    for(i=0;i<config->ps_config.n_lm;i++) {
        config->ps_config.lms[i] = ngram_model_read(config->ps_config.ps_cmd_ln,config->ps_config.lmset[i],NGRAM_AUTO,logmath_retain(lm));
    }
    config->ps_config.lm = ngram_model_set_init(config->ps_config.ps_cmd_ln,config->ps_config.lms,config->ps_config.lmnames,config->ps_config.lmweights,config->ps_config.n_lm);
    
}

audio_chunk_t * audio_chunk_new(int16 * audio, size_t size) {
    audio_chunk_t * chunk = malloc(size*sizeof(int16));
    chunk->audio = audio;
    chunk->size = size;
    return chunk;
}

void audio_chunk_free(audio_chunk_t * chunk) {
    free(chunk->audio);
    free(chunk);
}



/* get best hypothesis as a linked list of items */
asr_stream_t * get_asr_stream_from_utterance(manager_data_t * pool, ps_seg_t * seg) {
    debug("start");
    
    ps_seg_t * segment = seg;
    asr_stream_t * stream = NULL;
    
    while(segment!=NULL) {
        const char * word = ps_seg_word(segment);
        if(strcmp(word,"<s>") && strcmp(word,"</s>")) {
            asr_stream_t * prev = stream;
            stream = asr_stream_new_item(remove_variant(word),prev,NULL,-1,-1,is_non_word(word));
            if(prev!=NULL) {
                prev->next = stream;
            }
            int start_frm = -1, end_frm = -1;
            ps_seg_frames(segment,&start_frm,&end_frm);
            stream->t_start = start_frm*pool->samples_frame_shift/pool->sampling_rate+pool->segment_time_offset;
            stream->t_end = end_frm*pool->samples_frame_shift/pool->sampling_rate+pool->segment_time_offset;
        }
        segment = ps_seg_next(segment);
    }
    
    return asr_stream_rewind(stream);
}

/* merge best hypothesis into the stream of recognised items */
void merge_utt_into_stream(manager_data_t * pool, asr_stream_t * utt) {
    debug("start");
    
    if(utt==NULL) {
        return;
    }
    
    /* naive method (ester) */
    /* get longest common word */
    /* merge from that word */
    /* if none found, consider that last but one word in pool is best */
    
    asr_stream_t * left_utt = asr_stream_get_end(pool->asr_stream);
    if(left_utt==NULL) {
        /* no left utterance, nothing to merge into */
        pool->asr_stream = utt;
        return;
    }
    
    float32 right_t_start = utt->t_start;
    
    float32 greatest_length = 0;
    asr_stream_t * cut_left = NULL; /* cut left utterance after (not including)... */
    asr_stream_t * cut_right = NULL; /* cut right utterance at (including)... */
    while(left_utt!=NULL && left_utt->t_end>=right_t_start) {
        asr_stream_t * right_utt = utt;
        while(right_utt!=NULL && right_utt->t_start<left_utt->t_end) {
            if(((right_utt->t_end>=left_utt->t_start && right_utt->t_start<=left_utt->t_start) || (right_utt->t_end>=left_utt->t_end && right_utt->t_start<=left_utt->t_end))
                && (left_utt->t_end-left_utt->t_start+right_utt->t_end-right_utt->t_start)>greatest_length
                && left_utt->word!=NULL && right_utt->word!=NULL
                && !strcmp(right_utt->word,left_utt->word)) {
                cut_left = left_utt;
                cut_right = right_utt;
                greatest_length = left_utt->t_end-left_utt->t_start+right_utt->t_end-right_utt->t_start;
            }
            right_utt = right_utt->next;
        }
        left_utt = left_utt->prev;
    }
    
    /* if no common word found */
    /* remove last word of left utterance */
    /* then, first words of right utterance */
    /* until time consistent merge is possible */
    if(cut_left==NULL) {
        cut_left = asr_stream_get_end(pool->asr_stream)->prev;
        if(cut_left==NULL) cut_left = asr_stream_get_end(pool->asr_stream);
        float32 left_t_start = cut_left->t_start;
        asr_stream_t * right_utt = utt;
        while(right_utt->next!=NULL && right_utt->next->t_end<=left_t_start) {
            right_utt = right_utt->next;
        }
        cut_right = (right_utt->next!=NULL)?right_utt->next:right_utt;
    }
    
    /* if timings are inconsistent (may happen for words overlapping with chunk borders) */
    /* just fiddle with them (it doesn't really matter, only for asr scoring) */
    if(cut_left!=NULL && cut_right!=NULL && cut_right->next!=NULL && cut_left->t_end>cut_right->next->t_start) {
        cut_left->t_end = (cut_left->t_end+cut_right->next->t_start)/2;
        cut_left->t_start = min(cut_left->t_start,cut_left->t_end);
        cut_right->next->t_start = cut_left->t_end;
        cut_right->next->t_end = max(cut_right->next->t_end,cut_right->next->t_start);
    }
    
    asr_stream_free(cut_left->next);
    cut_left->next = cut_right->next;
    if(cut_left->next!=NULL) {
        cut_left->next->prev = cut_left;
    }
    asr_stream_free_backward(cut_right);
}
