#ifndef CM_H
#define CM_H

/* 
   Module where features for confidence measure are implemented.
   It provides a feature_set type to store a list of features to compute

   Features are regrouped in 4 categories (word/sentence level) X (asr/mt).
   For each category, we use an array of all available features,
   and the results for a given translation pair are stored in an array.
   (NB: not in this file, but this piece of info is important anyway)

   A feature in an array is given by
   - a pointer to the computation function
   - its name in a plain char*
   - a boolean which tells whether it is enabled
   - the index of the feature in the results array
   (i.e. it is a numbering of the enabled features;
   therefore the result of feature i of tab is stored at index tab[i].idx of the
   corresponding results array)

*/

#include "language_model.h"

struct fann;
struct cm_config_s;


/* This structure holds references to all the common resources
   for confidence measure */
typedef struct cm_resources_s {
  /* language model instances */
  lmi_t asr_lm;
  lmi_t asr_backward_lm;
  lmi_t mt_lm;
  lmi_t mt_backward_lm;

  /* neural network instances */
  struct fann *asr_word_level_ann;
  struct fann *asr_sent_level_ann;
  struct fann *mt_word_level_ann;
  struct fann *mt_sent_level_ann;
} cm_resources_t;

cm_resources_t *cm_resources_new(struct cm_config_s *config, sphinx_lm_t lm);

void cm_resources_free(cm_resources_t *config);


struct translation_pair_s;
struct feature_set_s;

/* Types of word (wlff) and sentence (slff) level feature function for confidence measure */
typedef float (*wlff)(cm_resources_t *res, struct feature_set_s *set, struct translation_pair_s *info, int i);
typedef float (*slff)(cm_resources_t *res, struct feature_set_s *set, struct translation_pair_s *info);

/* Represents a word level feature */
typedef struct {
  wlff fun;		 /* pointer to the computation function */
  char * name;		 /* name of the feature, for dynamic access */
  int enabled;		 /* true iff this feature must be computed */
  int idx;		 /* index of the feature in the list of enabled features */
} word_level_feature_t;
/* Represents a sentence level feature */
typedef struct {
  slff fun;
  char * name;
  int enabled;
  int idx;
} sent_level_feature_t;


typedef struct feature_set_s {
  word_level_feature_t * mt_wlf;
  int mt_wlf_count;
  int mt_wlf_total;
  sent_level_feature_t * mt_slf;
  int mt_slf_count;
  int mt_slf_total;
  word_level_feature_t * asr_wlf;
  int asr_wlf_count;
  int asr_wlf_total;
  sent_level_feature_t * asr_slf;
  int asr_slf_count;
  int asr_slf_total;
} feature_set_t;

/* New feature functions set, all enabled */
feature_set_t * feature_set_new();
/* New feature functions set, enabling only thoses named in 'names' (NULL terminated array) */
feature_set_t * feature_set_new_by_name(char **names);

void feature_set_free(feature_set_t *set);

void compute_features(feature_set_t *set, cm_resources_t *res, struct translation_pair_s *info);

#endif
