/*
Copyright (c) 2011, Sylvain Raybaud <sylvain.raybaud@gmail.com>
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
Neither the name of the PAROLE team, LORIA, INRIA nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdio.h>
#include <glib.h>
#include <pthread.h>
#include <sys/time.h>
#include <time.h>

#include <ngram_model.h>
#include <pocketsphinx.h>
#include <libxml/tree.h>

#include <cutils.h>

#include "engine.h"
#include "common.h"
#include "maparole.h"
#include "interface_ps.h"
#include "interface_moses.h"
#include "confidence_measure.h"
#include "language_model.h"

mp_engine_t mp_start_engine(mp_config_t * cfg) {
    debug("starting");

    mp_engine_t engine = {0};

    /* start engine */

    engine.ps = ps_init(cfg->ps_config.ps_cmd_ln);
    if (engine.ps == NULL) {
        myerror("Could not intialize PocketSphinx decoder");
        exit(EXIT_PS_ERROR);
    }
    
    if(cfg->ps_config.n_lm>0 && !cfg->htk_input) {
        init_ps_lm(cfg);
        cfg->ps_config.lm = ps_update_lmset(engine.ps,cfg->ps_config.lm);
        if(cfg->ps_config.lm==NULL) {
            myfail(EXIT_PS_ERROR,"Could not initialize LM set");
        }
    }

    if (!moses_static_data(&(cfg->moses_config))) {
        myerror("Could not initialize Moses decoder");
        exit(EXIT_MOSES_ERROR);
    }

    time(&(engine.time_started));
    struct tm * date = localtime(&(engine.time_started));
    engine.timestamp = malloc(20*sizeof(char));
    sprintf(engine.timestamp,"%i.%.2i.%.2i-%.2i.%.2i.%.2i",date->tm_year,date->tm_mon,date->tm_mday,date->tm_hour,date->tm_min,date->tm_sec);

    return engine;
}

void mp_terminate(mp_config_t * config, mp_engine_t * engine) {
    debug("starting");

    free(config->inputfilenames);
    free(config->ps_config.lms);
    free(engine->timestamp);

    if(!config->lattice_input)
      ps_terminate(&(config->ps_config),engine);
    moses_terminate(&config->moses_config);
    xmlCleanupParser();		/* free memory allocated by libxml2 */
}

void ps_terminate(ps_config_t * cfg,mp_engine_t * engine) {
    int i;
    debug("starting");
    /* debug("ps LMs"); */
    /* int i; */
    /* ngram_model_free(cfg->lm); */
    /* for(i=0;i<cfg->n_lm;i++) { */
    /*     ngram_model_free(cfg->lms[i]); */
    /* } */
    for (i=0 ; i<cfg->n_lm ; i++) {
      free(cfg->lmset[i]);
      free(cfg->lmnames[i]);
    }
    free(cfg->lmset);
    free(cfg->lmnames);
    free(cfg->lmweights);
    logmath_free(cfg->lmath);
    ps_free(engine->ps);
}

void mp_start_translation(mp_config_t * cfg, mp_engine_t * engine) {
    debug("starting");
    myverbose(cfg->verbose,"starting translation manager");
    
    int n = 0;
    while(n<cfg->n_input) {
        
        manager_data_t pool = manager_data_init(cfg,engine,cfg->inputfilenames==NULL?NULL:cfg->inputfilenames[n]);

        pthread_t asr_buffer_manager_thread;
        pthread_t asr_manager_thread;
        pthread_t asr_stream_manager_thread;
//         pthread_t lattice_loader_thread;
        pthread_t segment_manager_thread;
        pthread_t translation_manager_thread;
        
        int asr_buffer_manager_created  = pthread_create(&asr_buffer_manager_thread, NULL,  asr_buffer_manager, (void*)&pool);
        int asr_manager_created         = pthread_create(&asr_manager_thread,        NULL,  asr_manager,        (void*)&pool);
        int asr_stream_manager_created  = pthread_create(&asr_stream_manager_thread, NULL,  asr_stream_manager, (void*)&pool);
//         int lattice_loader_created      = pthread_create(&lattice_loader_thread,     NULL,  lattice_loader,     (void*)&pool);
        int segment_manager_created     = pthread_create(&segment_manager_thread,    NULL,  segment_manager,    (void*)&pool);
        int translation_manager_created = pthread_create(&translation_manager_thread,NULL,  translation_manager,(void*)&pool);
        
        if(asr_buffer_manager_created || asr_manager_created || asr_stream_manager_created || segment_manager_created || translation_manager_created) {
            manager_data_free(&pool);
            myfail(EXIT_UNKNOWN_ERROR,"Could not start threads.");
        }
        
        
        /* end */
        myverbose(cfg->verbose,"[engine] all threads launched, waiting for threads to finish.");
        
        pthread_join(asr_buffer_manager_thread,NULL);
        pthread_join(asr_manager_thread,NULL);
        pthread_join(asr_stream_manager_thread,NULL);
//         pthread_join(lattice_loader_thread,NULL);
        pthread_join(segment_manager_thread,NULL);
        pthread_join(translation_manager_thread,NULL);
        
        myverbose(cfg->verbose,"freeing pool");
        manager_data_free(&pool);
        
        n += 1;
    }
    
}

manager_data_t manager_data_init(mp_config_t * cfg, mp_engine_t * engine, const char * input) {
    debug("starting");
    myverbose(cfg->verbose,"populating pool");

    manager_data_t pool = {0};

    pool.mp_config = cfg;
    pool.mp_engine = engine;

    pool.lm = cfg->ps_config.lm;
    pool.sentence_start_id = pool.lm!=NULL?ngram_wid(pool.lm,"<s>"):0;
    pool.lm_zero = pool.lm!=NULL?ngram_zero(pool.lm):-99;
    pthread_mutex_init(&pool.lm_mutex,NULL);
    
    /* data for audio */
    
    if(!cfg->lattice_input) {
        
        fe_t * fe = ps_get_fe(engine->ps);
        fe_get_input_size(fe,&pool.samples_frame_shift,&pool.samples_per_frame);
        pool.sampling_rate = cmd_ln_float_r(cfg->ps_config.ps_cmd_ln,"-samprate");
        pool.n_channel = 1; /* no choice yet */

        pool.input_buffer_size = cfg->asr_buffer_size;

        /* convert durations to sizes (number of samples (int16)). make sure the frame count is an integer */
        pool.asr_segment_size = pool.sampling_rate*cfg->asr_segment_length;
        pool.asr_segment_size += pool.samples_frame_shift-((pool.asr_segment_size-pool.samples_per_frame)%pool.samples_frame_shift); /* round number of frames */
        pool.asr_segment_overlap = pool.sampling_rate*cfg->asr_segment_overlap;
        if(pool.asr_segment_overlap>(pool.asr_segment_size/2)) pool.asr_segment_overlap = pool.asr_segment_size/2 ;
        pool.segment_time_shift = (float32)(pool.asr_segment_size-pool.asr_segment_overlap)/pool.sampling_rate;
        pool.segment_time_offset = 0;
        
        pool.input_buffer = g_queue_new();
        pthread_mutex_init(&pool.input_buffer_mutex,NULL);
        pthread_cond_init(&pool.input_buffer_one_pushed,NULL);
        pthread_cond_init(&pool.asr_hyp_queue_one_hyp_pushed,NULL);

        pool.asr_hyp_queue = g_queue_new();
        pthread_mutex_init(&pool.asr_hyp_queue_mutex,NULL);
        pthread_cond_init(&pool.input_buffer_one_poped,NULL);

        pool.asr_stream = NULL;
        pthread_mutex_init(&pool.asr_stream_mutex,NULL);
        pthread_cond_init(&pool.asr_stream_updated,NULL);
        
        myverbose(cfg->verbose,"pool info:");
        myverbose(cfg->verbose,"  sampling rate = %fHz",pool.sampling_rate);
        myverbose(cfg->verbose,"  frame shift = %i samples",pool.samples_frame_shift);
        myverbose(cfg->verbose,"  asr buffer = %i segments",cfg->asr_buffer_size);
        myverbose(cfg->verbose,"  asr segment = %is (%li samples)",cfg->asr_segment_length,pool.asr_segment_size);
        myverbose(cfg->verbose,"  asr segment overlap = %is (%li samples)",cfg->asr_segment_overlap,pool.asr_segment_overlap);
        myverbose(cfg->verbose,"  frame size = %i samples",pool.samples_per_frame);
        myverbose(cfg->verbose,"  inter segments time shift = %fs",pool.segment_time_shift);
        
    }
    
    
    /* data for mt */

    pool.segment_queue = g_queue_new();
    pthread_mutex_init(&pool.segment_queue_mutex,NULL);
    pthread_cond_init(&pool.segment_queue_one_segment_pushed,NULL);

    /* cut on '.', ';', ':', '?', '!', '</s>' */
    /* should also cut on ',' */
    if(pool.lm!=NULL) {
        pool.unk_id = ngram_unknown_wid(pool.lm);
        int id_ptr = 0;
        int32 id;
        pool.mt_cut_symbols_id = malloc(8*sizeof(int32));
        if((id = ngram_wid(pool.lm,","))!=pool.unk_id)
            pool.mt_cut_symbols_id[id_ptr++] = id;
        if((id = ngram_wid(pool.lm,";"))!=pool.unk_id)
            pool.mt_cut_symbols_id[id_ptr++] = id;
        if((id = ngram_wid(pool.lm,":"))!=pool.unk_id)
            pool.mt_cut_symbols_id[id_ptr++] = id;
        if((id = ngram_wid(pool.lm,"?"))!=pool.unk_id)
            pool.mt_cut_symbols_id[id_ptr++] = id;
        if((id = ngram_wid(pool.lm,"!"))!=pool.unk_id)
            pool.mt_cut_symbols_id[id_ptr++] = id;
        if((id = ngram_wid(pool.lm,"."))!=pool.unk_id)
            pool.mt_cut_symbols_id[id_ptr++] = id;
        if((id = ngram_wid(pool.lm,"</s>"))!=pool.unk_id)
            pool.mt_cut_symbols_id[id_ptr++] = id;
        pool.mt_cut_symbols_id[id_ptr] = pool.unk_id;
    }
    
    
    /* verbosity */
    
    if(input==NULL) {
        myverbose(cfg->verbose,"translating speech from stdin (mic?)");
        pool.in.file = stdin;
        pool.in.filename = NULL;
    }
    else {
        myverbose(cfg->verbose,"translating speech from '%s'",input);
        if(cfg->lattice_input) {
            pool.in_filename = malloc((strlen(input)+1)*sizeof(char));
            strcpy(pool.in_filename,input);
        }
        else {
            pool.in = myfopen(input,"rb");
        }
    }
    if(cfg->out_to_stdout) {
        myverbose(cfg->verbose,"writing output to stdout");
        pool.out.file = stdout;
        pool.out.filename = NULL;
    }
    else {
        char * outfilename = gen_filename(input,cfg->prefix,(cfg->compress_out)?".s2tt.gz":".s2tt",(input==NULL)?engine->timestamp:NULL);
        myverbose(cfg->verbose,"writing output to '%s'",outfilename);
        pool.out = myfopen(outfilename,"a");
        free(outfilename);
    }

    if(cfg->save_asr) {
        pool.save_asr_filename = gen_filename(input,cfg->prefix,(cfg->compress_out)?".asr.gz":".asr",(input==NULL)?engine->timestamp:NULL);
        myverbose(cfg->verbose,"saving asr utterances to '%s'.",pool.save_asr_filename);
    }
    if(cfg->save_asr_lattice) {
        pool.save_asr_lattice_filename = gen_filename(input,cfg->prefix,".pslat",(input==NULL)?engine->timestamp:NULL);
        myverbose(cfg->verbose,"saving lattices to '%s-<id>'.",pool.save_asr_filename);
    }
    if(cfg->save_ctm) {
        pool.save_ctm_filename = gen_filename(input,cfg->prefix,(cfg->compress_out)?".ctm.gz":".ctm",(input==NULL)?engine->timestamp:NULL);
        myverbose(cfg->verbose,"saving CTM to '%s'",pool.save_ctm_filename);
    }
    if(cfg->save_raw_mt_seg) {
        pool.save_raw_mt_seg_filename = gen_filename(input,cfg->prefix,(cfg->compress_out)?".raw_mt_seg.gz":".raw_mt_seg",(input==NULL)?engine->timestamp:NULL);
        myverbose(cfg->verbose,"saving raw mt segments to '%s'",pool.save_raw_mt_seg_filename);
    }
    if(cfg->save_clean_mt_seg) {
        pool.save_clean_mt_seg_filename = gen_filename(input,cfg->prefix,(cfg->compress_out)?".clean_mt_seg.gz":".clean_mt_seg",(input==NULL)?engine->timestamp:NULL);
        myverbose(cfg->verbose,"saving clean mt segments to '%s'",pool.save_clean_mt_seg_filename);
    }
    
    if(!pool.mp_config->no_cm) {
        sphinx_lm_t sphinx_lm_cpy = {cfg->ps_config.lm, cfg->ps_config.lmath, pool.lm_mutex};
        pool.cm_res = cm_resources_new(&pool.mp_config->cm_config, sphinx_lm_cpy);
        
        pool.features = feature_set_new();
    }

    pool.xml_result_doc = xmlNewDoc(BAD_CAST "1.0");
    pool.xml_result_node = xmlNewNode(NULL, BAD_CAST "s2tt_result");
    xmlDocSetRootElement(pool.xml_result_doc, pool.xml_result_node);
    
    return pool;
}


void manager_data_free(manager_data_t * pool) {
    debug("starting");
    free(pool->mt_cut_symbols_id);
    if(!pool->mp_config->out_to_stdout) myfclose(pool->out);
    if(pool->mp_config->inputfilenames!=NULL) myfclose(pool->in);
    free(pool->save_asr_filename);
    free(pool->save_ctm_filename);
    free(pool->save_raw_mt_seg_filename);
    free(pool->save_clean_mt_seg_filename);
    if(pool->asr_stream!=NULL && pool->asr_stream->word!=NULL) {
        mywarning("there are untranslated tokens in asr_stream");
        asr_stream_free(pool->asr_stream);
    }
    
    if(!pool->mp_config->lattice_input) {
        while (! g_queue_is_empty(pool->asr_hyp_queue))
        translation_pair_free((translation_pair_t*)g_queue_pop_head(pool->asr_hyp_queue));
            g_queue_free(pool->asr_hyp_queue);
        while (! g_queue_is_empty(pool->input_buffer))
        audio_chunk_free((audio_chunk_t*)g_queue_pop_head(pool->input_buffer));
            g_queue_free(pool->input_buffer);
    }
    
    g_queue_free(pool->segment_queue);

    feature_set_free(pool->features);
    cm_resources_free(pool->cm_res);
    xmlFreeDoc(pool->xml_result_doc);
}



/* read chunks of audio and queue them */
void *asr_buffer_manager(void * data) {
    debug("starting");

    manager_data_t * pool = data;
    
    if(pool->mp_config->lattice_input)
        pthread_exit(NULL);
    
    int16 * overlap = malloc(pool->asr_segment_overlap*sizeof(int16));

    int first_round = 1;
    int input_depleted = 0;
    while(!input_depleted) {

        int16 * audio = malloc(pool->asr_segment_size*sizeof(int16));
        size_t n_samples;
        if(first_round) {
            /* if this is the first round, there is no overlaping segment to copy */
            n_samples = myfread(pool->in,audio,sizeof(int16),pool->asr_segment_size);
        }
        else {
            /* first copy overlaping segment, then read input */
            memcpy(audio,overlap,sizeof(int16)*pool->asr_segment_overlap);
            n_samples = myfread(pool->in,&(audio[pool->asr_segment_overlap]),sizeof(int16),pool->asr_segment_size-pool->asr_segment_overlap);
            n_samples += pool->asr_segment_overlap;
        }
        
        if(n_samples<pool->asr_segment_size) {
            /* less data was read than requested. it means input is depleted. this was the last round */
            myverbose(pool->mp_config->verbose,"[buffer manager] input is depleted.");
            input_depleted = 1;
        }

        pthread_mutex_lock(&pool->input_buffer_mutex);
        if(g_queue_get_length(pool->input_buffer)>=pool->input_buffer_size) {
            myverbose(pool->mp_config->verbose,"[buffer manager] waiting for some room in the buffer");
            pthread_cond_wait(&pool->input_buffer_one_poped,&pool->input_buffer_mutex);
            myverbose(pool->mp_config->verbose,"[buffer manager] received signal to go");
        }
        
        if(n_samples>pool->asr_segment_overlap || (first_round && n_samples>0)) {
            /* new data was read. create chunk */
            audio_chunk_t * chunk = audio_chunk_new(audio,n_samples);
            g_queue_push_head(pool->input_buffer,chunk);
        }
        else {
            /* if chunk is created, it will be freed by asr manager when consumed. otherwise, it needs to be freed here. */
            free(audio);
        }

        if(input_depleted) {
            pool->input_buffer_manager_is_dead = 1;
        }
        /* signal the new chunk and release mutex */
        pthread_cond_signal(&pool->input_buffer_one_pushed);
        pthread_mutex_unlock(&pool->input_buffer_mutex);

        if(!input_depleted) {
            /* if this wasn't last round, save overlaping segment for next round */
            memcpy(overlap,&(audio[pool->asr_segment_size-pool->asr_segment_overlap]),sizeof(int16)*pool->asr_segment_overlap);
        }
    
    first_round = 0;
    }
    
    myverbose(pool->mp_config->verbose,"[buffer manager] bye.");
    free(overlap);
    pthread_exit(NULL);
}


/* generate ASR utterances and queue them */
void *asr_manager(void * data) {
    debug("starting");

    manager_data_t * pool = data;
    
    if(pool->mp_config->lattice_input)
        pthread_exit(NULL);
    
    int chunk_id = 0;
    int input_depleted = 0;
    while(!input_depleted) {
    
        pthread_mutex_lock(&pool->input_buffer_mutex);
        
        if(g_queue_is_empty(pool->input_buffer) && !pool->input_buffer_manager_is_dead) {
            myverbose(pool->mp_config->verbose,"[asr manager] waiting for a new audio chunk");
            pthread_cond_wait(&pool->input_buffer_one_pushed,&pool->input_buffer_mutex);
            myverbose(pool->mp_config->verbose,"[asr manager] received signal to go");
        }

        audio_chunk_t * chunk = (audio_chunk_t*)g_queue_pop_tail(pool->input_buffer);

        pthread_cond_signal(&pool->input_buffer_one_poped);
        pthread_mutex_unlock(&pool->input_buffer_mutex);

        if(chunk!=NULL) {
            /* process raw */
            
            translation_pair_t * t_item = translation_pair_new();
            translation_pair_set_audio(t_item,chunk);
            
            pthread_mutex_lock(&pool->lm_mutex); /* sphinx functions aren't thread-safe */
            
            int nb_frames = process_raw(pool,chunk);
            debug("processing raw");
            if(nb_frames<0) {
                audio_chunk_free(chunk);
                pthread_mutex_unlock(&pool->lm_mutex);
                asr_manager_terminate(pool);
                myerror("something went wrong while decoding audio");
                pthread_exit(NULL);
            }
            debug("getting asr result");
            translation_pair_get_asr_result(t_item,pool);
            debug("got utterance");
            if(pool->mp_config->one_best && pool->mp_config->verbose) {
                char * string = stream_to_string(translation_pair_get_src_stream(t_item),-1);
                myverbose(pool->mp_config->verbose,"[asr manager] got utterance '%s'.",string);
                free(string);
            }
            else
                myverbose(pool->mp_config->verbose,"[asr manager] got lattice.");
/*            debug("saving asr result");
            if(pool->mp_config->save_asr) {
                translation_pair_save_src(t_item,pool->save_asr_filename,pool);
            }*/
            if(pool->mp_config->save_asr_lattice) {
                char * latfname = malloc(sizeof(char)*(strlen(pool->save_asr_lattice_filename)+10));
                sprintf(latfname,"%s-%.6i",pool->save_asr_lattice_filename,chunk_id);
                ps_lattice_write(translation_pair_get_src_lattice(t_item),latfname);
                free(latfname);
            }
            
            pool->segment_time_offset += pool->segment_time_shift;

            pthread_mutex_unlock(&pool->lm_mutex);
            
            if(pool->mp_config->seg_method!=SEG_NONE && pool->mp_config->seg_method!=SEG_AUDIO) {
                pthread_mutex_lock(&pool->asr_hyp_queue_mutex);
                g_queue_push_head(pool->asr_hyp_queue,t_item);
                pthread_cond_signal(&pool->asr_hyp_queue_one_hyp_pushed);
                pthread_mutex_unlock(&pool->asr_hyp_queue_mutex);
            }
            else {
                pthread_mutex_lock(&pool->segment_queue_mutex);
                g_queue_push_head(pool->segment_queue,t_item);
                pthread_cond_signal(&pool->segment_queue_one_segment_pushed);
                pthread_mutex_unlock(&pool->segment_queue_mutex);
            }
	    /* TODO free the pushed t_item somewhere */

        }
        else {
            /* buffer manager could not fill the buffer, it means input is depleted */
            myverbose(pool->mp_config->verbose,"[asr manager] input is depleted.");
            asr_manager_terminate(pool);
            input_depleted = 1;
        }
        chunk_id += 1;
    }
    myverbose(pool->mp_config->verbose,"[asr manager] bye.");
    pthread_exit(NULL);
}

/* read lattices from files and queue them */
// void *lattice_loader(void * data) {
//     debug("starting");
//     
//     manager_data_t * pool = data;
//     
//     if(!pool->cfg->lattice_input)
//         pthread_exit(NULL);
//     
//     pthread_mutex_t * dest_mutex = pool->mp_config->mt_segment?&pool->asr_hyp_queue_mutex:&pool->segment_queue_mutex;
//     pthread_cond_t * dest_cond = pool->mp_config->mt_segment?&pool->asr_hyp_queue_one_hyp_pushed:&pool->segment_queue_one_segment_pushed;
//     
// 
//     pthread_mutex_lock(dest_mutex);
//     if(g_queue_get_length(pool->input_buffer)>=pool->input_buffer_size) {
//         myverbose(pool->mp_config->verbose,"[lattice loader] waiting for some room in the input buffer");
//         pthread_cond_wait(&pool->input_buffer_one_poped,&pool->lattice_buffer_mutex);
//         myverbose(pool->mp_config->verbose,"[lattice loader] there's room in input buffer. Go.");
//     }
//     
//     
//     ps_lattice_t * lattice = ps_lattice_read(pool->mp_engine->ps,pool->in.filename);
//     
//     g_queue_push_head(pool->input_buffer,lattice);
//     
//     /* signal the new chunk and release mutex */
//     pthread_cond_signal(dest_cond);
//     pthread_mutex_unlock(dest_mutex);
//     
//     myverbose(pool->mp_config->verbose,"[lattice loader] bye.");
//     pthread_exit(NULL);
// }

/* get ASR utterances and merge them into stream */
void *asr_stream_manager(void * data) {
    debug("starting");

    manager_data_t * pool = data;
    
    if(pool->mp_config->seg_method==SEG_NONE || pool->mp_config->seg_method==SEG_AUDIO)
        pthread_exit(NULL);

    int input_depleted = 0;
    while(!input_depleted) {
        pthread_mutex_lock(&pool->asr_hyp_queue_mutex);
        if(g_queue_get_length(pool->asr_hyp_queue)<=1 && !pool->asr_manager_is_dead) {
            myverbose(pool->mp_config->verbose,"[stream manager] waiting for a new utterance");
            pthread_cond_wait(&pool->asr_hyp_queue_one_hyp_pushed,&pool->asr_hyp_queue_mutex);
            myverbose(pool->mp_config->verbose,"[stream manager] received signal to go");
        }
        
        translation_pair_t * t_item = (translation_pair_t*)g_queue_pop_tail(pool->asr_hyp_queue);
        pthread_mutex_unlock(&pool->asr_hyp_queue_mutex);
        pthread_mutex_lock(&pool->asr_stream_mutex);
        if(t_item==NULL) {
            /* asr manager could not fill the queue, it means input is depleted */
            myverbose(pool->mp_config->verbose,"[stream manager] input is depleted.");
            pool->asr_stream_manager_is_dead = 1;
            input_depleted = 1;
        }
        else {
            merge_utt_into_stream(pool,t_item->src_stream);
            if(pool->mp_config->verbose && pool->asr_stream!=NULL) {
                char * str = stream_to_string(pool->asr_stream,-1);
                myverbose(pool->mp_config->verbose,"[stream manager] recognised so far and not yet translated:\n '%s'",str);
                free(str);
            }
            else if(pool->mp_config->verbose && pool->asr_stream_lattice!=NULL) {
                myverbose(pool->mp_config->verbose,"[stream manager] untranslated lattice stream exists, but can't be verbosed.");
            }
        }

        pthread_cond_signal(&(pool->asr_stream_updated));
        pthread_mutex_unlock(&pool->asr_stream_mutex);
    }
    myverbose(pool->mp_config->verbose,"[stream manager] bye.");
    pthread_exit(NULL);
}

/* extract mt segments from stream */
void *segment_manager(void * data) {
    debug("starting");

    manager_data_t * pool = data;
    
    if(pool->mp_config->seg_method==SEG_NONE || pool->mp_config->seg_method==SEG_AUDIO)
        pthread_exit(NULL);

    int input_depleted = 0;
    while(!input_depleted) {
        pthread_mutex_lock(&pool->asr_stream_mutex);
        int chunk_length = 0;
        while((chunk_length = get_mt_segment_from_stream(pool))==0 && !pool->asr_stream_manager_is_dead) {
            myverbose(pool->mp_config->verbose,"[segment manager] waiting for a stream update");
            pthread_cond_wait(&pool->asr_stream_updated,&pool->asr_stream_mutex);
            myverbose(pool->mp_config->verbose,"[segment manager] received signal to go");
        }
        pthread_mutex_unlock(&pool->asr_stream_mutex);

        if(chunk_length==0)  /* if no segment found, but we are here, it means input is depleted. just take whatever remains */
            chunk_length = -1;
        
        debug("writing ctm");
        if(pool->mp_config->save_ctm) write_ctm(pool,chunk_length);
        
        debug("removing extracted segment from pool stream");
        char * segment_str = stream_to_string(pool->asr_stream,chunk_length);
        debug("got segment '%s'",segment_str);
        
        pool->asr_stream = asr_stream_free_n(pool->asr_stream,chunk_length);
        if(pool->asr_stream!=NULL) {
            pool->asr_stream->prev = NULL;
        }

        pthread_mutex_lock(&pool->segment_queue_mutex);
        if(segment_str==NULL) {
            /* input is depleted and stream is empty. die. */
            myverbose(pool->mp_config->verbose,"[segment manager] input is depleted.");
            input_depleted = 1;
            pool->segment_manager_is_dead = 1;
        }
        else {
            translation_pair_t * segment = translation_pair_new_from_string(segment_str);
            debug("pushing segment '%s'",translation_pair_get_src_string(segment));
            g_queue_push_head(pool->segment_queue,segment);
        }
        pthread_cond_signal(&pool->segment_queue_one_segment_pushed);
        pthread_mutex_unlock(&pool->segment_queue_mutex);

     free(segment_str);
    }
    myverbose(pool->mp_config->verbose,"[segment manager] bye.");
    pthread_exit(NULL);
}


/* translate mt segments */
void *translation_manager(void * data) {
    debug("starting");

    manager_data_t * pool = data;

    int input_depleted = 0;
    while(!input_depleted) {
        
        translation_pair_t * src;
        
        if(pool->mp_config->lattice_input) {
            ps_lattice_t * lattice = ps_lattice_read(pool->mp_engine->ps,pool->in_filename);
            src = translation_pair_new_from_lattice(lattice);
            input_depleted = 1;
        }
        else {
            pthread_mutex_lock(&pool->segment_queue_mutex);
            /*debug("segments queue contains %i elements",g_queue_get_length(pool->segment_queue));*/
            if(g_queue_is_empty(pool->segment_queue) && !pool->segment_manager_is_dead) {
                myverbose(pool->mp_config->verbose,"[translation manager] waiting for a new segment");
                pthread_cond_wait(&pool->segment_queue_one_segment_pushed,&pool->segment_queue_mutex);
                myverbose(pool->mp_config->verbose,"[translation manager] received signal to go");
            }
            src = (translation_pair_t *)g_queue_pop_tail(pool->segment_queue);
            pthread_mutex_unlock(&pool->segment_queue_mutex);
        }

        if(src==NULL) {
            /* input depleted */
            myverbose(pool->mp_config->verbose,"[translation manager] input is depleted. bye.");
            input_depleted = 1;
            pool->translation_manager_is_dead = 1;
        }
        else {
            
            char * translation = moses_translate(pool->mp_config,pool->mp_engine,src);
            myverbose(pool->mp_config->verbose,"[translation manager] got translation '%s'",translation);
            myfprintf(pool->out,"%s\n",translation);
            myfflush(pool->out);

            translation_pair_set_target(pool, src, translation);
            if(!pool->mp_config->no_cm)
                compute_features(pool->features, pool->cm_res, src);
            translation_pair_add_xml(src,pool->xml_result_node);

            debug("freeing translation_pair");
            translation_pair_free(src);
            debug("freeing hyp string");
            free(translation);
        }
    }

    xmlSaveFormatFileEnc(pool->out.filename, pool->xml_result_doc, "UTF-8", 1);
    debug("bye.");
    pool->translation_manager_is_dead = 1;
    pthread_exit(NULL);
}
