/*
Copyright (c) 2011, Sylvain Raybaud <sylvain.raybaud@gmail.com>
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
Neither the name of the PAROLE team, LORIA, INRIA nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <math.h>
#include <glib.h>

#include <cutils.h>

#include <logmath.h>
#include <ps_lattice.h>

#include <Manager.h>
#include <StaticData.h>
#include <Parameter.h>
#include <Sentence.h>
#include <WordLattice.h>
#include <TranslationSystem.h>
#include <Factor.h>
using namespace Moses;

#include <vector>
#include <string>
using namespace std;

#include "interface_moses.h"
#include "interface_moses_cplusplus.h"
#include "maparole.h"
#include "engine.h"
#include "common.h"
#include "lattice.h"

    
/* shall return a pointer to a Param object */
int moses_read_config_file(const char * cfgfilename, moses_config_t * cfg) {
    Parameter * params = new Parameter();
    bool status = params->LoadParam(string(cfgfilename));
    cfg->moses_parameters = params;
    return status?1:0;
}

void moses_terminate(moses_config_t * config) {
  Parameter * params = (Parameter*) config->moses_parameters;
  delete params;
}

int moses_static_data(moses_config_t * config) {
    return StaticData::LoadDataStatic(((Parameter*)config->moses_parameters))?1:0;
}

static string moses_get_hyp(const Hypothesis* hypo) {
    
    string current("");
    TargetPhrase p = hypo->GetTargetPhrase();
    for (size_t pos = 0 ; pos<p.GetSize() ; pos++) {
        const Factor *factor = p.GetFactor(pos, 0);
        current += factor->GetString()+string(" ");
    }
    
    const Hypothesis * prev = hypo->GetPrevHypo();
    if(prev != NULL)
        return moses_get_hyp(prev)+string(" ")+current;
    return current;
}


static char * moses_translate_cn(mp_config_t * config, mp_engine_t * engine, ps_lattice_t * lat) {
    
    const TranslationSystem& system = StaticData::Instance().GetTranslationSystem(TranslationSystem::DEFAULT); /* there is only one translation system for now */
    const StaticData &staticData = StaticData::Instance();
    const vector<FactorType> &inputFactorOrder = staticData.GetInputFactorOrder();
    
    PCN::CN cn = slf2cn(lat);
    WordLattice * lattice = new WordLattice();
    lattice->InitializeFromPCNDataType(cn,inputFactorOrder);
//     InputType * input = new WordLattice();
//     input->InitializeFromPCNDataType(*cn,inputFactorOrder);
    
    Manager * manager = new Manager(*lattice,staticData.GetSearchAlgorithm(), &system);
    manager->ProcessSentence();
    const Hypothesis* hypo = manager->GetBestHypothesis();
    
    string hyp = moses_get_hyp(hypo);
    char * hyp_ret = (char*)malloc((strlen(hyp.c_str())+1)*sizeof(char));
    strcpy(hyp_ret,hyp.c_str());
    
    delete lattice;
    delete manager;
//     delete input;
    return hyp_ret;
}


static char * moses_translate_string(mp_config_t * config, mp_engine_t * engine, char const * text) {

    const TranslationSystem& system = StaticData::Instance().GetTranslationSystem(TranslationSystem::DEFAULT); /* there is only one translation system for now */
    const StaticData &staticData = StaticData::Instance();
    const vector<FactorType> &inputFactorOrder = staticData.GetInputFactorOrder();
    
    string source(text);
    istringstream * in = new istringstream(source+"\n",istringstream::in);
    InputType * input = new Sentence(Input);
    input->Read(*in,inputFactorOrder);
    
    Manager * manager = new Manager(*input,staticData.GetSearchAlgorithm(), &system);
    manager->ProcessSentence();
    const Hypothesis* hypo = manager->GetBestHypothesis();

    string hyp = moses_get_hyp(hypo);
    char * hyp_ret = (char*)malloc((strlen(hyp.c_str())+1)*sizeof(char));
    strcpy(hyp_ret,hyp.c_str());

    delete in;
    delete manager;
    delete input;
    return hyp_ret;
}

char * moses_translate(mp_config_t * config, mp_engine_t * engine, translation_pair_t * t) {
    if(!config->one_best) {
        return moses_translate_cn(config,engine,translation_pair_get_src_lattice(t));
    }
    else {
        return moses_translate_string(config,engine,translation_pair_get_src_string(t));
    }
}


PCN::CN slf2cn(ps_lattice_t * lat) {
    debug("start");
    
    PCN::CN cn;
    
    GRegex * noise_re = mk_noise_re();
    GSList * ordered_nodes = ps_lattice_topological_sort(lat);
    
    
    GSList * node_elmt = ordered_nodes;
    while(node_elmt!=NULL) {
        PCN::CNCol col;
        ps_latnode_t * node = (ps_latnode_t*)node_elmt->data;
        ps_latlink_iter_t * alt = ps_latnode_exits(node);
        while(alt!=NULL) {
            ps_latlink_t * link = ps_latlink_iter_link(alt);
            vector<float> scores;
            
            int32 l_postprob_i = ps_latlink_prob(lat,link,NULL);
            float64 l_postprob_f = logmath_exp(ps_lattice_get_logmath(lat),l_postprob_i);
            scores.push_back((float)l_postprob_f);
            
            const char * w = ps_latlink_baseword(lat,link);
            string word(match_noise(noise_re,w)?"*EPS*":w);
            pair<string,vector<float> > linkdata(word,scores);
            
            ps_latnode_t * dest = ps_latlink_nodes(link,NULL);
            int distance = g_slist_index(node_elmt,dest);
            
            PCN::CNAlt cnalt(linkdata,size_t(distance));
            col.push_back(cnalt);
            
            alt = ps_latlink_iter_next(alt);
        }
        cn.push_back(col);
    }
    
    g_regex_unref(noise_re);
    
    return cn;
}
