#ifndef _LANGUAGE_MODEL_H_
#define _LANGUAGE_MODEL_H_

#include <pocketsphinx.h>


/* 
   Module providing a generic way to use Sphinx's and Moses language models.
   It allows to retrieve loaded LMs from running instances.

   Genericity is achieved through those types:
   - wid_t: a word index in a given LM
   - lmi_t: a generic interface
   - lmi_data_t: the actual LM used by the functions

   For example, to use a Sphinx LM :
   - fill a sphinx_lm_t structure
   - create a lmi_t with sphinx_lmi_new(...)
   Then you can transparently use the lmi_t lmi, e.g.
   - wid_t foo = lmi.wid(lmi.data, "foo");
   - float p = lmi.score_bigram(lmi.data, foo, foo); // Gives you p("foo foo")
 */

/* Word index in a LM */
typedef union wid_s {
    int32 idx;          /* for Sphinx, it's a simple int */
    void *adr;            /* for Moses, it's a pointer to an Object */
} wid_t;

/* Structure providing a generic access to a language model implementation (lmi)
 * in an object-based fashion: the first argument of every "member" function call
 * should be the 'data' field of the structure. */
typedef void * lmi_data_t;
typedef struct lmi_s {
    /* p(a) */
    float (*score_unigram)(lmi_data_t, wid_t a);
    /* p(b|a) */
    float (*score_bigram)(lmi_data_t, wid_t a, wid_t b);
    /* p(c|a,b) */
    float (*score_trigram)(lmi_data_t, wid_t a, wid_t b, wid_t c);
    /* p(t[n-1]|t[n-2],...,t[0]) */
    float (*score_ngram)(lmi_data_t, int n, wid_t *t);
    /* true iff the LM contains 'a' */
    int (*unigram_exists)(lmi_data_t, wid_t a);
    /* true iff the LM contains 'a b' */
    int (*bigram_exists)(lmi_data_t, wid_t a, wid_t b);
    /* true iff the LM contains 'a b c' */
    int (*trigram_exists)(lmi_data_t, wid_t a, wid_t b, wid_t c);
    /* Get the index of 'word' in this LM */
    wid_t (*wid)(lmi_data_t, char * word);
    /* free the data of the LM
       (do not call this if the LM is retrieved from the running instance */
    void (*free_data)(lmi_data_t);
    /* Pointer to the actual data of the implementation */
    lmi_data_t * data;
    /* Miscellaneous */
    wid_t bos;            /* index of <s> */
    wid_t eos;            /* index of </s> */
} lmi_t;

/* Structure referencing a Sphinx LM */
typedef struct sphinx_lm_s {
    ngram_model_t *lm;        /* The actual LM */
    logmath_t *lmath;     /* Sphinx object for log-proba convertions
    note: it is a field of Sphinx's internal "struct ngram_model_s",
    but it is not accessible in ngram_model_t, hence this structure */
    pthread_mutex_t mutex;
} sphinx_lm_t;

sphinx_lm_t *sphinx_open_lm(char *filename, logmath_t *lmath);

/* Return by value, just need to free 'data' */
lmi_t sphinx_lmi_new(sphinx_lm_t *data);

#ifdef __cplusplus
extern "C" {
#endif

  /* Pointer to a Moses LM (C++ type:  Moses::LanguageModelImplementation*) */
  typedef void* moses_lm_t;

  /* idem, by value */
  lmi_t moses_lmi_new(moses_lm_t *data);

  /* Retrieve the model language from the running instance of Moses */
  moses_lm_t moses_get_lm();
  /* Open a model language */
  moses_lm_t moses_lm_open(char *fileName);

#ifdef __cplusplus
} /* extern "C" */
#endif


#endif
