/*
Copyright (c) 2011, Sylvain Raybaud <sylvain.raybaud@gmail.com>
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
Neither the name of the PAROLE team, LORIA, INRIA nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdio.h>
#include <glib.h>
#include <pthread.h>
#include <sys/time.h>

#include <ngram_model.h>
#include <pocketsphinx.h>

#include <cutils.h>

#include "maparole.h"
#include "interface_ps.h"
#include "interface_moses.h"
#include "common.h"

#include "engine.h"

mp_engine_t mp_start_engine(mp_config_t * cfg) {
    
    mp_engine_t engine = {0};
    
    /* start engine */
    
    if(cfg->input_is_audio) {
        engine.ps = ps_init(cfg->ps_config.ps_cmd_ln);
        if (engine.ps == NULL) {
            myerror("Could not intialize PocketSphinx decoder");
            exit(EXIT_PS_ERROR);
        }
    }
    
    if (!moses_static_data(&(cfg->moses_config))) {
        myerror("Could not intialize Moses decoder");
        exit(EXIT_MOSES_ERROR);
    }
    
    return engine;
}

void mp_terminate(mp_config_t * config, mp_engine_t * engine) {
    debug("ps engine");
    ps_free(engine->ps);
    debug("closing files");
    if(config->save_asr_filename) {
        fclose(config->save_asr_file);
        free(config->save_asr_filename);
    }
    if(config->save_raw_mt_seg_filename) {
        fclose(config->save_raw_mt_seg_file);
        free(config->save_raw_mt_seg_filename);
    }
    if(config->save_clean_mt_seg_filename) {
        fclose(config->save_clean_mt_seg_file);
        free(config->save_clean_mt_seg_filename);
    }
}


void mp_start_translation(mp_config_t * cfg, mp_engine_t * engine) {
    debug("start");
    myverbose(cfg->verbose,"starting translation manager");
    
    manager_data_t pool = manager_data_init(cfg,engine);
    
    pthread_t asr_buffer_manager_thread;
    pthread_t asr_manager_thread;
    pthread_t translation_manager_thread;
    
    int asr_buffer_manager_created = pthread_create(&asr_buffer_manager_thread,NULL,asr_buffer_manager,(void*)&pool);
    int asr_manager_created = pthread_create(&asr_manager_thread,NULL,asr_manager,(void*)&pool);
    int translation_manager_created = pthread_create(&translation_manager_thread,NULL,translation_manager,(void*)&pool);
    
    if(asr_buffer_manager_created || asr_manager_created || translation_manager_created) {
        manager_data_free(&pool);
        myfail(EXIT_UNKNOWN_ERROR,"Could not start threads.");
    }
    
    
    /* end */
    myverbose(cfg->verbose,"waiting for threads to finish...");
    
    pthread_join(asr_buffer_manager_thread,NULL);
    pthread_join(asr_manager_thread,NULL);
    pthread_join(translation_manager_thread,NULL);
    
    myverbose(cfg->verbose,"freeing pool");
    manager_data_free(&pool);
    
}

manager_data_t manager_data_init(mp_config_t * cfg, mp_engine_t * engine) {
    debug("start");
    myverbose(cfg->verbose,"populating pool");
    
    manager_data_t pool;
    
    pool.mp_config = cfg;
    pool.mp_engine = engine;
    
    fe_t * fe = ps_get_fe(engine->ps);
    fe_get_input_size(fe,NULL,&pool.samples_per_frame);
    int sampling_rate = (int)cmd_ln_float_r(cfg->ps_config.ps_cmd_ln,"-samprate");
    
    if(cfg->asr_buffer_size%2==1) {
        cfg->asr_buffer_size += 1;
        myinfo("Increasing number of bufferised segments by 1 (now %i) to make it even.",cfg->asr_buffer_size);
    }
    /* convert durations to sizes (number of int16) */
    pool.asr_segment_size = sampling_rate*cfg->asr_segment_length;
    pool.asr_segment_overlap = sampling_rate*cfg->asr_segment_overlap;
    pool.overlap_frames_offset = (pool.asr_segment_size-pool.asr_segment_overlap)/pool.samples_per_frame;
    
    /* compute buffer size from segment size and overlap */
    pool.asr_buffer_size = (pool.asr_segment_size-pool.asr_segment_overlap)*cfg
    ->asr_buffer_size/2+pool.asr_segment_overlap;
    pool.asr_buffer2_size = pool.asr_buffer_size;
    
    pool.asr_buffer = malloc(pool.asr_buffer_size*sizeof(int16));
    pool.asr_buffer2 = malloc(pool.asr_buffer2_size*sizeof(int16));
    pool.asr_segment = malloc(pool.asr_segment_size*sizeof(int16));
    pool.asr_buffer_segment_ptr = 0;
    
    pthread_mutex_init(&pool.asr_buffer_mutex,NULL);
    pthread_cond_init(&pool.asr_buffer2_empty,NULL);
    pthread_cond_init(&pool.asr_buffer2_refill_done,NULL);
    pool.asr_buffer2_refill_in_progress = 0;
    pthread_mutex_init(&pool.lm_mutex,NULL);
    pthread_cond_init(&pool.asr_new_utterance_cond,NULL);
    
    pool.asr_stream = NULL;
    pthread_mutex_init(&pool.asr_stream_mutex,NULL);
    
    pool.lm = ngram_model_set_select(ps_get_lmset(engine->ps),"default");
    pool.sentence_start_id = ngram_wid(pool.lm,"<s>");
    pool.lm_zero = ngram_zero(pool.lm);
    
    /* cut on '.', ';', ':', '?', '!', '</s>' */
    /* should also cut on ',' */
    pool.unk_id = ngram_unknown_wid(pool.lm);
    int id_ptr = 0;
    int32 id;
    pool.mt_cut_symbols_id = malloc(8*sizeof(int32));
    if((id = ngram_wid(pool.lm,","))!=pool.unk_id)
        pool.mt_cut_symbols_id[id_ptr++] = id;
    if((id = ngram_wid(pool.lm,";"))!=pool.unk_id)
        pool.mt_cut_symbols_id[id_ptr++] = id;
    if((id = ngram_wid(pool.lm,":"))!=pool.unk_id)
        pool.mt_cut_symbols_id[id_ptr++] = id;
    if((id = ngram_wid(pool.lm,"?"))!=pool.unk_id)
        pool.mt_cut_symbols_id[id_ptr++] = id;
    if((id = ngram_wid(pool.lm,"!"))!=pool.unk_id)
        pool.mt_cut_symbols_id[id_ptr++] = id;
    if((id = ngram_wid(pool.lm,"."))!=pool.unk_id)
        pool.mt_cut_symbols_id[id_ptr++] = id;
    if((id = ngram_wid(pool.lm,"</s>"))!=pool.unk_id)
        pool.mt_cut_symbols_id[id_ptr++] = id;
    pool.mt_cut_symbols_id[id_ptr] = pool.unk_id;
    
    myverbose(cfg->verbose,"pool info:");
    myverbose(cfg->verbose,"  sampling rate = %i",sampling_rate);
    myverbose(cfg->verbose,"  frame size = %i samples",pool.samples_per_frame);
    myverbose(cfg->verbose,"  overlap frames offset = %i frames",pool.overlap_frames_offset);
    myverbose(cfg->verbose,"  asr buffer = %li samples",pool.asr_buffer_size);
    myverbose(cfg->verbose,"  asr segment = %li samples",pool.asr_segment_size);
    myverbose(cfg->verbose,"  asr segment overlap = %li samples",pool.asr_segment_overlap);
    
    myverbose(cfg->verbose,"opening audio file '%s'.",cfg->audiofilename);
    pool.audiofile = fopen(cfg->audiofilename,"rb");
    if(pool.audiofile==NULL) {
        myfail(EXIT_PS_ERROR,"cannot open input file '%s'.",cfg->audiofilename);
    }
    pool.input_depleted = 0;
    
    pool.asr_buffer_segment_ptr = 0;
    size_t n_samples_read = fread(pool.asr_buffer,2,pool.asr_buffer_size,pool.audiofile);
    if(n_samples_read==0) {
        manager_data_free(&pool);
        myfail(EXIT_PS_ERROR,"cannot read input file.");
    }
    asr_buffer2_fill(&pool);
    
    return pool;
}


void manager_data_free(manager_data_t * pool) {
    debug("start");
    free(pool->asr_buffer);
    free(pool->asr_buffer2);
    free(pool->asr_segment);
    free(pool->mt_cut_symbols_id);
    if(pool->audiofile!=NULL)
        fclose(pool->audiofile);
    if(pool->asr_stream!=NULL && pool->asr_stream->word!=NULL) {
        mywarning("there are untranslated tokens in asr_stream");
        asr_stream_free(pool->asr_stream);
    }
}



/* make sure that asr_buffer always has enough data */
/* this task could be handled by asr_manager...     */
/* I guess it's only useful if I/O is very slow     */
void *asr_buffer_manager(void * data) {
    debug("start");
    
    manager_data_t * pool = data;
    
    struct timeval now;
    struct timespec timeout;
    
    while(1) {
        
        /* wait until asr_manager signals an empty buffer */
        pthread_mutex_lock(&pool->asr_buffer_mutex);
        while(!pool->asr_buffer2_refill_in_progress) {
            gettimeofday(&now,NULL);
            timeout.tv_sec = (now.tv_usec<(1000000-DEFAULT_WAIT))?(now.tv_sec):(now.tv_sec+1);
            timeout.tv_nsec = ((now.tv_usec+DEFAULT_WAIT)%1000000)*1000;
            pthread_cond_timedwait(&(pool->asr_buffer2_empty),(&pool->asr_buffer_mutex),&timeout);
        }
        
        size_t n_new_samples = asr_buffer2_fill(pool);
        pool->asr_buffer2_refill_in_progress = 0;
        pthread_cond_signal(&(pool->asr_buffer2_refill_done));
        
        pthread_mutex_unlock(&(pool->asr_buffer_mutex));
        
        if(n_new_samples==0) {
            myverbose(pool->mp_config->verbose,"[buffer manager] input is depleted. bye.");
            break;
        }
    }
    pthread_exit(0);
}

void *asr_manager(void * data) {
    debug("start");
    
    manager_data_t * pool = data;
    
    while(1) {
        /* get raw samples */
        size_t n_samples = get_asr_segment(pool); /* will block until it can read something */
        if(n_samples==0) {
            /* input depleted */
            myverbose(pool->mp_config->verbose,"nothing left to process. bye.");
            break;
        }
        
        pthread_mutex_lock(&pool->lm_mutex);
        
        /* process raw */
        int new_utterance = ps_start_utt(pool->mp_engine->ps,NULL);
        new_utterance = ps_process_raw(pool->mp_engine->ps,pool->asr_segment,n_samples,FALSE,TRUE);
        new_utterance = ps_end_utt(pool->mp_engine->ps);
        
        pthread_mutex_unlock(&pool->lm_mutex);
        
        /* get hypothesis */
        ps_seg_t * seg = ps_seg_iter(pool->mp_engine->ps,NULL);
        asr_stream_t * utt = get_asr_stream_from_utterance(seg);
        
        pthread_cond_signal(&(pool->asr_new_utterance_cond));
        
        myverbose(pool->mp_config->verbose,"[ASR] got utterance '%s'",stream_to_string(utt,-1));
        if(pool->mp_config->save_asr_file) {
            fprintf(pool->mp_config->save_asr_file,"%s\n",stream_to_string_raw(utt,-1));
        }
        
        merge_utt_into_stream(pool,utt);
        
        myverbose(pool->mp_config->verbose,"[ASR] recognised so far and not yet translated:\n '%s'",stream_to_string(pool->asr_stream,-1));
    }
    pthread_exit(0);
}

void *translation_manager(void * data) {
    debug("start");
    
    manager_data_t * pool = data;
    
    while(1) {
        /* get mt friendly segment from asr stream */
        /* block until you get it */
        myverbose(pool->mp_config->verbose,"waiting for an mt-friendly segment.");
        char * segment = get_mt_segment_from_stream(pool);
        if(segment==NULL) {
            /* input depleted */
            myverbose(pool->mp_config->verbose,"[translation manager] input is depleted. bye.");
            break;
        }
        if(pool->mp_config->save_raw_mt_seg_file) {
            fprintf(pool->mp_config->save_raw_mt_seg_file,"%s\n",segment);
        }
        
        char * clean_segment = clean_for_mt(segment);
        myverbose(pool->mp_config->verbose,"[translation manager] translating '%s'",clean_segment);
        if(pool->mp_config->save_clean_mt_seg_file) {
            fprintf(pool->mp_config->save_clean_mt_seg_file,"%s\n",clean_segment);
        }
        
        char * translation = moses_translate_string(pool->mp_config,pool->mp_engine,clean_segment);
        myverbose(pool->mp_config->verbose,"[translation manager] got translation '%s'",translation);
        printf("%s\n",translation);
        free(segment);
        free(clean_segment);
	free(translation);
    }
    pthread_exit(0);
}

/* get a segment of input audio for ASR */
/* store it in the pool and return number of samples read */
/* block until you get it */
size_t get_asr_segment(manager_data_t * pool) {
    debug("start");
    
    size_t * segment_ptr = &pool->asr_buffer_segment_ptr;
    size_t * buffer_size = &pool->asr_buffer_size;
    size_t * buffer2_size = &pool->asr_buffer2_size;
    size_t overlap = pool->asr_segment_overlap;
    
    
    myverbose(pool->mp_config->verbose,"buffer info attempt at getting ASR segment:");
    myverbose(pool->mp_config->verbose,"  buffer:    size: %li   address: %p",*buffer_size,(void*)pool->asr_buffer);
    myverbose(pool->mp_config->verbose,"  buffer2:   size: %li   address: %p",*buffer2_size,(void*)pool->asr_buffer2);
    myverbose(pool->mp_config->verbose,"  segment pointer: %li",*segment_ptr);
    
    
//     if(*segment_ptr==*buffer_size-overlap || *segment_ptr==*buffer_size) {
    if(*segment_ptr==*buffer_size-overlap) {
        
        /* if a refill operation is in progress, wait until it's finished */
        pthread_mutex_lock(&(pool->asr_buffer_mutex));
        if(pool->asr_buffer2_refill_in_progress) {
            pthread_cond_wait(&(pool->asr_buffer2_refill_done),&(pool->asr_buffer_mutex));
        }
        
        /* buffer1 depleted. attempting to use buffer2 and refill */
        if(*buffer2_size==0) {
            /* it's over */
            pool->input_depleted = 1;
	    pthread_cond_signal(&(pool->asr_new_utterance_cond));
            pthread_mutex_unlock(&pool->asr_buffer_mutex);
            myverbose(pool->mp_config->verbose,"nothing left in asr buffer.");
            return 0;
        }
        else {
            /* input not depleted, move buff2 to buff1 and refill buff2 */
            int16 * buff_tmp = pool->asr_buffer;
            pool->asr_buffer = pool->asr_buffer2;
            pool->asr_buffer2 = buff_tmp;
            *segment_ptr = 0;
            *buffer_size = *buffer2_size;
            pool->asr_buffer2_refill_in_progress = 1;
            pthread_cond_signal(&(pool->asr_buffer2_empty));
            pthread_mutex_unlock(&pool->asr_buffer_mutex);
        }
    }
    
    /* populate segment */
    size_t n_samples = min(pool->asr_segment_size,(*buffer_size-*segment_ptr));
    memcpy(pool->asr_segment,&(pool->asr_buffer[*segment_ptr]),2*n_samples);
    *segment_ptr += n_samples-overlap;
    
    return n_samples;
}

/* fill asr_buffer2 */
/* *** CAUTION *** */
/* this function assumes that pool->asr_buffer_mutex is locked! */
size_t asr_buffer2_fill(manager_data_t * pool) {
    debug("start");
    
    size_t overlap = pool->asr_segment_overlap;
    int16 * buffer = pool->asr_buffer;
    int16 * buffer2 = pool->asr_buffer2;
    size_t buffer_size = pool->asr_buffer_size;
    size_t buffer2_size = pool->asr_buffer2_size;
    
    /* copy overlapping parts of buff1 in buffer2 */
    memcpy(buffer2,&(buffer[buffer_size-overlap]),2*overlap);
    
    /* read new data from file */
    size_t n_samples_read = fread(&(buffer2[overlap]),2,buffer2_size-overlap,pool->audiofile);
    
    /* update buffer size (changes if input is depleted) */
    pool->asr_buffer2_size = (n_samples_read>0)?n_samples_read+overlap:0;
    return n_samples_read;
}

/* get best hypothesis as a linked list of items */
asr_stream_t * get_asr_stream_from_utterance(ps_seg_t * seg) {
    debug("start");
    
    ps_seg_t * segment = seg;
    asr_stream_t * stream = NULL;
    
    while(segment!=NULL) {
        const char * word = ps_seg_word(segment);
        if(strcmp(word,"<s>") && strcmp(word,"</s>")) {
            asr_stream_t * prev = stream;
            stream = asr_stream_new_item(remove_variant(word),prev,NULL,-1,-1,is_non_word(word));
            if(prev!=NULL) {
                prev->next = stream;
            }
            ps_seg_frames(segment,&(stream->start_frm),&(stream->end_frm));
        }
        segment = ps_seg_next(segment);
    }
    
    return asr_stream_rewind(stream);
}

/* merge best hypothesis into the stream of recognised items */
void merge_utt_into_stream(manager_data_t * pool, asr_stream_t * utt) {
    debug("start");
    
    if(utt==NULL) {
        return;
    }
    
    pthread_mutex_lock(&(pool->asr_stream_mutex));
    
    /* naive method (ester) */
    /* get longest common word */
    /* merge from that word */
    /* if none found, consider that last but one word in pool is best */
    
    asr_stream_t * left_utt = asr_stream_get_end(pool->asr_stream);
    if(left_utt==NULL) {
        /* no left utterance, nothing to merge into */
        pool->asr_stream = utt;
        pthread_mutex_unlock(&(pool->asr_stream_mutex));
        return;
    }
    
    int left_last_frame = left_utt->end_frm;
    int right_first_frame = utt->start_frm+pool->overlap_frames_offset;
    
    int greatest_length = 0;
    asr_stream_t * cut_left = NULL; /* cut left utterance after (not including)... */
    asr_stream_t * cut_right = NULL; /* cut right utterance before (including)... */
    while(left_utt!=NULL && left_utt->end_frm>right_first_frame) {
        asr_stream_t * right_utt = utt;
        while(right_utt!=NULL
            && right_utt->start_frm+pool->overlap_frames_offset<left_last_frame) {
            if(left_utt->word!=NULL && right_utt->word!=NULL
                && !strcmp(right_utt->word,left_utt->word)
                && left_utt->end_frm-left_utt->start_frm+right_utt->end_frm-right_utt->start_frm>greatest_length) {
                cut_left = left_utt;
                cut_right = right_utt;
                greatest_length = left_utt->end_frm-left_utt->start_frm+right_utt->end_frm-right_utt->start_frm;
            }
            right_utt = right_utt->next;
        }
        left_utt = left_utt->prev;
    }
    
    /* if no common word found, we decide that left utterance */
    /* is more reliable up to the penultimate item */
    /* and right one after its second item */
    /*if(cut_left==NULL) {
        debug("no merge point found, trying my best guess");
        cut_left = asr_stream_get_end(pool->asr_stream)->prev;
        cut_right = utt->next;
    }*/
    
    /* if no common word found just append */
    if(cut_left==NULL) {
        utt->prev = asr_stream_get_end(pool->asr_stream);
        asr_stream_get_end(pool->asr_stream)->next = utt;
    }
    else {
        asr_stream_free(cut_left->next);
        cut_left->next = cut_right->next;
        if(cut_left->next!=NULL) {
            cut_left->next->prev = cut_left;
        }
        asr_stream_free_backward(cut_right);
    }
    
    pthread_mutex_unlock(&(pool->asr_stream_mutex));
    
}

