/*
Copyright (c) 2011, Sylvain Raybaud <sylvain.raybaud@gmail.com>
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
Neither the name of the PAROLE team, LORIA, INRIA nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>

#include <cutils.h>

#include "common.h"
#include "parse_cl.h"
#include "maparole.h"

void parse_cl(mp_config_t * config, int argc, char ** argv) {
    
    static const char * usage = "Usage: %s <options> [input files] [< input audio] [> translated text]\nOptions are:\n\
    --compress or -z: compress output (no effect if -o is used)\n\
    --stdout or -o: output translated text to stdout instead of <file>.s2tt\n\
    --prefix or -P: prefix for output files\n\
    --moses-config or -m: moses configuration file (defaults to moses.ini)\n\
    --pocketsphinx-config or -p: pocketsphinx configuration file (defaults to ps.cfg)\n\
    --s2tt-config or -s: speech to text translation configuration file (defaults to s2tt.cfg)\n\
    --source-lm or -l: source language model (for recognition)\n\
    --target-lm or -L: target language model (for translation)\n\
    --hmm or -H: source language accoustic model (for recognition)\n\
    --dict or -d: source language dictionary (for recognition)\n\
    --fdict or -f: filler/noise dictionary (for recognition)\n\
    --asr-segment or -S: duration of speech chunks to recognise. if too low, sar accuracy will be poor. if too high, it will not work at all. I suggest you stick to the default (10s)\n\
    --asr-buffer or -B: number of ASR segment to bufferize (default=10)\n\
    --asr-segment-overlap or -O: how much asr segments overlap. bigger = slower but better accuracy. Must be bigger than one frame and smaller than one segment. (in seconds, default=3).\n\
    --mt-segmentation or -t (emit|prefer|best|best-prefer|silence|length|mix|none): choose ASR hypothesis resegmetnation method. (default=mix)\n\
    --max-mt-chunk or -c: maximum length of segments passed to MT system (default=40)\n\
    --save-asr or -A: save raw ASR hypothesis (including noise and pronunciation variants) to <file>.asr or <timestamp>.asr if input is mic or stdin.\n\
    --save-raw-mt-seg or -R: save raw segments passed to MT system (including noise and pronunciation variants) to <file>.raw_mt_seg or <timestamp>.raw_mt_seg if input is mic or stdin.\n\
    --save-clean-mt-seg or -C: save clean segments passed to MT system (retokenised, no acoustic indications) to <file>.clean_mt_seg or <timestamp>.clean_mt_seg if input is mic or stdin.\n\
    --verbose or -v: print out some details on stderr\n\
    --help or -h: print this message and exit\n";
    
    
    static struct option long_options[] = {
        {"verbose",             no_argument,       NULL, 'v'}, /* verbosity */
        {"compress",            no_argument,       NULL, 'z'},
        {"stdout",              no_argument,       NULL, 'o'},
        {"prefix",              required_argument, NULL, 'P'},
        {"moses-config",        required_argument, NULL, 'm'},
        {"pocketsphinx-config", required_argument, NULL, 'p'},
        {"s2tt-config",         required_argument, NULL, 's'},
        {"source-lm",           required_argument, NULL, 'l'},
        {"target-lm",           required_argument, NULL, 'L'},
        {"hmm",                 required_argument, NULL, 'H'},
        {"dict",                required_argument, NULL, 'd'},
        {"fdict",               required_argument, NULL, 'f'},
        {"mdef",                required_argument, NULL, 'M'},
        {"asr-segment",         required_argument, NULL, 'S'},
        {"asr-segment-overlap", required_argument, NULL, 'O'},
        {"asr-buffer",          required_argument, NULL, 'B'},
        {"max-mt-chunk",        required_argument, NULL, 'c'},
        {"save-asr",            no_argument,       NULL, 'A'},
        {"save-raw-mt-seg",     no_argument,       NULL, 'R'},
        {"save-clean-mt-seg",   no_argument,       NULL, 'C'},
        {"help",                no_argument,       NULL, 'h'},
        {0, 0, 0, 0}
    };
    
    int help = 0;
    
    int option_code = 1;
    
    while(option_code!=-1) {
        
        int option_index = 0;
        option_code = getopt_long(argc, argv, "s:p:m:l:L:H:f:M:d:S:O:B:t:c:P:ARCozvh",long_options, &option_index);
        switch(option_code) {
            case 0:  /* long option setting a flag */
                break;
            case 'v':
                config->verbose = 1;
                break;
            case 'z':
                config->compress_out = 1;
                break;
            case 'h':
                help = 1;
                break;
            case 'P':
                config->prefix = optarg;
                break;
            case 'm':
                config->cfg_moses_filename = optarg;
                break;
            case 'p':
                config->cfg_ps_filename = optarg;
                break;
            case 's':
                config->cfg_s2tt_filename = optarg;
                break;
            case 'l':
                config->ps_config.lm_filename = optarg;
                break;
            case 'L':
                config->moses_config.lm_filename = optarg;
                break;
            case 'H':
                config->ps_config.hmm_filename = optarg;
                break;
            case 'f':
                config->ps_config.fdict_filename = optarg;
                break;
            case 'M':
                config->ps_config.mdef_filename = optarg;
                break;
            case 'd':
                config->ps_config.dict_filename = optarg;
                break;
            case 'S':
                config->asr_segment_length = atoi(optarg);
                break;
            case 'O':
                config->asr_segment_overlap = atoi(optarg);
                break;
            case 'B':
                config->asr_buffer_size = atoi(optarg);
                break;
            case 't':
                if      (!strcmp(optarg,"emit")) config->seg_method = SEG_EMIT;
                else if (!strcmp(optarg,"prefer")) config->seg_method = SEG_PREFER;
                else if (!strcmp(optarg,"best")) config->seg_method = SEG_BEST;
                else if (!strcmp(optarg,"best-prefer")) config->seg_method = SEG_BEST_PREFER;
                else if (!strcmp(optarg,"silence")) config->seg_method = SEG_SILENCE;
                else if (!strcmp(optarg,"length")) config->seg_method = SEG_LENGTH;
                else if (!strcmp(optarg,"mix")) config->seg_method = SEG_MIX;
                else if (!strcmp(optarg,"none")) config->seg_method = SEG_NONE;
                else myfail(EXIT_CONFIG_ERROR,"invalid segmentation method '%s'",optarg);
                break;
            case 'c':
                config->max_mt_chunk = atoi(optarg);
                break;
            case 'A':
                config->save_asr = 1;
                break;
            case 'R':
                config->save_raw_mt_seg = 1;
                break;
            case 'C':
                config->save_clean_mt_seg = 1;
                break;
            case 'o':
                config->out_to_stdout = 1;
                break;
            case -1:  /* end of options */
                break;
            case '?':  /* unknown option */
                help = 1;
                break;
            default:  /* serious error */
                myerror("you are not reading this.");
                abort();
        }
        
    }
    
    if(help) {
        printf(usage,argv[0]);
        exit(EXIT_SUCCESS);
    }
    
    /* input files names */
    int i = optind;
    config->inputfilenames = (optind<argc)?malloc((argc-optind)*sizeof(char*)):NULL;
    while(i<argc) {
        config->inputfilenames[i-optind] = argv[i];
        i += 1;
    }
    config->n_input = (optind<argc)?(argc-optind):1; /* if no file provided, input is mic/stdin */
    
    
    return;
    
}

int mp_check_cl(mp_config_t * cfg) {
    
    int status = 0;
    
    if(cfg->asr_segment_length<1) {
        myerror("Invalid ASR segment length '%i'",cfg->asr_segment_length);
        status = EXIT_CONFIG_ERROR;
    }
    if(cfg->asr_segment_overlap<0 || cfg->asr_segment_overlap>=cfg->asr_segment_length) {
        myerror("Invalid ASR segment overlap '%i'",cfg->asr_segment_overlap);
        status = EXIT_CONFIG_ERROR;
    }
    if(cfg->asr_buffer_size<1) {
        myerror("Invalid ASR buffer size '%i'",cfg->asr_buffer_size);
        status = EXIT_CONFIG_ERROR;
    }
    if(cfg->max_mt_chunk<=0) {
        myerror("Invalid maximum mt chunk size '%i'",cfg->max_mt_chunk);
        status = EXIT_CONFIG_ERROR;
    }
    
    return status;
}
