/*
Copyright (c) 2011, Sylvain Raybaud <sylvain.raybaud@gmail.com>
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
Neither the name of the PAROLE team, LORIA, INRIA nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef _engine_h_
#define _engine_h_

#include <stdio.h>
#include <pthread.h>

#include <pocketsphinx.h>

#include "maparole.h"

typedef struct asr_stream_s {
    char * word;
    int start_frm; /* start frame */
    int end_frm; /* end frame */
    int non_word;
    struct asr_stream_s * next;
    struct asr_stream_s * prev;
} asr_stream_t;

typedef struct manager_data_s {
    
    mp_config_t * mp_config;
    mp_engine_t * mp_engine;
    ngram_model_t * lm;
    int32 sentence_start_id;
    int32 unk_id;
    int32 lm_zero;
    
    int input_depleted;
    
    int samples_per_frame;
    int overlap_frames_offset;
    
    /* double buffering: fill buffer2, read buffer1 */
    int16 * asr_buffer;
    int16 * asr_buffer2;
    size_t asr_buffer_segment_ptr;
    size_t asr_buffer_size;
    size_t asr_buffer2_size; /* normaly same as buffer1, except toward the end of input */
    pthread_mutex_t asr_buffer_mutex;
    pthread_cond_t asr_buffer2_empty;
    pthread_cond_t asr_buffer2_refill_done;
    pthread_cond_t asr_new_utterance_cond;
    int asr_buffer2_refill_in_progress;
    
    int16 * asr_segment;
    size_t asr_segment_size;
    size_t asr_segment_overlap;
    char * asr_hyp;
    
    asr_stream_t * asr_stream;
    pthread_mutex_t asr_stream_mutex;
    
    pthread_mutex_t lm_mutex;
    
    myFile in;
    myFile out;
    myFile save_asr_file;
    myFile save_clean_mt_seg_file;
    myFile save_raw_mt_seg_file;
    
    /* cut on '.', ';', ':', '?', '!', '</s>' */
    int32 * mt_cut_symbols_id;
    
} manager_data_t;


void *asr_buffer_manager(void*);
void *asr_manager(void*);
void *translation_manager(void*);

manager_data_t manager_data_init(mp_config_t *, mp_engine_t*, const char *);
void manager_data_free(manager_data_t * pool);
asr_stream_t * asr_stream_new_item(char * word, asr_stream_t * prev, asr_stream_t * next, int start, int end, int non_word);
/* fill asr buffer2 */
/* *** CAUTION *** */
/* this function assumes that pool->asr_buffer_mutex is locked! */
size_t asr_buffer2_fill(manager_data_t * pool);

size_t get_asr_segment(manager_data_t*);
asr_stream_t * get_asr_stream_from_utterance(ps_seg_t*);
asr_stream_t * asr_stream_get_end(asr_stream_t * stream);
asr_stream_t * asr_stream_rewind(asr_stream_t * stream);
void asr_stream_free(asr_stream_t * stream);
asr_stream_t * asr_stream_free_n(asr_stream_t * stream, int n);
void asr_stream_free_backward(asr_stream_t * stream);

/* preprocess string for mt: */
/* tokenize and stuff. */
char * clean_for_mt(char * sentence);

/* get a segment of input audio for ASR */
/* store it in the pool and return number of samples read */
size_t get_asr_segment(manager_data_t * pool);

/* get best hypothesis as a linked list of items */
asr_stream_t * get_asr_stream_from_utterance(ps_seg_t * seg);

/* merge best hypothesis into the stream of recognised items */
void merge_utt_into_stream(manager_data_t * pool, asr_stream_t * utt);

char * stream_to_string(asr_stream_t * from, int length);
char * stream_to_string_raw(asr_stream_t * from, int length);

char * remove_variant(const char * word);
int is_non_word(const char * word);
int is_silence(asr_stream_t * item);

/* get first word item in the stream (forward) */
asr_stream_t * asr_stream_get_first_word(asr_stream_t *);
/* get stream length */
int asr_stream_length(asr_stream_t * stream);

/* get mt friendly segment from asr stream */
/* block until you get it */
char * get_mt_segment_from_stream(manager_data_t * pool);
int cut_on_wid(manager_data_t * pool, int32 wid);

/* cut when 3-gram LM produces </s> */
int get_mt_segment_emit(manager_data_t * pool);
/* cut when 3-gram LM says </s> more likely that actual next item */
int get_mt_segment_prefer(manager_data_t * pool);
/* cut when 3-gram LM says </s> most likely */
int get_mt_segment_at_best(manager_data_t * pool);
/* if tehre exists a position within range where EoS more likely than actual item, cut at the most likely of these positions */
int get_mt_segment_best_prefer(manager_data_t * pool);
/* acoustic: cut on silences */
int get_mt_segment_silence(manager_data_t * pool);
/* dummy method: cut after n words */
int get_mt_segment_length(manager_data_t * pool);
/* cut according to a mix of all previous criteria */
int get_mt_segment_mix(manager_data_t * pool);

char * gen_filename(const char * base, const char * prefix, char * suffix, char * timestamp);

#endif /* _engine_h_ */
