/*
Copyright (c) 2011, Sylvain Raybaud <sylvain.raybaud@gmail.com>
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
Neither the name of the PAROLE team, LORIA, INRIA nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <pthread.h>
#include <sys/time.h>

#include <ngram_model.h>

#include "engine.h"
#include "maparole.h"
#include "common.h"

/* acoustic based method */
/* cut on ++silence++ */
int get_mt_segment_silence(manager_data_t * pool) {
//     debug("starting");

    if(pool->asr_stream==NULL)
        return 0;
    
    asr_stream_t * stream = pool->asr_stream;
    int i = 1;
    while((pool->mp_config->max_mt_chunk<=0 || i<=pool->mp_config->max_mt_chunk) && stream!=NULL) {
        //debug("loop");
        if(is_silence(stream)) {
            return i;
        }
        stream = stream->next;
        i += 1;
    }
    
    return 0;
}

/* dummy method: cut after n words max */
int get_mt_segment_length(manager_data_t * pool) {
//     debug("starting");

    if(pool->asr_stream==NULL)
        return 0;
    
    if(pool->mp_config->max_mt_chunk<=0)
        return 0;
    
    int i = 1;
    asr_stream_t * stream = pool->asr_stream;
    while(i<pool->mp_config->max_mt_chunk && stream!=NULL) {
        //debug("loop");
        stream = stream->next;
        i++;
    }
    
    if(i==pool->mp_config->max_mt_chunk) {
        return i;
    }
    else {
        return 0;
    }
}

/* pure N-gram method */
/* cut when token </s> (or any cutable item) has the highest probability in 4-gram LM*/
/* return last word (not </s>) */
int get_mt_segment_emit(manager_data_t * pool) {
    debug("starting");
    
    if(pool->asr_stream==NULL)
        return 0;
    
    int len = 1;
    /* reverse context: trigram[0]=w_-1, trigram[1]=w_-2, trigram[2]=w_-3 */
    int32 bigram[3] = {0,0};
    bigram[0] = pool->sentence_start_id;
    
    asr_stream_t * stream = asr_stream_get_first_word(pool->asr_stream);
    while((pool->mp_config->max_mt_chunk<=0 || len<=pool->mp_config->max_mt_chunk) && stream!=NULL) {
        //debug("loop");
        
        bigram[1] = bigram[0];
        bigram[0] = ngram_wid(pool->lm,stream->word);
        ngram_iter_t * bigram_iter = ngram_ng_iter(pool->lm,bigram[0],&(bigram[1]),(bigram[1]==0)?0:1);
        
        int32 best_score = pool->lm_zero;
        int32 eos_score = pool->lm_zero;
        ngram_iter_t * trigram_iter = (bigram_iter!=NULL)?ngram_iter_successors(bigram_iter):NULL;
        
        while(trigram_iter!=NULL) {
            //debug("loop");
            int32 score = 0;
            int32 backoff = 0;
            const int32 * wids = ngram_iter_get(trigram_iter,&score,&backoff);
            
            if(score>eos_score && cut_on_wid(pool,wids[0])) {
                eos_score = score;
            }
            if(score>best_score) {
                best_score = score;
            }
            trigram_iter = ngram_iter_next(trigram_iter);
        }
        if(eos_score>=best_score) {
            return len;
        }
        stream = asr_stream_get_first_word(stream->next);
        len += 1;
    }
    
    return 0;
}

/* pure N-gram method */
/* cut when token </s> (or any cutable item) has a higher probability in 4-gram LM than actual following word */
/* return last word (not </s>) */
/* it tends to make short segments (2 to 6 words, avg=2.5) */
int get_mt_segment_prefer(manager_data_t * pool) {
    debug("starting");
    
    if(pool->asr_stream==NULL)
        return 0;
    
    int len = 1;
    asr_stream_t * stream = asr_stream_get_first_word(pool->asr_stream);
    
    /* reverse context: trigram[0]=w_i, trigram[1]=w_i-1, trigram[2]=w_i-2 */
    int32 trigram[3] = {0,0,0};
    trigram[1] = pool->sentence_start_id;
    if(stream!=NULL) trigram[0] = ngram_wid(pool->lm,stream->word);
    
    while((pool->mp_config->max_mt_chunk<=0 || len<=pool->mp_config->max_mt_chunk) && stream!=NULL) {
        //debug("loop");
        
        asr_stream_t * next_word = asr_stream_get_first_word(stream->next);
        if(next_word==NULL) {
            stream = NULL;
            break;
        }
        
        trigram[2] = trigram[1];
        trigram[1] = trigram[0];
        trigram[0] = ngram_wid(pool->lm,next_word->word);
        
        /* get score of actual item */
	pthread_mutex_lock(&(pool->lm_mutex));
        int32 score = pool->lm_zero;
        int32 n_used;
        if(trigram[1]==0)
            score = ngram_ng_score(pool->lm,trigram[0],NULL,0,&n_used);
        else if(trigram[2]==0)
            score = ngram_ng_score(pool->lm,trigram[0],&(trigram[1]),1,&n_used);
        else
            score = ngram_ng_score(pool->lm,trigram[0],&(trigram[1]),2,&n_used);
        
        /* get score of breaker */
        int32 eos_score = pool->lm_zero;
        int i = 0;
        while(pool->mt_cut_symbols_id[i]!=pool->unk_id) {
            //debug("loop");
            if(trigram[1]==0)
                eos_score = max(eos_score,ngram_ng_score(pool->lm,pool->mt_cut_symbols_id[i],NULL,0,&n_used));
            else if(trigram[2]==0)
                eos_score = max(eos_score,ngram_ng_score(pool->lm,pool->mt_cut_symbols_id[i],&(trigram[1]),1,&n_used));
            else
                eos_score = max(eos_score,ngram_ng_score(pool->lm,pool->mt_cut_symbols_id[i],&(trigram[1]),2,&n_used));
            i++;
        }
	pthread_mutex_unlock(&(pool->lm_mutex));
        
        /* decide */
        if(eos_score>=score) {
            return len;
        }
        stream = next_word;
        len += 1;
    }
    
    if(len>=pool->mp_config->max_mt_chunk && pool->mp_config->max_mt_chunk>0) {
        return len;
    }
    
    return 0;
}

/* pure N-gram method */
/* if there exist a position within the window where */
/* token </s> (or any cutable item) has a higher probability in 3-gram LM than actual following word */
/* then cut after such a position where the probability of </s> is the highest */
int get_mt_segment_best_prefer(manager_data_t * pool) {
//     debug("starting");
    
    if(pool->asr_stream==NULL)
        return 0;
    
    int len = 1;
    asr_stream_t * stream = asr_stream_get_first_word(pool->asr_stream);
    
    /* reverse context: trigram[0]=w_i, trigram[1]=w_i-1, trigram[2]=w_i-2 */
    int32 trigram[3] = {0,0,0};
    trigram[1] = pool->sentence_start_id;
    if(stream!=NULL) trigram[0] = ngram_wid(pool->lm,stream->word);
    
    int best_cut = 0;
    int32 best_cut_score = pool->lm_zero;
    
    while((pool->mp_config->max_mt_chunk<=0 || len<=pool->mp_config->max_mt_chunk) && stream!=NULL) {
        //debug("loop");
        
        asr_stream_t * next_word = asr_stream_get_first_word(stream->next);
        if(next_word==NULL) {
            break;
        }
        
        trigram[2] = trigram[1];
        trigram[1] = trigram[0];
        trigram[0] = ngram_wid(pool->lm,next_word->word);
        
        /* get score of actual item */
        pthread_mutex_lock(&(pool->lm_mutex));
        int32 score = pool->lm_zero;
        int32 n_used;
        if(trigram[1]==0)
            score = ngram_ng_score(pool->lm,trigram[0],NULL,0,&n_used);
        else if(trigram[2]==0)
            score = ngram_ng_score(pool->lm,trigram[0],&(trigram[1]),1,&n_used);
        else
            score = ngram_ng_score(pool->lm,trigram[0],&(trigram[1]),2,&n_used);

        /* get score of breaker */
        int32 eos_score = pool->lm_zero;
        int i = 0;
        while(pool->mt_cut_symbols_id[i]!=pool->unk_id) {
            //debug("loop");
            if(trigram[1]==0)
                eos_score = max(eos_score,ngram_ng_score(pool->lm,pool->mt_cut_symbols_id[i],NULL,0,&n_used));
            else if(trigram[2]==0)
                eos_score = max(eos_score,ngram_ng_score(pool->lm,pool->mt_cut_symbols_id[i],&(trigram[1]),1,&n_used));
            else
                eos_score = max(eos_score,ngram_ng_score(pool->lm,pool->mt_cut_symbols_id[i],&(trigram[1]),2,&n_used));
            i++;
        }
        pthread_mutex_unlock(&(pool->lm_mutex));
        
        /* decide */
        if(eos_score>=score && eos_score>=best_cut_score) {
            best_cut = len;
            best_cut_score = eos_score;
        }
        stream = next_word;
        len += 1;
    }
    if(len>=pool->mp_config->max_mt_chunk && best_cut!=0) {
        return best_cut;
    }
    
    return 0;
}

int get_mt_segment_at_best(manager_data_t * pool) {
//     debug("starting");
    
    if(pool->asr_stream==NULL)
        return 0;
    
    int len = 1;
    asr_stream_t * stream = asr_stream_get_first_word(pool->asr_stream);
    int best_cut = 0;
    int32 best_cut_score = pool->lm_zero;
    
    /* reverse context: bigram[0]=w_i, bigram[1]=w_i-1*/
    int32 bigram[2] = {0,0};
    bigram[0] = pool->sentence_start_id;
    
    while((pool->mp_config->max_mt_chunk<=0 || len<=pool->mp_config->max_mt_chunk) && stream!=NULL) {
        //debug("loop");
        
        bigram[1] = bigram[0];
        bigram[0] = ngram_wid(pool->lm,stream->word);
        
        /* get score of breaker */
	pthread_mutex_lock(&(pool->lm_mutex));
        int n_used = 0;
        int32 score = pool->lm_zero;
        int i = 0;
        while(pool->mt_cut_symbols_id[i]!=pool->unk_id) {
            //debug("loop");
            if(bigram[0]==0)
                score = score,ngram_ng_score(pool->lm,pool->mt_cut_symbols_id[i],NULL,0,&n_used);
            else if(bigram[1]==0)
                score = score,ngram_ng_score(pool->lm,pool->mt_cut_symbols_id[i],bigram,1,&n_used);
            else
                score = score,ngram_ng_score(pool->lm,pool->mt_cut_symbols_id[i],bigram,2,&n_used);
            if(score>=best_cut_score) {
                best_cut_score = score;
                best_cut = len;
            }
            i++;
        }
	pthread_mutex_unlock(&(pool->lm_mutex));
        
        stream = asr_stream_get_first_word(stream->next);
        len += 1;
    }
    
    if(len>=pool->mp_config->max_mt_chunk && pool->mp_config->max_mt_chunk>0 && best_cut!=0) {
        return best_cut;
    }
    
    return 0;
}

/* mix acoustic and N-gram method, with maximum chunk length */
/* cut if cutable symbol more likely than following word */
/* if none found, cut on first silence */
/* if that fails, cut where cutable symbol is most likely in the range [0,max_chunk] */
/* return last word (not </s>) */
int get_mt_segment_mix(manager_data_t * pool) {
//     debug("starting");
    
    if(pool->asr_stream==NULL)
        return 0;
    
    int chunk = get_mt_segment_best_prefer(pool);
    if(chunk>0 || asr_stream_length(pool->asr_stream)<pool->mp_config->max_mt_chunk) return chunk; /* don't feel compelled to cut if max length not reached */
    
    chunk = get_mt_segment_silence(pool);
    if(chunk>0) return chunk;
    
    chunk = get_mt_segment_at_best(pool);
    if(chunk>0) return chunk;
    
    return 0;
}

/* get mt friendly segment from asr stream */
/* block until you get it */
char * get_mt_segment_from_stream(manager_data_t * pool) {
//     debug("start");
    
    char * seg = NULL;
    while(seg==NULL) {
//         //debug("loop");
        /* loop until you get the goddamn segment */
        pthread_mutex_lock(&(pool->asr_stream_mutex));
        
        int chunk_length = 0;
        
        /* cut when 3-gram LM produces </s> */
        if      (pool->mp_config->seg_method==SEG_EMIT) chunk_length = get_mt_segment_emit(pool);
        
        /* cut when 3-gram LM says </s> more likely that actual next item */
        else if (pool->mp_config->seg_method==SEG_PREFER) chunk_length = get_mt_segment_prefer(pool);
        
        /* cut when 3-gram LM says </s> most likely */
        else if (pool->mp_config->seg_method==SEG_BEST) chunk_length = get_mt_segment_at_best(pool);
        
        /* if tehre exists a position within range where EoS more likely than actual item, cut at the most likely of these positions */
        else if (pool->mp_config->seg_method==SEG_BEST_PREFER) chunk_length = get_mt_segment_best_prefer(pool);
        
        /* acoustic: cut on silences */
        else if (pool->mp_config->seg_method==SEG_SILENCE) chunk_length = get_mt_segment_silence(pool);
        
        /* dummy method: cut after n words */
        else if (pool->mp_config->seg_method==SEG_LENGTH) chunk_length = get_mt_segment_length(pool);
        
        /* cut according to a mix of all previous criteria */
        else if (pool->mp_config->seg_method==SEG_MIX) chunk_length = get_mt_segment_mix(pool);
        
        if(chunk_length!=0) {
            myverbose(pool->mp_config->verbose,"found mt segment.");
            seg = stream_to_string(pool->asr_stream,chunk_length);
            pool->asr_stream = asr_stream_free_n(pool->asr_stream,chunk_length);
            if(pool->asr_stream!=NULL) {
                pool->asr_stream->prev = NULL;
            }
        }
        
        if(chunk_length==0 && !pool->input_depleted) {
            struct timeval now;
            struct timespec timeout;
            gettimeofday(&now,NULL);
            timeout.tv_sec = now.tv_sec+1;
            timeout.tv_nsec = now.tv_usec*1000;
//             debug("waiting for signal from asr_manager...");
            pthread_cond_timedwait(&(pool->asr_new_utterance_cond),&(pool->asr_stream_mutex),&timeout);
//             debug("asr manager signaled a new utterance, or timedwait timed out.");
        }
        else if(chunk_length==0 && pool->input_depleted) {
            myverbose(pool->mp_config->verbose,"flushing whatever remains of asr_stream to mt.");
            seg = stream_to_string(pool->asr_stream,-1);
            asr_stream_free(pool->asr_stream);
            pool->asr_stream = NULL;
            pthread_mutex_unlock(&(pool->asr_stream_mutex));
            return seg;
        }
        
        pthread_mutex_unlock(&(pool->asr_stream_mutex));
        
//         if(chunk_length==0) {
//             usleep(DEFAULT_WAIT); /* stupid: if you got no segment, wait until asr produces a new utterance */
//         }
    }
    
    return seg;
    
}


int cut_on_wid(manager_data_t * pool, int32 wid) {
    int i = 0;
    while(pool->mt_cut_symbols_id[i]!=pool->unk_id) {
        //debug("loop");
        if(wid==pool->mt_cut_symbols_id[i]) {
            return TRUE;
        }
        i++;
    }
    return FALSE;
}
