#ifndef CM_H
#define CM_H

#include "language_model.h"

/** \file
   Module where confidence measures are implemented.
   It provides
   - a feature_set_s structure to store a list of features to compute,
   and creation/destruction functions.
   - a cm_resources_s structure to hold the resources needed,
   and creation/destruction functions.
   - a compute_features() function which computes features for a translation_pair from those structures.

   Features are regrouped in 4 categories (word/sentence level) X (asr/mt).\n
   For each category, we use an array of all available features,
   and the results for a given translation pair are stored in arrays
   within the translation_pair_t.

   <b>Usage</b>
   - feature_set_new() or feature_set_new_by_name() is called once to get a list of
   either all or some of the available features, stored in the pool.
   - the created set is passed to each compute_features() call to compute this set of features
   for a given translation_pair. (see compute_features() for details)\n
   Features and confidence measures are then available in the translation_pair_t structure.

   <b>Implementation details</b>

   A feature (word_level_feature_s or sent_level_feature_s) is given by
   - a pointer to the computation function (same prototype for all)
   - its name in a plain char*
   - a boolean which tells whether it is enabled or not
   - the index of the feature in the results array
   (i.e. it is a numbering of the enabled features;
   therefore the result of feature i of tab is stored at index tab[i].idx of the
   corresponding results array (see the end of the example below for a better understanding)

   The enabled/index thing allows to use only a subset of the features, while still being able
   to retrieve the result of a feature X for another feature Y efficiently.
   \n\n

   For each of the 4 categories, there's a static array and an enum in confidence_measure.c, as well as an array in translation_pair_s.
   - The static array contains for each feature: a pointer to the function, and the name of the feature\n
   - The enum contains for each feature an entry with the same name (in caps), which value is
   the index of the feature in the array.
   - The array in translation_pair is where the values of the features are stored.


   <b>Example & adding a feature</b>

   For word level features (wlf) of translated (mt) features, the static array could be: \n
   <pre>
   static word_level_feature_t mt_wlf[] = {
     {feature_1, "feature_1"},
     {feature_2, "feature_2"}
   };</pre>
   And the enum must then be:
   <pre>  enum mt_wordLevelFeature {FEATURE_1, FEATURE_2};</pre>
   Also, in translation_pair_s, there is a field:
   <pre>  float ** mt_wordLevelFeatures;</pre>


   To add a new feature <i>foo</i>, you have to:
   - write the function <i>foo</i> with prototype <i>wlff</i>
   (word level feature function, see the typedef section)
   - BOTH add <i>{foo, "foo"}</i> to the static array <i>mt_wlf</i>
   and <i>FOO</i> to the enum <i>mt_wordLevelFeature</i>
   (in confidence_measure.c) at the SAME position.

   <pre>
  static word_level_feature_t mt_wlf[] = {
    {feature_1, "feature_1"},
    {feature_2, "feature_2"}
    {foo, "foo"}
  };</pre>
   <pre>  enum mt_wordLevelFeature {FEATURE_1, FEATURE_2, FOO};</pre>

   where <i>"foo"</i> is the name you give to feature_set_new_by_name() to enable <i>foo</i>,\n
   and FOO is the value you use in the code to access the feature in <i>feature_set_s.mt_wlf</i>.

   Thus, You can
   - use the name "foo" to enable foo with feature_set_new_by_name(). e.g.\n
   <pre>
  char ** t = {"feature_1", "foo"};
  feature_set_t * set = feature_set_new_by_name(t);</pre>
   - use this set to tell compute_features() which features to compute:\n
   <pre>  compute_features(pair, set);</pre>
   - access the computed value in your code (e.g. in another feature) at\n
   <pre>  pair->mt_wordLevelFeature[i_word][set->mt_wlf[FOO].idx]</pre>
   (note the indirection <i>mt_wlf[FOO]<b>.idx</b></i>: in this example <i>FOO</i> is 2
   (the index of foo in feature_set_s.mt_wlf)
   and <i>mt_wlf[FOO].idx</i> is 1
   (the index of foo in the results array: translation_pair_s.asr_wordLevelFeatures))

   <b>Neural networks</b>

   One neural network is used for each category of features.\n
   The input of a network is the array of the features for a word or a sentence,
   and the output is the confidence measure for this element.\n
   Care must be taken to use networks that have been trained with features produced
   using the same feature_set_s that is used by compute_features().\n
   It is obvious why the same list of features must be used for training and executing the networks,
   we just want to emphasise that neural networks allow no mismatch checking whatsoever,
   so beware of your scripts, or fear the apocalypse.
   (or worse, irrelevant values without a word of warning from s2tt)

*/

struct fann;
struct cm_config_s;
struct translation_pair_s;

/** Holds references to all the resources
   for confidence measures (language models, neural networks) */
typedef struct cm_resources_s {
  /* language model instances */
  lmi_t asr_lm;			/**< source language model */
  lmi_t asr_backward_lm;	/**< backward source language model */
  lmi_t mt_lm;			/**< target language model */
  lmi_t mt_backward_lm;		/**< backward target language model */

  /* neural network instances */
  struct fann *asr_word_level_ann; /**< neural network */
  struct fann *asr_sent_level_ann; /**< neural network */
  struct fann *mt_word_level_ann;  /**< neural network */
  struct fann *mt_sent_level_ann;  /**< neural network */
} cm_resources_t;

/** Create a new resources structure for confidence measures from configuration.
    It loads the language models and neural networks given in the configuration.
    \param config the configuration (filenames, etc)
    \param lm the source language model: it is already open and stored in the pool,
    and must be passed as an argument. (Unlike Moses' existing LM which can be retrieve with moses_get_lm())
    \return a new cm_resources_s, to be freed with cm_resources_free()*/
cm_resources_t *cm_resources_new(struct cm_config_s *config, sphinx_lm_t lm);
/** Free all memory allocated by cm_resources_new. */
void cm_resources_free(cm_resources_t *res);


struct translation_pair_s;
struct feature_set_s;

/** Type of word level feature function (wlff) */
typedef float (*wlff)(cm_resources_t *res, struct feature_set_s *set, struct translation_pair_s *info, int i);
/** Type of sentence level feature function (slff) */
typedef float (*slff)(cm_resources_t *res, struct feature_set_s *set, struct translation_pair_s *info);

/** Represents a word level feature */
typedef struct word_level_feature_s {
  wlff fun;		 /**< pointer to the computation function */
  char * name;		 /**< name of the feature, for easier dynamic access */
  int enabled;		 /**< true iff this feature must be computed */
  int idx;		 /**< index of the feature in the list of enabled features */
} word_level_feature_t;
/** Represents a sentence level feature */
typedef struct sent_level_feature_s {
  slff fun;		 /**< pointer to the computation function */
  char * name;		 /**< name of the feature, for easier dynamic access */
  int enabled;		 /**< true iff this feature must be computed */
  int idx;		 /**< index of the feature in the list of enabled features */
} sent_level_feature_t;

/** Represents a set of available features */
typedef struct feature_set_s {
  word_level_feature_t * mt_wlf;  /**< array of word level features for translation */
  int mt_wlf_count;		  /**< number of enabled features */
  int mt_wlf_total;		  /**< number of available features (size of the array) */
  sent_level_feature_t * mt_slf;  /**< array of sentence level features for translation */
  int mt_slf_count;		  /**< number of enabled features */
  int mt_slf_total;		  /**< number of available features (size of the array) */
  word_level_feature_t * asr_wlf; /**< array of word level features for recognition */
  int asr_wlf_count;		  /**< number of enabled features */
  int asr_wlf_total;		  /**< number of available features (size of the array) */
  sent_level_feature_t * asr_slf; /**< array of sentence level features for recognition */
  int asr_slf_count;		  /**< number of enabled features */
  int asr_slf_total;		  /**< number of available features (size of the array) */
} feature_set_t;

/** Create a new feature functions set, all enabled */
feature_set_t * feature_set_new();
/** Create a new feature functions set, enabling only thoses listed in <i>names</i>.
    The names are hard-coded in confidence_measure.c.\n
    If a name does not exist, a warning is printed and execution continues.
    \param names a NULL terminated array of feature names */
feature_set_t * feature_set_new_by_name(char **names);
/** Free all memory allocated by feature_set_new* */
void feature_set_free(feature_set_t *set);

/** Compute the features for a given translation pair, and the resulting confidence measures.
    Warning: it assumes all necessary fields were set in <i>info</i>
    (i.e. source and target sentences for now).\n
    However, it takes care of allocating the memory in <i>info</i> to store the results.
    (calling translation_pair_alloc_features())

    compute_features() will compute a value for each sentence level feature in <i>set</i>, and a
    value for each word (in both source and target) and for each word level feature in <i>set</i>.\n
    It will then feed thoses values into the 4 neural networks referenced
    in 'res' to get the final confidence measures: one for source, one for target,
    and one for each word in both the source and the target.\n
    Everything is stored in the translation_pair_s <i>info</i>.

    \param set the set of features to compute
    \param res the resources used in computation (language models, etc)
    \param info the translation_pair to compute features for, and where to store the results.
*/
void compute_features(feature_set_t *set, cm_resources_t *res, struct translation_pair_s *info);

/** Print the arrays of features on stdin (utils) */
void print_features(feature_set_t *features, struct translation_pair_s *info);

#endif
