#include <cutils.h>
#include "common.h"
#include "language_model.h"

#include <string>
#include <vector>

// Moses headers
#include <FFState.h>
#include <StaticData.h>
#include <TranslationSystem.h>
#include <ScoreIndexManager.h>
#include <LM/Factory.h>
#include <LM/Ken.h>
#include <lm/model.hh>
#include <lm/max_order.hh>
#include <lm/return.hh>

#define min(a,b) ((a)<(b))?(a):(b)
#define max(a,b) ((a)>(b))?(a):(b)

/** \ file
 * Functions for querying Moses' implementations of language models.
 */

using namespace Moses;

/** Wrapper for Moses' Word allowing the manipulation of wids.

    Internally, a wid in Moses is a Factor*, but language models functions use Word instances,
    where a word only contains an array of Factor*.

    In single factor models, the wid is the only Factor* in the array, we use this class to retrieve the wid of a Word, or to rebuild a Word from a wid.
*/
class MyWord : public Moses::Word {
public:
  MyWord() {}
  /** Make a Word from a wid */
  MyWord(Factor *wid) {
    m_factorArray[0] = wid;
  }
  /** Get the wid of the word */
  const Factor * getWid() {
    return m_factorArray[0];
  }
};

moses_lm_t moses_get_lm() {
  // Just digging through the Moses layers
  TranslationSystem ts = StaticData::Instance().GetTranslationSystem(std::string("default"));
  LMList list = ts.GetLanguageModels();
  LanguageModel *lm = *(list.begin());
  return lm;
}

moses_lm_t moses_lm_open(char *fileName) {
  std::vector<FactorType> factorTypes(1);
  factorTypes[0] = 0;
  std::string file(fileName);
  try {
    LanguageModel *lm =
      LanguageModelFactory::
      CreateLanguageModel(Ken,
			  factorTypes,
			  3,
			  file,
			  (ScoreIndexManager&) StaticData::Instance().GetScoreIndexManager(),
			  0 /*unused for ken LM*/);
    return lm;
  }
  catch (std::exception& e) {
    myfail(EXIT_MOSES_ERROR, "%s", e.what());
  }
}


float moses_score_unigram(lmi_data_t data, wid_t a) {

    const lm::base::Model &model = *((LanguageModel*)data)->GetKen();
    lm::ngram::State outState;

    float score = model.Score(model.NullContextMemory(), a.m_wid, &outState);

    return pow(score,10.0);
}

float moses_score_bigram(lmi_data_t data, wid_t a, wid_t b) {

    const lm::base::Model &model = *((LanguageModel*)data)->GetKen();
    lm::ngram::State inState; inState.length = 1; inState.words[0] = a.m_wid;
    lm::ngram::State outState;

    float score = model.Score(&inState, b.m_wid, &outState);

    return pow(score,10.0);
}

float moses_score_trigram(lmi_data_t data, wid_t a, wid_t b, wid_t c) {

    const lm::base::Model &model = *((LanguageModel*)data)->GetKen();
    lm::ngram::State inState; inState.length = 2; inState.words[0] = b.m_wid; inState.words[1] = a.m_wid;
    lm::ngram::State outState;

    float score = model.Score(&inState, c.m_wid, &outState);

    return exp(score);
}

float moses_score_ngram(lmi_data_t data, int n, wid_t *t) {

    const lm::base::Model &model = *((LanguageModel*)data)->GetKen();
    lm::ngram::State inState; inState.length = min(n-1,lm::ngram::kMaxOrder-1);
    lm::ngram::State outState;

    int i;
    for(i=0; i<(min(n,lm::ngram::kMaxOrder)-1); i++) {
        inState.words[i] = t[n-2-i].m_wid;
    }

    float score = model.Score(&inState, t[n-1].m_wid, &outState);

    return pow(score,10.0);
}

int moses_unigram_exists(lmi_data_t data, wid_t a) {

    const lm::base::Model &model = *((LanguageModel*)data)->GetKen();
    lm::ngram::State outState;

    struct lm::FullScoreReturn ret = model.FullScore(model.NullContextMemory(), a.m_wid, &outState);

    return ret.ngram_length;
}

int moses_bigram_exists(lmi_data_t data, wid_t a, wid_t b) {

    const lm::base::Model &model = *((LanguageModel*)data)->GetKen();
    lm::ngram::State inState; inState.length = 1; inState.words[0] = a.m_wid;
    lm::ngram::State outState;

    struct lm::FullScoreReturn ret = model.FullScore(&inState, b.m_wid, &outState);

    return ret.ngram_length==2;
}

int moses_trigram_exists(lmi_data_t data, wid_t a, wid_t b, wid_t c) {
    const lm::base::Model &model = *((LanguageModel*)data)->GetKen();
    lm::ngram::State inState; inState.length = 2; inState.words[0] = b.m_wid; inState.words[1] = a.m_wid;
    lm::ngram::State outState;

    struct lm::FullScoreReturn ret = model.FullScore(&inState, c.m_wid, &outState);

    return ret.ngram_length==3;
}

wid_t moses_wid(lmi_data_t data, char * word) {
    const lm::base::Model &model = *((LanguageModel*)data)->GetKen();
    const lm::base::Vocabulary &voc = model.BaseVocabulary();
    std::string w(word);
    wid_t wid;
    wid.m_wid = voc.Index(w);
    return wid;
}

void moses_free_data(lmi_data_t data) {
  LanguageModel *_lm = (LanguageModel*) data;
  delete _lm;
}

lmi_t moses_lmi_new(moses_lm_t *data) {
  lmi_t moses = {
    moses_score_unigram,
    moses_score_bigram,
    moses_score_trigram,
    moses_score_ngram,
    moses_unigram_exists,
    moses_bigram_exists,
    moses_trigram_exists,
    moses_wid,
    moses_free_data,
    (lmi_data_t*) data,
    moses_wid(data, (char*)"<s>"),
    moses_wid(data, (char*)"</s>")
  };
  return moses;
}

