/*
Copyright (c) 2011, Sylvain Raybaud <sylvain.raybaud@gmail.com>
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
Neither the name of the PAROLE team, LORIA, INRIA nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef _engine_h_
#define _engine_h_

#include <stdio.h>
#include <pthread.h>
#include <glib.h>
#include <semaphore.h>
#include <libxml/tree.h>

#include <pocketsphinx.h>
#include <src/libpocketsphinx/dict.h>

#include "maparole.h"

/** \file
    The s2tt engine.

    ... 

    <b>Parallel decoders</b>

    s2tt can instanciate multiple Sphinx decoders.\n
    One decoder is created for each acoustic model specified by a
    "--acoustic-model hmm[:dict[:fdict[:mdef]]]" option.\n
    They are run in parallel in threads managed by the asr_manager() thread, and the
    best decoder is determined by a call to find_best_decoder().\n
    see asr_decoder() for details

    ...

    <b>XML output</b>

    The final output of s2tt is written to a XML file using libxml2 (from Gnome).
    The output looks like this:
<pre>
  <s2tt_result>
  <sentence cm="0.515427">
    <text> foo bar baz </text>
    <word cm="0.515427">foo</word>
    <word cm="0.515427">bar</word>
    <word cm="0.515427">baz</word>
  </sentence>
  <sentence cm="0.515427">
    ...
  </sentence>
</s2tt_result>
</pre>

    The necessary resources are kept in the pool.\n
    A sentence from a translation_pair 't' is added to the s2tt_result node with add_node_to_xml(),\n
    and the whole tree is actually written to the disc with save_xml().

 */

struct feature_set_s;
struct cm_resources_s;
union wid_s;
struct sentence_s;

#ifdef __cplusplus
extern "C" {
#endif

/** Stream of words: implemented as a doubly linked link.
    Produced by asr_manager and fed into asr_stream_manager. */
typedef struct asr_stream_s {
    char * word;        /**< the recognised word/filler/noise */
    s3wid_t wid;        /**< corresponding wid in sphinx' dictionary */
    int n_phonemes;
    const char ** phonemes;   /**< the corresponding phonemes, as recognized by the decoder */
    float32 t_start;	/**< start time */
    float32 t_end;	/**< end time */
    float32 cm;		/**< confidence measure. not yet implemented */
    int non_word;	/**< true if the element is not a word (if it is a filler/noise) */
    struct asr_stream_s * next;	/**< next element in the list, or NULL if none*/
    struct asr_stream_s * prev;	/**< previous element in the list, or NULL if none */
} asr_stream_t;

typedef void asr_stream_lattice_t; /* TODO */


/** Audio chunk. Produced by asr_buffer_manager and fed into asr_manager. */
typedef struct audio_chunk_s {
    int16 * audio;		/**< raw audio data */
    int size;			/**< number of samples */
    float time_start;		/**< time position of first sample */
} audio_chunk_t;


/**
 * Structure representing an information unit
 * for a (source, target) pair of sentences.
 * The features can be allocated and computed by compute_features()
 * when the source and target are set.
 */
typedef struct translation_pair_s {
    struct sentence_s *source;	/**< The recognised sentence if ASR output is a one-best */
    struct sentence_s *target;	/**< The translated sentence */
    ps_lattice_t * src_lattice;	/**< The source lattice for translation if ASR output is a lattice */
    audio_chunk_t * src_audio;	/**< The audio chunk corresponding to the source */
    asr_stream_t * src_stream;	/**< The stream output by ASR for the source sentence above. */
    
    float startTime; 		/**< start time of the audio */
    float endTime;		/**< end time of the audio */

    float translation_score;	/**< Score given by Moses */
    
    /* filled by computeFeatures() */

    /** array storing the sentence level features for translation */
    float * mt_sentLevelFeatures;
    /** internal : memory for 2d array below */
    float * mt_wordLevelFeaturesBuffer;
    /** array storing the word level features for translation;
	wordLevelFeatures[i][j] = feature j for word i */
    float ** mt_wordLevelFeatures;
    /** array storing the sentence level features for recognition */
    float * asr_sentLevelFeatures;
    /** internal : memory for 2d array below */
    float * asr_wordLevelFeaturesBuffer;
    /** array storing the word level features for recognition;
	wordLevelFeatures[i][j] = feature j for word i */
    float ** asr_wordLevelFeatures;

    /** Confidence measure for recognised sentence,
	computed by a neural network, from above features. */
    float asr_sent_cm;
    /** Array of confidence measures: one for each word in source sentence,
	computed by a neural network, from above features. */
    float *asr_word_cm;
    /** Confidence measure for translated sentence,
	computed by a neural network, from above features. */
    float mt_sent_cm;
    /** Array of confidence measures: one for each word in target sentence,
	computed by a neural network, from above features. */
    float *mt_word_cm;
} translation_pair_t;

struct manager_data_s;

/** Parameter for the asr_decoder thread function. */
typedef struct decoder_thread_data_s {
  struct manager_data_s * pool; /**< The pool */
  int thread_number; 		/**< The thread number */
} decoder_thread_data_t;

/** Internal structure holding the necessary data for one decoder when running several decoders
    in parallel on the same audio extract.
    It is used only in decoder_set_t. */
typedef struct decoder_s {
    ps_decoder_t * instance;	/**< the actual Sphinx decoder */
    
    /* Variables relevant to parallel decoding (with 'find_best_decoder') */
    pthread_t thread; /**< the pthread used by find_best_decoder() for this decoder*/
    decoder_thread_data_t thread_data; /**< The argument for the thread function */
    sem_t sem_start;	      /**< semaphore to wake up the decoder */
    sem_t sem_finish; /**< semaphore for the decoder to signal the end of a decoding */
    float score;      /**< where the decoder writes its result */
} decoder_t;

/** Internal structure holding the data for parallel decoders management.
    Only used in manager_data_t */
typedef struct decoders_set_s {
    int time_to_die;	 /**< flag checked by decoder threads; if true, the threads must end */
    int count;		 /**< number of decoders */
    decoder_t *list;	 /**< list of the decoders: array of size 'count' */
    int current;	 /**< index of the decoder to use for ASR, positioned by find_best_decoder() */
} decoders_set_t;

/** Data pool shared by the threads. */
typedef struct manager_data_s {
    
    int input_buffer_manager_is_dead; /**< Synchronisation flag */
    int asr_manager_is_dead;	      /**< Synchronisation flag */
    int asr_stream_manager_is_dead;   /**< Synchronisation flag */
    int segment_manager_is_dead;      /**< Synchronisation flag */
    int translation_manager_is_dead;  /**< Synchronisation flag */
    
    mp_config_t * mp_config;
    mp_engine_t * mp_engine;
    
    ngram_model_t * lm;
    pthread_mutex_t lm_mutex;
    int32 sentence_start_id;
    int32 unk_id;
    int32 lm_zero;

    struct feature_set_s *features; /**< The features to compute for processed data */
    struct cm_resources_s *cm_res;  /**< The resources for confidence measures*/
  
    float32 sampling_rate;	/**< Sampling rate of the raw audio processed by Sphinx. */
    int samples_per_frame;
    int samples_frame_shift;

    size_t asr_segment_size; 	/**< number of samples in a segment */
    size_t asr_segment_overlap;	/**< number of samples in the overlap */
    
    float32 segment_time_shift;
    float32 segment_time_offset;
    
    int n_channel; /* always 1 so far */

    /* "best decoder search" related variables */

    decoders_set_t decoders;	/**< the available decoders */
    audio_chunk_t audio_extract; /**< where parallel decoders read input */
    float penultimate_reco_score; /**< score given by Sphinx for the last but one transcription by asr_manager */
    float last_reco_score; /**< score given by Sphinx for the last transcription by asr_manager */
    int last_utt_contains_noise; /**< True if noise was detected in last utterance */
    int last_utt_contains_silence; /**< True if silence was detected in last utterance */
    
    /** audio buffer: queue of chunks */
    GQueue * input_buffer;
    size_t input_buffer_size;
    pthread_mutex_t input_buffer_mutex;
    pthread_cond_t input_buffer_one_pushed;
    pthread_cond_t input_buffer_one_poped;
    
    /** queue of transcribed chunks --- no max size */
    GQueue * asr_hyp_queue;
    pthread_mutex_t asr_hyp_queue_mutex;
    pthread_cond_t asr_hyp_queue_one_hyp_pushed;
    
    /** stream (doubly linked items list) of merged transcribed chunks --- no max size */
    asr_stream_t * asr_stream;
    asr_stream_lattice_t * asr_stream_lattice;
    pthread_mutex_t asr_stream_mutex;
    pthread_cond_t asr_stream_updated;
    
    /** queue of mt segments --- no max size */
    GQueue * segment_queue;
    pthread_mutex_t segment_queue_mutex;
    pthread_cond_t segment_queue_one_segment_pushed;

    /** the XML node where to append generated translations */
    xmlNodePtr xml_result_node;
    /** the XML document to use when writing to disc */
    xmlDocPtr xml_result_doc;

    char * in_filename;
    myFile in;			/**< Audio input file */
    myFile out;			/**< File used as a log for s2tt results */
    char * save_xml_filename;   /**< Final output XML file of translation */
    char * save_asr_filename;
    char * save_ctm_filename;
    char * save_clean_mt_seg_filename;
    char * save_raw_mt_seg_filename;
    
    /** cut on '.', ';', ':', '?', '!', '&lt;/s>' */
    int32 * mt_cut_symbols_id;
    
} manager_data_t;



/* Thread functions */

/** Thread: runs a given decoder on a small chunk of audio.
    Multiple instances of this threads are used to run decoders in parallel;
    they are triggered by find_best_decoder to determine the best acoustic model for ASR.

    IO: each decoder is run on pool.audio_extract, and
    stores the score it got in pool.decoders.list[i].score.
    
    Those threads are managed within the asr_manager() thread:
    they are created by start_decoder_threads(), killed by terminate_decoder_threads(),
    and they are triggered by find_best_decoder().

    Synchronisation is done using pthread's semaphores:
    each thread waits for input with pool.decoders.list[i].sem_wait,
    and signals it has finished working with endpool.decoders.list[i].sem_post.
*/

void *asr_decoder(void * data);
/** Thread: takes care of replenishing the audio buffer */
void *asr_buffer_manager(void * data);
/** Thread: handles the recognition process */
void *asr_manager(void * data);
/** Thread: merges transcribed audio chunks into stream */
void *asr_stream_manager(void * data);
/** Thread: extracts mt friendly segments from stream and push them to segment queue */
void *segment_manager(void * data);
/** Thread: handles the translation process */
void *translation_manager(void * data);

manager_data_t manager_data_init(mp_config_t *, mp_engine_t*, const char *);
void manager_data_free(manager_data_t * pool);

asr_stream_t * asr_stream_get_end(asr_stream_t * stream);
asr_stream_t * asr_stream_rewind(asr_stream_t * stream);
void asr_stream_free(asr_stream_t * stream);
asr_stream_t * asr_stream_free_n(asr_stream_t * stream, int n);
void asr_stream_free_backward(asr_stream_t * stream);

/** Find the best decoder for the next audio and set it as current.
    It extracts a few seconds of audio from 'next_chunk',
    runs every available decoder in pool.decoders on the sample,
    and sets whichever gives the highest recognition score as the current one for ASR.
    (pool.decoders.current)

    It is called by asr_manager() at an appropriate time.\n
    The decision whether to call or not to call find_best_decoder() is taken depending
    on values stored in the pool; those values are updated at relevant times.
    See the code of asr_manager() for details.
    \param pool the pool
    \param next_chunk pointer to the next audio chunk to be decoded
*/
void find_best_decoder(manager_data_t * pool, audio_chunk_t * next_chunk);

/** get best hypothesis as a linked list of items */
asr_stream_t * get_asr_stream_from_utterance(manager_data_t * pool, ps_seg_t * seg);

/** merge best hypothesis into the stream of recognised items */
void merge_utt_into_stream(manager_data_t * pool, asr_stream_t * utt);

void write_ctm(manager_data_t * pool, int length);

int is_silence(asr_stream_t * item);

/** get first word item in the stream (forward) */
asr_stream_t * asr_stream_get_first_word(asr_stream_t *);
/** get stream length */
int asr_stream_length(asr_stream_t * stream);

/** get mt friendly segment from asr stream */
int get_mt_segment_from_stream(manager_data_t * pool);
/** cut when 3-gram LM produces &lt;/s> */
int get_mt_segment_emit(manager_data_t * pool);
/** cut when 3-gram LM says &lt;/s> more likely that actual next item */
int get_mt_segment_prefer(manager_data_t * pool);
/** cut when 3-gram LM says &lt;/s> most likely */
int get_mt_segment_at_best(manager_data_t * pool);
/** if there exists a position within range where EoS more likely than actual item, cut at the most likely of these positions */
int get_mt_segment_best_prefer(manager_data_t * pool);
/** acoustic: cut on silences */
int get_mt_segment_silence(manager_data_t * pool);
/** dummy method: cut after n words */
int get_mt_segment_length(manager_data_t * pool);
/** cut according to a mix of all previous criteria */
int get_mt_segment_mix(manager_data_t * pool);
int cut_on_wid(manager_data_t * pool, int32 wid);

char * gen_filename(const char * base, const char * prefix, char * suffix, char * timestamp);

audio_chunk_t * audio_chunk_new(int16 * audio, size_t size);
void audio_chunk_free(audio_chunk_t * chunk);

int process_raw(manager_data_t * pool, audio_chunk_t * chunk);

void asr_manager_terminate(manager_data_t * pool);

/** Add a node to the XML tree of s2tt results
    \param pool the pool
    \param info the sentence to add */
xmlNodePtr add_node_to_xml(manager_data_t * pool, translation_pair_t * info);
/** Write the results to the disk.
    It saves pool->xml_result_doc to pool->save_xml_filename.
    \param pool the pool */
void save_xml(manager_data_t * pool);

/** Create a new empty translation_pair.
 * Note that it does not allocate memory for the features arrays, since they depend
 * on the source and the target. Instead, this is done by compute_features() */
translation_pair_t * translation_pair_new();
/** Create a new translation_pair from with a lattice as source */
translation_pair_t * translation_pair_new_from_lattice(ps_lattice_t * source);

/** Getter */
void translation_pair_get_asr_result   (translation_pair_t *, manager_data_t *);
/** Getter */
char * translation_pair_get_src_string (translation_pair_t *);
/** Getter */
asr_stream_t * translation_pair_get_src_stream   (translation_pair_t *);
/** Getter */
ps_lattice_t * translation_pair_get_src_lattice(translation_pair_t *);
/** Getter */
char * translation_pair_get_tgt_string (translation_pair_t *);
/** Generate source phoneme sequence */
char * translation_pair_get_src_phonemes_string (translation_pair_t *);

/** Setter */
void translation_pair_set_audio        (translation_pair_t *, audio_chunk_t *);
/** Set the target-related fields in item */
void translation_pair_set_target(translation_pair_t *item, char *target_str, const char ** phonemes);
/** Allocate the memory for the features.
    It can't be done at creation since the amount of memory needed depends on the number of words of
    both the source and the target.
    It assumes the source and the target to be set.
    Note that it is called automatically by compute_features(), so you don't need to worry about it. */
void translation_pair_alloc_features(struct feature_set_s * feat, translation_pair_t *item);
/** Free any memory allocated for a translation_pair */
void translation_pair_free             (translation_pair_t *);


#ifdef __cplusplus
}
#endif

#endif /* _engine_h_ */
