/*
Copyright (c) 2011, Sylvain Raybaud <sylvain.raybaud@gmail.com>
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
Neither the name of the PAROLE team, LORIA, INRIA nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <pthread.h>
#include <glib.h>
#include <libxml/tree.h>

#include <ngram_model.h>
#include <logmath.h>
#include <pocketsphinx.h>
#include <src/libpocketsphinx/dict.h>

#include "sentence.h"
#include "engine.h"
#include "maparole.h"
#include "common.h"
#include "confidence_measure.h"

static asr_stream_t * asr_stream_new_item(char * word, s3wid_t wid, asr_stream_t * prev, asr_stream_t * next, float32 start, float32 end, int non_word) {
    /* debug("starting"); */
    asr_stream_t * new_item = malloc(sizeof(asr_stream_t));
    new_item->word = word;
    new_item->wid = wid;
    new_item->next = next;
    new_item->prev = prev;
    new_item->t_start = start;
    new_item->t_end = end;
    new_item->cm = 0;
    new_item->non_word = non_word;
    new_item->n_phonemes = 0;
    return new_item;
}


/*                         ***     CAUTION     ***                               */
/* asr low level manipulation functions assume the mutex on the stream is locked! */
void asr_stream_free(asr_stream_t * stream) {
    /* debug("starting"); */
    while(stream!=NULL) {
        asr_stream_t * current = stream;
        stream = current->next;
        free(current->word);
        free(current);
    }
}

asr_stream_t * asr_stream_free_n(asr_stream_t * stream, int n) {
    /* debug("starting"); */
    int i = 1;
    while(stream!=NULL && (n<0 || i<=n)) {
        asr_stream_t * current = stream;
        stream = current->next;
        free(current->word);
        free(current);
        i += 1;
    }
    return stream;
}


void asr_stream_free_backward(asr_stream_t * stream) {
    /* debug("starting"); */
    while(stream!=NULL) {
        asr_stream_t * current = stream;
        stream = current->prev;
        free(current->word);
        free(current);
    }
}

/* get last item of a stream */
/* NULL if stream is NULL (empty) */
asr_stream_t * asr_stream_get_end(asr_stream_t * stream) {
    /* debug("starting"); */
    asr_stream_t * item = stream;
    while(item!=NULL && item->next!=NULL) {
        item = item->next;
    }
    return item;
}
/* get first item of a stream */
/* NULL if stream is NULL (empty) */
asr_stream_t * asr_stream_rewind(asr_stream_t * stream) {
    /* debug("starting"); */
    asr_stream_t * item = stream;
    while(item!=NULL && item->prev!=NULL) {
        item = item->prev;
    }
    return item;
}


void write_ctm(manager_data_t * pool, int length) {
    debug("start");
    if(pool->asr_stream==NULL || length==0) {
        return;
    }
    
    debug("opening file");
    myFile f = myfopen(pool->save_ctm_filename,"a");
    
    debug("starting to write");
    int nb_words = 1;
    asr_stream_t * stream = pool->asr_stream;
    while((length<0 || nb_words<=length) && stream!=NULL) {
        if(stream->word!=NULL) {
            myfprintf(f,"%s %i %f %f %s %f\n",pool->in_filename,pool->n_channel,stream->t_start,stream->t_end-stream->t_start,stream->word,stream->cm);
        }
        stream = stream->next;
        nb_words += 1;
    }
    
    myfclose(f);
}

asr_stream_t * asr_stream_get_first_word(asr_stream_t * stream) {
    /* debug("starting"); */
    asr_stream_t * current = stream;
    while(current!=NULL && (current->word==NULL || current->non_word))
        current = current->next;
    return current;
}

int is_silence(asr_stream_t * item) {
    if(item->word!=NULL
        && (!strcmp(item->word,"++silence++") || !strcmp(item->word,"<sil>") || !strcmp(item->word,"SIL")))
        return 1;
    else
        return 0;
}

int asr_stream_length(asr_stream_t * stream) {
    /* debug("starting"); */
    int len = 0;
    asr_stream_t * current = stream;
    while(current!=NULL) {
        current = current->next;
        len += 1;
    }
    return len;
}

char * gen_filename(const char * base, const char * prefix, char * suffix, char * timestamp)  {
    debug("starting");
    
    if(!base && !prefix && !timestamp)
        myfail(EXIT_UNKNOWN_ERROR,"cannot conjure a filename out of thin air.");
        
    if(base==NULL) {
        size_t len = 1;
        if(suffix!=NULL) len += strlen(suffix);
        if(prefix!=NULL) len += strlen(prefix);
        if(timestamp!=NULL) len += strlen(timestamp);
        char * outfname = malloc(len*sizeof(char));
        sprintf(outfname,"%s%s%s",prefix!=NULL?prefix:"",timestamp!=NULL?timestamp:"",suffix!=NULL?suffix:"");
        return outfname;
    }
    
    if(prefix!=NULL) {
        gchar ** base_split = g_strsplit_set(base,"/\\",0);
        int i=-1; while(base_split[i+1]!=NULL) i++;
        size_t len = 1;
        if(i>=0) len += strlen(base_split[i]);
        if(suffix!=NULL) len += strlen(suffix);
        if(prefix!=NULL) len += strlen(prefix);
        if(timestamp!=NULL) len += strlen(timestamp);
        char * outfname = malloc(len*sizeof(char));
        sprintf(outfname,"%s%s%s%s",prefix,(i>=0)?base_split[i]:"",timestamp!=NULL?timestamp:"",suffix!=NULL?suffix:"");
        g_strfreev(base_split);
        return outfname;
    }
    
    size_t len = 1+strlen(base);
    if(suffix!=NULL) len += strlen(suffix);
    if(timestamp!=NULL) len += strlen(timestamp);
    char * outfname = malloc(len*sizeof(char));
    sprintf(outfname,"%s%s%s",base,timestamp!=NULL?timestamp:"",suffix!=NULL?suffix:"");
    return outfname;
    
}

void init_ps_lm(mp_config_t * config) {
    debug("start");
    
    if(config->ps_config.n_lm==0)
        myfail(EXIT_CONFIG_ERROR,"no LM specified");
    
    logmath_t * lmath = logmath_init(config->ps_config.lmbase,0,1);
    config->ps_config.lms = malloc(config->ps_config.n_lm*sizeof(ngram_model_t*));
    int i;
    for(i=0;i<config->ps_config.n_lm;i++) {
      config->ps_config.lms[i] = ngram_model_read(config->ps_config.decoders.cfg[0].ps_cmd_ln,config->ps_config.lmset[i],NGRAM_AUTO, lmath);
    }
    config->ps_config.lm = ngram_model_set_init(config->ps_config.decoders.cfg[0].ps_cmd_ln,config->ps_config.lms,config->ps_config.lmnames,config->ps_config.lmweights,config->ps_config.n_lm);
    /* Storing info for CM */
    config->ps_config.lmath = lmath;
}

audio_chunk_t * audio_chunk_new(int16 * audio, size_t size) {
    audio_chunk_t * chunk = malloc(size*sizeof(int16));
    chunk->audio = audio;
    chunk->size = size;
    return chunk;
}

void audio_chunk_free(audio_chunk_t * chunk) {
    if(chunk==NULL)
        return;
    free(chunk->audio);
    free(chunk);
}


static const char ** get_phonemes(dict_t * dict, s3wid_t wid, int * n_phonemes) {
    
    int32 pronlen = dict_pronlen(dict,wid);
    const char ** phonemes = malloc((pronlen+1)*sizeof(char*));
    phonemes[pronlen] = NULL;
    
    for(int32 i=0; i<pronlen; i++) {
        phonemes[i] = dict_ciphone_str(dict,wid,i);
    }
    
    if(n_phonemes!=NULL) *n_phonemes = pronlen;
    
    return phonemes;
}

/* get best hypothesis as a linked list of items */
asr_stream_t * get_asr_stream_from_utterance(manager_data_t * pool, ps_seg_t * seg) {
    debug("start");
    
    dict_t * dict = pool->mp_engine->dict_list[pool->decoders.current];
    
    ps_seg_t * segment = seg;
    asr_stream_t * stream = NULL;
    
    while(segment!=NULL) {
        const char * __word__ = ps_seg_word(segment);
        char * word = malloc((strlen(__word__)+1)*sizeof(char)); strcpy(word,__word__);
        
        if(strcmp(word,"<s>") && strcmp(word,"</s>")) {
            
            asr_stream_t * prev = stream;
            
            s3wid_t wid = dict_wordid(dict,word);
            dict_word2basestr(word);
            stream = asr_stream_new_item(word,wid,prev,NULL,-1,-1,dict_filler_word(dict,wid));
            stream->phonemes = get_phonemes(dict,wid,&stream->n_phonemes);
            
            if(prev!=NULL) {
                prev->next = stream;
            }
            
            int start_frm = -1, end_frm = -1;
            ps_seg_frames(segment,&start_frm,&end_frm);
            stream->t_start = start_frm*pool->samples_frame_shift/pool->sampling_rate+pool->segment_time_offset;
            stream->t_end = end_frm*pool->samples_frame_shift/pool->sampling_rate+pool->segment_time_offset;
            
        }
        segment = ps_seg_next(segment);
    }
    
    return asr_stream_rewind(stream);
}

/* merge best hypothesis into the stream of recognised items */
void merge_utt_into_stream(manager_data_t * pool, asr_stream_t * utt) {
    debug("start");
    
    if(utt==NULL) {
        return;
    }
    
    /* naive method (ester) */
    /* get longest common word */
    /* merge from that word */
    /* if none found, consider that last but one word in pool is best */
    
    asr_stream_t * left_utt = asr_stream_get_end(pool->asr_stream);
    if(left_utt==NULL) {
        /* no left utterance, nothing to merge into */
        pool->asr_stream = utt;
        return;
    }
    
    float32 right_t_start = utt->t_start;
    
    float32 greatest_length = 0;
    asr_stream_t * cut_left = NULL; /* cut left utterance after (not including)... */
    asr_stream_t * cut_right = NULL; /* cut right utterance at (including)... */
    while(left_utt!=NULL && left_utt->t_end>=right_t_start) {
        asr_stream_t * right_utt = utt;
        while(right_utt!=NULL && right_utt->t_start<left_utt->t_end) {
            if(((right_utt->t_end>=left_utt->t_start && right_utt->t_start<=left_utt->t_start) || (right_utt->t_end>=left_utt->t_end && right_utt->t_start<=left_utt->t_end))
                && (left_utt->t_end-left_utt->t_start+right_utt->t_end-right_utt->t_start)>greatest_length
                && left_utt->word!=NULL && right_utt->word!=NULL
                && !strcmp(right_utt->word,left_utt->word)) {
                cut_left = left_utt;
                cut_right = right_utt;
                greatest_length = left_utt->t_end-left_utt->t_start+right_utt->t_end-right_utt->t_start;
            }
            right_utt = right_utt->next;
        }
        left_utt = left_utt->prev;
    }
    
    /* if no common word found */
    /* remove last word of left utterance */
    /* then, first words of right utterance */
    /* until time consistent merge is possible */
    if(cut_left==NULL) {
        cut_left = asr_stream_get_end(pool->asr_stream)->prev;
        if(cut_left==NULL) cut_left = asr_stream_get_end(pool->asr_stream);
        float32 left_t_start = cut_left->t_start;
        asr_stream_t * right_utt = utt;
        while(right_utt->next!=NULL && right_utt->next->t_end<=left_t_start) {
            right_utt = right_utt->next;
        }
        cut_right = (right_utt->next!=NULL)?right_utt->next:right_utt;
    }
    
    /* if timings are inconsistent (may happen for words overlapping with chunk borders) */
    /* just fiddle with them (it doesn't really matter, only for asr scoring) */
    if(cut_left!=NULL && cut_right!=NULL && cut_right->next!=NULL && cut_left->t_end>cut_right->next->t_start) {
        cut_left->t_end = (cut_left->t_end+cut_right->next->t_start)/2;
        cut_left->t_start = min(cut_left->t_start,cut_left->t_end);
        cut_right->next->t_start = cut_left->t_end;
        cut_right->next->t_end = max(cut_right->next->t_end,cut_right->next->t_start);
    }
    
    asr_stream_free(cut_left->next);
    cut_left->next = cut_right->next;
    if(cut_left->next!=NULL) {
        cut_left->next->prev = cut_left;
    }
    asr_stream_free_backward(cut_right);
}


int process_raw(manager_data_t * pool, audio_chunk_t * chunk) {
    decoders_set_t * deco = &pool->decoders;
    int start_utt = ps_start_utt(deco->list[deco->current].instance,NULL);
    int nb_frames = ps_process_raw(deco->list[deco->current].instance,chunk->audio,chunk->size,FALSE,TRUE);
    int end_utt = ps_end_utt(deco->list[deco->current].instance);

    /* saving score */
    int32 score;
    const char *hyp = ps_get_hyp(deco->list[deco->current].instance, &score, NULL);
    pool->penultimate_reco_score = pool->last_reco_score;
    pool->last_reco_score = logmath_exp(pool->mp_config->ps_config.lmath, score);
    debug("got utterance '%s' (score: %f)", hyp, pool->last_reco_score);
    hyp = NULL;
    return (start_utt>=0 && end_utt>=0)?nb_frames:-1;
}


void asr_manager_terminate(manager_data_t * pool) {
    pthread_mutex_lock(&pool->asr_hyp_queue_mutex);
    pool->asr_manager_is_dead = 1;
    pthread_cond_signal(&pool->asr_hyp_queue_one_hyp_pushed);
    pthread_mutex_unlock(&pool->asr_hyp_queue_mutex);
}


translation_pair_t * translation_pair_new() {
    static translation_pair_t zz = {0};
    translation_pair_t *item = malloc(sizeof(translation_pair_t));
    *item = zz;
    return item;
}

translation_pair_t * translation_pair_new_from_lattice(ps_lattice_t * lattice) {
    translation_pair_t *item = translation_pair_new();
    item->src_lattice = ps_lattice_retain(lattice);
    return item;
}


void translation_pair_free(translation_pair_t *s) {
    /* debug("freeing src and tgt sentences"); */
    sentence_free(s->source);
    sentence_free(s->target);
    /* debug("freeing input audio, stream and lattice"); */
    audio_chunk_free(s->src_audio);
    asr_stream_free(s->src_stream);
    if(s->src_lattice)
      ps_lattice_free(s->src_lattice);
    /* debug("freeing CM structures"); */
    free(s->mt_sentLevelFeatures);
    free(s->mt_wordLevelFeaturesBuffer);
    free(s->mt_wordLevelFeatures);
    free(s->asr_sentLevelFeatures);
    free(s->asr_wordLevelFeaturesBuffer);
    free(s->asr_wordLevelFeatures);
    free(s->asr_word_cm);
    free(s->mt_word_cm);
    free(s);
}

void translation_pair_set_target(translation_pair_t *item, char *target_str,const char ** phonemes) {
  item->target = sentence_new(target_str,phonemes);
}

void translation_pair_alloc_features(feature_set_t * feat, translation_pair_t *item) {
  int i;

  item->mt_sentLevelFeatures = malloc(feat->mt_slf_count*sizeof(float));
  item->mt_wordLevelFeaturesBuffer = malloc(item->target->wordCount*feat->mt_wlf_count*sizeof(float));
  item->mt_wordLevelFeatures = malloc(item->target->wordCount*sizeof(float*));
  for (i=0 ; i<item->target->wordCount ; i++)
    item->mt_wordLevelFeatures[i] = item->mt_wordLevelFeaturesBuffer+(i*feat->mt_wlf_count);

  item->asr_sentLevelFeatures = malloc(feat->asr_slf_count*sizeof(float));
  item->asr_wordLevelFeaturesBuffer = malloc(item->source->wordCount*feat->asr_wlf_count*sizeof(float));
  item->asr_wordLevelFeatures = malloc(item->source->wordCount*sizeof(float*));
  for (i=0 ; i<item->source->wordCount ; i++)
    item->asr_wordLevelFeatures[i] = item->asr_wordLevelFeaturesBuffer+(i*feat->asr_wlf_count);

  item->mt_word_cm = malloc(item->target->wordCount*sizeof(float));
  item->asr_word_cm = malloc(item->source->wordCount*sizeof(float));
}

void translation_pair_get_asr_result(translation_pair_t * item, manager_data_t * pool) {
    decoders_set_t * deco = &pool->decoders;
    if(pool->mp_config->one_best) {
        ps_seg_t * seg = ps_seg_iter(deco->list[deco->current].instance,NULL);
        item->src_stream = get_asr_stream_from_utterance(pool,seg);
    }
    else {
        ps_lattice_t * lattice = ps_get_lattice(deco->list[deco->current].instance); /* possible race condition if new utterance is started before lattice_retain is called */
        item->src_lattice = ps_lattice_retain(lattice);
    }
}

char * translation_pair_get_src_string   (translation_pair_t * t) {
    return t->source!=NULL?sentence_get_string(t->source):NULL;
}

char * translation_pair_get_src_phonemes_string(translation_pair_t * t) {
    if(t->source==NULL) {
        return NULL;
    }
    size_t s_len = 1;
    char * s = malloc(s_len*sizeof(char)); s[0] = '\0';
    int i = 0;
    const char * p = t->source->phonemes[i];
    while(p!=NULL) {
        s_len += strlen(p)+1;
        s = realloc(s,s_len);
        strcat(s," ");
        strcat(s,p);
        p = t->source->phonemes[++i];
    }
    return s;
}

asr_stream_t * translation_pair_get_src_stream   (translation_pair_t * t) {
    return t->src_stream;
}

ps_lattice_t * translation_pair_get_src_lattice(translation_pair_t * t) {
    return t->src_lattice;
}

char * translation_pair_get_tgt_string   (translation_pair_t * t) {
    return t->target!=NULL?sentence_get_string(t->target):NULL;
}

void translation_pair_set_audio        (translation_pair_t * t, audio_chunk_t * a) {
    t->src_audio = a;
}

xmlNodePtr add_node_to_xml(manager_data_t * pool, translation_pair_t * pair) {
    debug("start");
    
    xmlNodePtr parent = pool->xml_result_node;
    xmlNodePtr word_node,pair_node,src,hyp;
    const int N = 256;
    char buff[N];
    int i;

    LIBXML_TEST_VERSION;

    pair_node = xmlNewChild(parent, NULL, BAD_CAST "pair", BAD_CAST "");
    src = xmlNewChild(pair_node, NULL, BAD_CAST "source", BAD_CAST "");
    hyp = xmlNewChild(pair_node, NULL, BAD_CAST "translation", BAD_CAST "");\
    
    xmlNewChild(src, NULL, BAD_CAST "text", BAD_CAST translation_pair_get_src_string(pair));
    if(pool->mp_config->phonemes) {
        char * s = translation_pair_get_src_phonemes_string(pair);
        xmlNewChild(src, NULL, BAD_CAST "phonemes", BAD_CAST s);
        free(s);
    }
    
    xmlNewChild(hyp, NULL, BAD_CAST "text", BAD_CAST translation_pair_get_tgt_string(pair));
    
    if(!pool->mp_config->no_confidence) {
        snprintf(buff,N, "%f",pair->mt_sent_cm);
        xmlNewProp(hyp, BAD_CAST "cm", BAD_CAST buff);

        /* for each word in target*/
        for (i=0 ; i<pair->target->wordCount ; i++) {
            /* create node */
            word_node = xmlNewChild(hyp, NULL, BAD_CAST "word", BAD_CAST pair->target->words[i]);
            snprintf(buff,N, "%f",pair->mt_word_cm[i]);
            /* add attribute 'cm' */
            xmlNewProp(word_node, BAD_CAST "cm", BAD_CAST buff);
        }
    }

    return pair_node;
}

void save_xml(manager_data_t * pool) {
    debug("start");
    xmlSaveFormatFileEnc(pool->save_xml_filename, pool->xml_result_doc, "UTF-8", 1);
}
