#ifndef _LANGUAGE_MODEL_H_
#define _LANGUAGE_MODEL_H_

#include <pocketsphinx.h>


/** \file 
   Module providing a generic way to use Moses' and Sphinx's language models.

   It allows to use loaded LMs from running instances,
   and gives an object-based interface to LMs.

   Genericity is achieved through those types:
   - wid_s: a word index in a given LM
   - lmi_s: a generic interface with "member" functions.
   - lmi_data_t: the actual LM data used by the functions, member of the above structure

   Only the creation of an lmi_t needs to be different for Sphinx and Moses
   (see sphinx_lmi_new() and moses_lmi_new() for details).

   <b>Usage: full example</b>
   <pre>
  lmi_t lmi = sphinx_lmi_new(sphinx_lm_open("foo.arpa", pool.lmath)); // Non generic creation
  ... // later 
  wid_t foo = lmi.wid(lmi.data, "foo"); // Generic use
  float p = lmi.score_bigram(lmi.data, foo, foo); // Gives you p("foo foo")
  ... // etc
  lmi.free_data(lmi.data);</pre>

   For the list of the "member" functions, see the documentation for lmi_s.
 */

/** Word index in a LM. */
typedef union wid_s {
    int32 idx;          /**< For Sphinx, it's a simple int. */
    void *adr;		/**< For Moses, it's a pointer to an object. */
} wid_t;

/** Pointer to Sphinx-or-Moses-specific data */
typedef void * lmi_data_t;

/** Structure providing a generic access to a language model implementation (lmi)
 * in an object-based fashion.
 * Hence, it has a set of "member" functions, and the first argument of those functions should be the <i>data</i> field of the structure.
 * e.g.
 *
 * wid_t wid_foo = lmi.wid(lmi.data, "foo"); 
 *
 * int foo_exists = lmi.unigram_exists(lmi.data, wid_foo);
 *
 * The structure is entirely filled at creation by sphinx_lmi_new() or moses_lmi_new()
   and nothing needs to be freed except <i>data</i> (with lmi_t.free_data()), if dynamically allocated. */
typedef struct lmi_s {
    /** p(a) */
    float (*score_unigram)(lmi_data_t, wid_t a);
    /** p(b|a) */
    float (*score_bigram)(lmi_data_t, wid_t a, wid_t b);
    /** p(c|a,b) */
    float (*score_trigram)(lmi_data_t, wid_t a, wid_t b, wid_t c);
    /** p(t[n-1]|t[n-2],...,t[0]) */
    float (*score_ngram)(lmi_data_t, int n, wid_t *t);
    /** true iff the LM contains 'a' */
    int (*unigram_exists)(lmi_data_t, wid_t a);
    /** true iff the LM contains 'a b' */
    int (*bigram_exists)(lmi_data_t, wid_t a, wid_t b);
    /** true iff the LM contains 'a b c' */
    int (*trigram_exists)(lmi_data_t, wid_t a, wid_t b, wid_t c);
    /** Get the index of 'word' in this LM */
    wid_t (*wid)(lmi_data_t, char * word);
    /** free the data of the LM
       (do not call this if the LM is retrieved from the running instance */
    void (*free_data)(lmi_data_t);
    /** Pointer to the actual data of the implementation */
    lmi_data_t * data;
    /* Miscellaneous */
    wid_t bos;			/**< Index of &lt;s> */
    wid_t eos;			/**< Index of &lt;/s> */
} lmi_t;

/** Structure referencing a Sphinx LM; to be used as a lmi_data_t */
typedef struct sphinx_lm_s {
    ngram_model_t *lm;	  /**< The actual LM */
    logmath_t *lmath;     /**< Sphinx object for log-proba convertions.
    Note: it is a field of Sphinx's internal "struct ngram_model_s",
    but it is not accessible in ngram_model_t, hence this structure */
    pthread_mutex_t mutex ;	/**< Mutex for concurrent access to Sphinx functions. */
} sphinx_lm_t;

/** Open and load in memory a language model with Sphinx.
    Warning: a new mutex is created for this LM, perhaps there should be only one mutex
    (pool.lm_mutex) shared for all models. (TODO: check)
    \return a structure for sphinx_lmi_new()
 */
sphinx_lm_t *sphinx_lm_open(char *filename, logmath_t *lmath);

/** Create a full lmi_t using Sphinx's LM from <i>data</i>.
    If <i>data</i> was created by sphinx_lm_open(), it can be retrieved in the lmi_t.data field to be freed:\n
    lmi.free_data(lmi.data)
    \param data a sphinx_lm_t returned by sphinx_lm_open() or filled manually with a pre-open model.
*/
lmi_t sphinx_lmi_new(sphinx_lm_t *data);

#ifdef __cplusplus
extern "C" {
#endif

  /** Pointer to a Moses LM. (C++ type:  Moses::LanguageModelImplementation*) */
  typedef void* moses_lm_t;

/** Create a full lmi_t using Moses' LM from <i>data</i>.
   * If <i>data</i> was created by moses_lm_open(), it can be retrieved in the lmi_t.data field to be freed:\n
    lmi.free_data(lmi.data)
   \param data a moses_lm_t returned by moses_lm_open() or moses_get_lm().\n
   */
  lmi_t moses_lmi_new(moses_lm_t *data);

  /** Retrieve the model language from the running instance of Moses. */
  moses_lm_t moses_get_lm();
  /** Open and load in memory a model language with Moses. */
  moses_lm_t moses_lm_open(char *fileName);

#ifdef __cplusplus
} /* extern "C" */
#endif


#endif
