#include <cutils.h>
#include "common.h"
#include "language_model.h"

#include <string>
#include <vector>

// Moses headers
#include <FFState.h>
#include <StaticData.h>
#include <TranslationSystem.h>
#include <ScoreIndexManager.h>
#include <LanguageModelFactory.h>
#include <LanguageModelKen.h>

/** \ file
 * Functions for querying Moses' implementations of language models.
 */

using namespace Moses;

/** Wrapper for Moses' Word allowing the manipulation of wids.

    Internally, a wid in Moses is a Factor*, but language models functions use Word instances,
    where a word only contains an array of Factor*.

    In single factor models, the wid is the only Factor* in the array, we use this class to retrieve the wid of a Word, or to rebuild a Word from a wid.
*/
class MyWord : public Moses::Word {
public:
  MyWord() {}
  /** Make a Word from a wid */
  MyWord(Factor *wid) {
    m_factorArray[0] = wid;
  }
  /** Get the wid of the word */
  const Factor * getWid() {
    return m_factorArray[0];
  }
};

moses_lm_t moses_get_lm() {
  // Just digging through the Moses layers
  TranslationSystem ts = StaticData::Instance().GetTranslationSystem(std::string("default"));
  LMList list = ts.GetLanguageModels();
  LanguageModel *lm = *(list.begin());
  return lm;
}

moses_lm_t moses_lm_open(char *fileName) {
  std::vector<FactorType> factorTypes(1);
  factorTypes[0] = 0;
  std::string file(fileName);
  try {
    LanguageModel *lm =
      LanguageModelFactory::
      CreateLanguageModel(Ken,
			  factorTypes,
			  3,
			  file,
			  (ScoreIndexManager&) StaticData::Instance().GetScoreIndexManager(),
			  0 /*unused for ken LM*/);
    return lm;
  }
  catch (std::exception& e) {
    myfail(EXIT_MOSES_ERROR, "%s", e.what());
  }
}


float moses_score_unigram(lmi_data_t data, wid_t a) {
  LanguageModelImplementation *lmi = ((LanguageModel*)data)->GetLMImplementation();

  /* Prepare the contextFactor argument for call */
  std::vector<const Word*> contextFactor(1);

  MyWord w1((Factor*)a.adr);
  contextFactor[0] = &w1;

  /* Call */
  FFState *outState = lmi->NewState(lmi->GetNullContextState());
  LMResult r = lmi->GetValueForgotState(contextFactor, *outState);

  delete outState;
  return exp(r.score);
}

float moses_score_bigram(lmi_data_t data, wid_t a, wid_t b) {
  LanguageModelImplementation *lmi = ((LanguageModel*)data)->GetLMImplementation();

  /* Prepare the contextFactor argument for call */
  std::vector<const Word*> contextFactor(2);
  MyWord w1((Factor*)a.adr);
  contextFactor[0] = &w1;
  MyWord w2((Factor*)b.adr);
  contextFactor[1] = &w2;
  /* Call */
  FFState *outState = lmi->NewState(lmi->GetNullContextState());
  LMResult r = lmi->GetValueForgotState(contextFactor, *outState);

  delete outState;
  return exp(r.score);
}

float moses_score_trigram(lmi_data_t data, wid_t a, wid_t b, wid_t c) {
  LanguageModelImplementation *lmi = ((LanguageModel*)data)->GetLMImplementation();

  /* Prepare the contextFactor argument for call */
  std::vector<const Word*> contextFactor(3);
  MyWord w1((Factor*)a.adr);
  contextFactor[0] = &w1;
  MyWord w2((Factor*)b.adr);
  contextFactor[1] = &w2;
  MyWord w3((Factor*)c.adr);
  contextFactor[2] = &w3;

  /* Call */
  FFState *outState = lmi->NewState(lmi->GetNullContextState());
  LMResult r = lmi->GetValueForgotState(contextFactor, *outState);

  delete outState;
  return exp(r.score);
}

float moses_score_ngram(lmi_data_t data, int n, wid_t *t) {
  LanguageModelImplementation *lmi = ((LanguageModel*)data)->GetLMImplementation();

  /* Prepare the contextFactor argument for call */
  std::vector<const Word*> contextFactor(n);
  MyWord *words = new MyWord[n];
  for (int i=0 ; i<n ; i++) {
    MyWord w((Factor*)t[i].adr);
    words[i] = w;
    contextFactor[i] = words+i;
  }

  /* Call */
  FFState *outState = lmi->NewState(lmi->GetNullContextState());
  LMResult r = lmi->GetValueForgotState(contextFactor, *outState);
  delete[] words;
  delete outState;
  return exp(r.score);
}

int moses_unigram_exists(lmi_data_t data, wid_t a) {
  LanguageModelImplementation *lmi = ((LanguageModel*)data)->GetLMImplementation();

  /* Prepare the contextFactor argument for call */
  std::vector<const Word*> contextFactor(1);
  MyWord w1((Factor*)a.adr);
  contextFactor[0] = &w1;

  /* Call */
  FFState *outState = lmi->NewState(lmi->GetNullContextState());
  LMResult r = lmi->GetValueForgotState(contextFactor, *outState);

  delete outState;
  return !r.unknown;
}

int moses_bigram_exists(lmi_data_t data, wid_t a, wid_t b) {
  LanguageModelKenBase *lmi = (LanguageModelKenBase*) ((LanguageModel*)data)->GetLMImplementation();

  /* Prepare the contextFactor argument for call */
  std::vector<const Word*> contextFactor(2);
  MyWord w1((Factor*)a.adr);
  contextFactor[0] = &w1;
  MyWord w2((Factor*)b.adr);
  contextFactor[1] = &w2;

  /* Call */
  FFState *outState = lmi->NewState(lmi->GetNullContextState());
  LMKenResult r = lmi->GetKenFullScoreForgotState(contextFactor, *outState);

  delete outState;
  return r.ngram_length==2;
}

int moses_trigram_exists(lmi_data_t data, wid_t a, wid_t b, wid_t c) {
  LanguageModelKenBase *lmi = (LanguageModelKenBase*) ((LanguageModel*)data)->GetLMImplementation();

  /* Prepare the contextFactor argument for call */
  std::vector<const Word*> contextFactor(3);
  MyWord w1((Factor*)a.adr);
  contextFactor[0] = &w1;
  MyWord w2((Factor*)b.adr);
  contextFactor[1] = &w2;
  MyWord w3((Factor*)c.adr);
  contextFactor[2] = &w3;

  /* Call */
  FFState *outState = lmi->NewState(lmi->GetNullContextState());
  LMKenResult r = lmi->GetKenFullScoreForgotState(contextFactor, *outState);

  delete outState;
  return r.ngram_length==3;
}

wid_t moses_wid(lmi_data_t data, char * word) {
  wid_t wid;
  MyWord w;
  const std::vector<FactorType>& factorOrder =
    StaticData::Instance().GetInputFactorOrder();

  w.CreateFromString(Output, factorOrder, word, true);
  wid.adr = (void *)w.getWid();
  return wid;
}

void moses_free_data(lmi_data_t data) {
  LanguageModel *_lm = (LanguageModel*) data;
  delete _lm;
}

lmi_t moses_lmi_new(moses_lm_t *data) {
  lmi_t moses = {
    moses_score_unigram,
    moses_score_bigram,
    moses_score_trigram,
    moses_score_ngram,
    moses_unigram_exists,
    moses_bigram_exists,
    moses_trigram_exists,
    moses_wid,
    moses_free_data,
    (lmi_data_t*) data,
    moses_wid(data, (char*)"<s>"),
    moses_wid(data, (char*)"</s>")
  };
  return moses;
}

