#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#
# Copyright (C) 2010, Sylvain Raybaud <sraybaud@crans.org>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Sylvain Raybaud nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY Sylvain Raybaud <sraybaud@crans.org> ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL Sylvain Raybaud BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#
# This piece of software turns a word lattice in SLF (HTK format) and converts it
# to PLF (Moses input format)
# SLF is documented there: http://labrosa.ee.columbia.edu/doc/HTKBook21/node257.html
# PLF is documented there: http://www.statmt.org/moses/?n=Moses.WordLattices
# I tried to keep most of the flexibility of SLF but some functionnalities are missing (see TODO below)
#
# Usage:  slf2plf [options] input_file > output
# or:     slf2plf [options] input_file.gz > output
# or:     slf2plf [options] < input_file > output
# Options:
#    -C, --dont-clean: don't remove noise information from input lattice [default: clean]
#    -S, --dont-sanitize: don't sanitize input words. Be careful (see help) [default: sanitize]
#    -D, --keep-dead-ends: don't remove nodes and link which only belong to path not terminating on sentence ends '</s>'. [default: prune dead ends]
#    -E, --no-exponential: sphinx lattice normaly contains log probabilities while Moses expect raw probabilities. Use this option you want to keep the score the way they are. [default: use exponential]
#    -n, --prune-noise: NOT TESTED! Remove transitions labeled with noise or silence in a way that preserves graph connectivity. Cannot be used with 'clean'. [default: don't prune]
#    -h, --help: print short help
#
# Principle:
#  1) Load graph into memory as a list of nodes and a list of links
#     each node has access to the list of its incoming and outgoing
#     links
#  2) Prune lattice of noise and paths not corresponding to complete sentences
#  3) Compute a topological ordering of nodes
#  4) Output lattice in moses input format
#
# The algorithm implemented for topological sort is the one described on the following wikipedia entry on the 6 august 2010:
#   http://en.wikipedia.org/wiki/Topological_sorting 
#
# TODO:
#  a) Make the program robust to missing accoustic and lm scores in the lattice --> DONE! (somewhat: missing scores are replaced with 0)
#  b) Make it possible to choose how output lattice score is computed from nodes and link attributes
#  c) Deal with sublattices
#  d) Make the program robust to incorrect input
#  e) Let the user provide a list of words to be ignored (silences and noise, for example)


from optparse import OptionParser
import gzip
import sys
import re
import copy
from math import exp


#general purpose tools to make the world a better place

def verbose_msg(opts,msg):
    if opts.verbose:
        sys.stderr.write(msg+'\n')

def open_plain_or_gz(filename,mode='r'):
    """Transparently open gziped or plain files, based on file extension"""
    return gzip.open(filename,mode) if filename.endswith(".gz") else open(filename,mode)

def verbose_msg(opts,msg):
    if opts.verbose:
        sys.stderr.write(msg+'\n')

def warn(msg):
    """Print a warning message to stderr. The program continues."""
    sys.stderr.write("*** WARNING *** %s\n" % msg)
    
def error(msg,code=1):
    """Print an error message to stderr and exists with the given code (default: 1)."""
    sys.stderr.write("*** ERROR *** %s\n" % msg)
    exit(code)


# regular expressions that will be used later

__attr_re__ = r"(?P<attribute>[^=]+)=(?P<quote>(\"|'')?)(?P<value>.*)(?P=quote)"  # filed_name=field_value    value may be "quoted"
__size_re__ = r"(?P<attribute>N|NODES|L|LINKS)=(?P<value>\d+)"  # for parsing size information in SLF
__node_re__ = r"I=(?P<id>\d+)(?P<attributes>(\s+%s)*)" % __attr_re__  # for parsing node definition in SLF
__link_re__ = r"J=(?P<id>\d+)(?P<attributes>(\s+%s)*)" % __attr_re__  # for parsing link information in SLF
__noise_re__ = r"\+\+[^\+]*\+\+"

# constants
EPSILON = "*EPS*"  # epsilon transition

# links to be cleaned out if requested
CLEAN = ['<sil>', 'euh', '!NULL', '!null', '<SIL>']

# word substitution list
REPLACE = {'!SENT_END': '.', '!sent_end': '.'}

# a few mandatory attributes in short and long forms. They will be stored in long format

ATTRIBUTES = {
    'E': 'END',
    'S': 'START',
    'W': 'WORD',
    'a': 'acoustic',
    'l': 'language'
    }

class Node:
    """ Node definition for using in word lattice """
    def __init__(self,index,attributes_string=None):
        """index is just the id of the node. attribute_string is a succession fo definition of the form: name=value"""
        self.index = index
        self.incoming = set()
        self.outgoing = set()
        self.attributes = {}
        if attributes_string:
            attributes = attributes_string.split()
            for attr_str in attributes:
                m = re.match(__attr_re__,attr_str)
                attr = __get_long_form__(m.group('attribute'))
                val = m.group('value')
                if attr in self.attributes:
                    warn("Node %i: attribute '%s' redefined." % (self.index,attr))
                self.attributes[attr] = val
    
    def has_attribute(self,attrname):
        return attrname in self.attributes
    
    def get(self,attr):
        if self.has_attribute(attr):
            return self.attributes[attr]
        else:
            return None

class Link:
    """ Link definition for using in word lattice """
    def __init__(self,index,attributes_string=None):
        """'index' is just the id of the link.
        'attribute_string' is a succession fo definition of the form: name=value
        'sanitize' is a flag indicating if words should be santized (e.g. escaping quote)"""
        self.index = index
        self.from_node = -1
        self.to_node = -1
        self.word = ""
        self.attributes = {}
        if attributes_string:
            attributes = attributes_string.split()
            for attr_str in attributes:
                m = re.match(__attr_re__,attr_str)
                attr = __get_long_form__(m.group('attribute'))
                val = m.group('value')
                if attr in self.attributes:
                    warn("Node %i: attribute '%s' redefined." % (self.index,attr))
                self.attributes[attr] = val
                if attr=="WORD":
                    self.word = val
        try:
            self.from_node = int(self.attributes['START'])
            self.to_node = int(self.attributes['END'])
        except:
            error("Incomplete definition of link %i" % self.index)
        #TODO: check that word attribute is defined either here or on destination node
    
    def has_attribute(self,attrname):
        return attrname in self.attributes
    
    def get(self,attr):
        if self.has_attribute(attr):
            return self.attributes[attr]
        else:
            return None

class Lattice:
    """ Class representing a word lattice with constructor and method for loading a lattice in HTK Standard Lattice Format.
    Also has a method for writing the lattice in Moses input format (PLF) """
    def __init__(self,lattfile=None,lines=None):  # can read lattice either from file or data in memory
        """ lattfile should be an open stream. lines should contain a reversed list of already read lines. At least one of them must be provided."""
        data = []
        if lattfile==None and (lines==None or len(lines)==0):
            error("Where shall I read the lattice from?")
        if lattfile!=None:
            if type(lattfile)==gzip.GzipFile:
                data = list(map(lambda x: x.decode(),lattfile.readlines()))  #not very memory efficient, but easier for parsing
            else:
                data = lattfile.readlines()  #not very memory efficient, but easier for parsing
            data.reverse()  # lines to be parsed will be stored on a stack
        if lines!=None:
            data = data+lines
        data = __clean__(data)  #remove comments and empty lines
        self.nodes_count = -1
        self.links_count = -1
        self.nodes = {}
        self.links = {}
        self.sublattices = {}
        self.lmscale = 1.0
        self.logbase = None
        self.attributes = {}
        self.__load_graph__(data)
    
    def has_attribute(self,attrname):
        return attrname in self.attributes
    
    def get(self,attr):
        if self.has_attribute(attr):
            return self.attributes[attr]
        else:
            return None
    
    def get_node(self,n):
        return self.nodes[n] if n in self.nodes else None
        
    def get_link(self,l):
        return self.links[l] if l in self.links else None
    
    def __load_graph__(self,lines):
        self.__parse_header__(lines)
        self.__parse_counts__(lines)
        self.__parse_nodes__(lines)
        self.__parse_links__(lines)
        if len(self.nodes)!=self.nodes_count or len(self.links)!=self.links_count:
            warn("Incorrect nodes or links count")
        if 'lmscale' in self.attributes:
            self.lmscale = float(self.get('lmscale'))
        if 'base' in self.attributes:
            self.logbase = float(self.get('base'))
     
    def __parse_header__(self,lines):
        attributes_tmp = {}
        go = True  # carry on as long as we are parsing a header
        while len(lines)>0 and go:
            l = lines.pop()
            m = re.match(__attr_re__,l)
            if m and not re.match(__size_re__+".*",l):  # should match an attribute but *not* a size definition
                attr = m.group('attribute')
                val = m.group('value')
                if attr in attributes_tmp:
                    warn("Attribute '%s' redefined." % attr)
                attributes_tmp[attr] = val
                if attr=="SUBLAT": # we are parsing a sublattice. go down, create it and then carry on.
                    sublat = Lattice(lines=lines)
                    # merge attributes
                    for k in attributes_tmp:
                        if k in sublat.attributes:
                            warn("Sublattice '%s': attribute '%s' redefined." % (val,k))
                    sublat.attributes.update(attributes_tmp)
                    attributes_tmp = {}
                    self.sublattices[val] = sublat
                    if lines.pop().strip()!=":":
                        warn("Missing colon at the end of sublattice '%s'. Continuing anyway, but things could go wrong." % val)
            else:
                go = False
        if not go:
            lines.append(l)
        self.attributes = attributes_tmp
    
    # If I understand correctly counts may be written in any order, on 1 or 2 lines
    # Therefore parsed lines need to be split
    def __parse_counts__(self,lines):
        go = True
        while len(lines)>0 and go:
            l = lines.pop()
            terms = l.split()
            while len(terms)>0 and go:
                t = terms.pop()
                m = re.match(__size_re__,t)
                if m:
                    if m.group('attribute')=='N' or m.group('attribute')=='NODES':
                        if self.nodes_count!=-1:
                            warn("Nodes count redefined.")
                        self.nodes_count = int(m.group('value'))
                    elif m.group('attribute')=='L' or m.group('attribute')=='LINKS':
                        if self.links_count!=-1:
                            warn("Links count redefined.")
                        self.links_count = int(m.group('value'))
                else:
                    go = False
        if not go:
            lines.append(l)
        
        
    def __parse_nodes__(self,lines):
        go = True
        while len(lines)>0 and go:
            l = lines.pop()
            m = re.match(__node_re__,l)
            if m:
                index = int(m.group('id'))
                n = Node(index,m.group('attributes'))
                if index in self.nodes:
                    warn("Node %i redefined." % index)
                self.nodes[index] = n
            else:
                go = False
        if not go:
            lines.append(l)
        
    def __parse_links__(self,lines):
        removed_links = 0
        go = True
        while len(lines)>0 and go:
            l = lines.pop()
            m = re.match(__link_re__,l)
            if m:
                index = int(m.group('id'))
                if index in self.links:
                    warn("Link %i redefined." % index)
                e = Link(index,m.group('attributes'))
                self.__update_nodes__(e)  # Information about nodes is contained in link definition
                self.links[index] = e
            else:
                go = False
        if not go:
            lines.append(l)
    
    def __update_nodes__(self,link):
        # in case the nodes were not explicitely defined
        if not (link.from_node in self.nodes):
            self.nodes[link.from_node] = Node(link.from_node)
        if not (link.to_node in self.nodes):
            self.nodes[link.to_node] = Node(link.to_node)
        # update connection information of nodes
        self.nodes[link.from_node].outgoing.add(link.index)
        self.nodes[link.to_node].incoming.add(link.index)
    
    def remove_orphan_nodes(self):
        """ Put orphaned nodes out of their misery """
        removed_nodes = 0
        for n in list(self.nodes.values()):
            if len(n.incoming)==0 and len(n.outgoing)==0:
                self.nodes.pop(n.index)
                self.nodes_count -= 1
    
    #NOT TESTED!!
    # this fonction completely removes links carrying only noise. it's better to replace them with epsilon transitions
    def prune_noise(self,cleanList):
        """ [NOT TESTED] Remove noise information from the graph without breaking connectivity.
        'cleanList' is a list of additionnal words that should cause a link to be removed.
        Should not be useful with Moses: use epsilon transitions instead."""
        for link in list(self.links.values()):
            # get a link that should be removed
            if 'WORD' in link.attributes and ((link.get('WORD') in cleanList) or (re.match(__noise_re__,link.get('WORD')))):
                src = self.nodes[link.from_node]
                tgt = self.nodes[link.to_node]
                # get all links out of its target node
                for l in tgt.outgoing:
                    # create a new, identical link
                    new = copy.deepcopy(self.links[l])
                    self.links_count += 1
                    new.index = self.links_count
                    # insert this link into lattice
                    self.links[new.index] = new
                    # add this link between src and target of l
                    new.from_node = src.index
                    src.outgoing.add(new.index)
                    self.nodes[new.to_node].incoming.add(new.index)
                # remove link
                src.outgoing.remove(link.index)
                tgt.incoming.remove(link.index)
                self.links.pop(link.index)
                self.links_count -= 1
        self.remove_orphan_nodes()
    
    def remove_dead_ends(self):
        """If the lattice contains terminal node whose incoming edges are not carrying the word '</s>' (dead ends) then Moses doesn't seem to be able to translate the lattice. Such nodes therefore need to be recursevely removed."""
        if utterance_has_no_end(self):
            warn("Utterance has no end ('</s>' or '!SENT_END'). Will not remove dead ends.")  # the lattice is probably not suitable for translation with Moses
            return
        end_nodes = get_ends(self)  # get all terminal nodes
        while end_nodes!=[]:
            node = end_nodes.pop()
            for l in set(node.incoming):
                if not is_end(self.links[l].word) and not is_end(node.get('WORD')):  # remove incoming edge if it's not an end-of-sentence
                    from_node = self.nodes[self.links[l].from_node]
                    from_node.outgoing.remove(l)
                    node.incoming.remove(l)
                    self.links.pop(l)
                    self.links_count -= 1
                    if from_node.outgoing==set([]):  # parent node is candidate for pruning if it has no outgoing edges
                        end_nodes.append(self.nodes[from_node.index])
            if node.incoming==set([]):  # if all incoming links have been pruned then remove the node
                self.nodes.pop(node.index)
                self.nodes_count -= 1
    
    # write lattice in PLF
    def write_plf(self,options):
        """ Write lattice in moses input format (PLF) on standard output.
        options.sanitize: sanitize input words (e.g. escape quotes)
        options.clean: replace noise and silence information with epsilon transitions
        options.exp: raise score to exp before writing"""
        # get topological ordering
        ordered_nodes = topological_ordering(self)
        # process sublattices
        # NOT THAT EASY
        # ordered nodes of a sublattice can be inserted at the position of the node they replace
        #for i,n in len(ordered_nodes):
            #if self.nodes[n].has_key('L'):  # this node is in fact a sublattice
                #sublat_ordered = self.sublattices
        # write lattice
        sys.stdout.write("(")
        for i,n in enumerate(ordered_nodes[:-1]):  # list all nodes in topological order
            __print_links_out__(self,ordered_nodes[i:],options)  # and write links leaving them
        sys.stdout.write(")\n")
        return

# word sanitizing
def __sanitize_word__(string):
    s = string.replace("'","\\'")
    return s


def __get_long_form__(string):
    if string in ATTRIBUTES:
        return ATTRIBUTES[string]
    else:
        return string

# remove comments and blank lines from input
def __clean__(lines):
    data = []
    for l in lines:
        if l.strip()!='' and l[0]!='#':
            data.append(re.sub("#.*\n",'',l))
    return data

# print a link out of the first node in the list 'remains'
def __print_links_out__(lattice,remains,options):
    n = lattice.nodes[remains[0]]
    sys.stdout.write("(")
    for l in n.outgoing:
        link = lattice.links[l]
        distance = remains.index(link.to_node)  # this is the distance between current node and the node it is linked to
        a = float(link.get('acoustic')) if link.has_attribute('acoustic') else 0
        l = float(link.get('language')) if link.has_attribute('language') else 0
        
        dest = lattice.get_node(link.to_node)
        w = link.get('WORD')
        w_n = dest.get('WORD')
        if w!=None and w_n!=None and w!=w_n:
            warn("Link %i and destination node %i carry different WORD attribute (%s!=%s). Choosing link's." % (link.index,dest.index,w,w_n))
        if w==None:
            w = w_n
        if w==None:
            warn("No word information for link %i! Marking as epsilon-transition." % link.index)
            w = EPSILON
        
        if options.clean:
            if w in CLEAN or re.match(__noise_re__,w):
                w = EPSILON
            if w in REPLACE:
                w = REPLACE[w]
            if len(w)>=2 and w[0]=="!":  #word is actually and event
                w = EPSILON
        if options.sanitize:
            w = __sanitize_word__(w)
        score = a+l*lattice.lmscale if not options.exp else (pow(lattice.logbase,a+l*lattice.lmscale) if lattice.logbase!=None else exp(a+l*lattice.lmscale))
        sys.stdout.write("('%s',%e,%i)," % (w,score,distance))
        if opts.verbose and options.exp and (score>1 or score<0):
            warn("Probability greater than 1: '%s': P=%e" % (w,score))
    sys.stdout.write("),")


def is_end(w):
    return w=="</s>" or w=="!SENT_END" or w=="!sent_end"

def utterance_has_no_end(lattice):
    for l in list(lattice.links.values()):
        if (l.has_attribute('WORD') and is_end(l.word)) or is_end(lattice.get_node(l.to_node).get('WORD')):
            return False
    return True

def get_ends(lattice):
    end_nodes = []
    for n in list(lattice.nodes.values()):
        if n.outgoing == set([]):
            end_nodes.append(n)
    return end_nodes

# topological sorting algorithm needs to be initialised with a list of initial nodes having no incoming link
def get_starting_nodes(lattice):
    """ Given an instance of class Lattice, output the list of all nodes with no incoming link """
    S = []
    for n in list(lattice.nodes.values()):
        if len(n.incoming)==0:
            S.append(n.index)
    return S

def topological_ordering(original_lattice):
    """ Given an instance of class Lattice, output a topological ordering of nodes (as a list of indexes) """
    lattice = copy.deepcopy(original_lattice)  # the lattice will be modified by the program. Work on a copy.
    start_nodes = get_starting_nodes(lattice)
    ordered_nodes = []
    while len(start_nodes)>0:
        n = start_nodes.pop()
        ordered_nodes.append(n)
        for l in set(lattice.nodes[n].outgoing):
            m = lattice.nodes[lattice.links[l].to_node]
            lattice.links.pop(l)
            m.incoming.remove(l)
            lattice.nodes[n].outgoing.remove(l)
            if len(m.incoming)==0:
                start_nodes.append(m.index)
    if len(lattice.links)>0:
        warn("Lattice has at least one cycle.")
    return ordered_nodes




USAGE = """
%prog [options] input_file > output
or: %prog [options] input_file.gz > output
or: %prog [options] < input_file > output
input_file is a graph in SLF format (HTK), possibly gziped (.gz extension required)."""

if __name__=='__main__':
    parser = OptionParser(usage=USAGE)
    parser.add_option('-C','--dont-clean',default=True,action="store_false",dest="clean",help="Word lattice graph generated by sphinx contain information about noise and silences which should be replaced with epsilon transition for translation. Use this option if you want to keep it. Cannot be used with 'prune-noise'. [clean]")
    parser.add_option('-n','--prune-noise',default=False,action="store_true",dest="prune_noise",help="Remove transitions labeled with noise or silence in a way that preserves graph connectivity. Cannot be used with 'clean'. [%default]")
    parser.add_option('-D','--keep-dead-ends',default=True,action="store_false",dest="remove_dead_ends",help="Remove nodes and link which only belong to path not terminating on sentence ends '</s>' or '!SENT_END'. [prune dead ends]")
    parser.add_option('-S','--dont-sanitize',default=True,action="store_false",dest="sanitize",help="Don't sanitize input words. Be careful! for example if there is a quote in your graph moses won't be able to decode it. [sanitize]")
    parser.add_option('-E','--no-exponential',default=True,action="store_false",dest="exp",help="Sphinx lattices normaly contain log probabilities while Moses expects plain probabilities. Use this option you want to keep the score the way they are. [use exponential]")
    parser.add_option('-v','--verbose',default=False,action="store_true",dest="verbose",help="verbosity [%default]")
    opts,args = parser.parse_args()
    if len(args)>1:
        parser.print_help()
        error("Too many arguments.",1)
    if opts.clean and opts.prune_noise:
        warn("'prune_noise' requested. Disabling 'clean'.")
        opts.clean = False
    input_file = sys.stdin if (len(args)==0 or args[0]=="-") else open_plain_or_gz(args[0])
    lattice = Lattice(input_file)
    
    verbose_msg(opts,"Lattice loaded:")
    verbose_msg(opts," N=%i - L=%i" % (lattice.nodes_count,lattice.links_count))
    verbose_msg(opts," logbase=%s" % (str(lattice.logbase) if lattice.logbase!=None else "default (not specified)"))
    verbose_msg(opts," lmscale=%f" % lattice.lmscale)
    
    if opts.prune_noise:
        lattice.prune_noise(CLEAN)
    if opts.remove_dead_ends:
        lattice.remove_dead_ends()
    lattice.write_plf(opts)
    if not (len(args)==0 or args[0]=="-"):
        input_file.close()
