/*
Copyright (c) 2011, Sylvain Raybaud <sylvain.raybaud@gmail.com>
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
Neither the name of the PAROLE team, LORIA, INRIA nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <glib.h>
#include <stdlib.h>
#include <string.h>

#include "slf.h"
#include "engine.h"

#define SLF_HEADER 1
#define SLF_NODE 2
#define SLF_LINK 3
#define SLF_COUNT 4

#define version_re "version=(?P<version>\\d*(\\.\\d*)?)"
#define utterance_re "utterance=(?P<utterance>[[:alnum:].-_/]+)"
#define lmscale_re "lmscale=(?P<lmscale>\\d*(\\.\\d*)?)"
#define lmname_re "lmname=(?P<lmname>[[:alnum:].-_/]+)"
#define base_re "base=(?P<base>\\d*(\\.\\d*)?)"
#define wdpenalty_re "wdpenalty=(?P<wdpenalty>\\d*(\\.\\d*)?)"
#define header_re "(" version_re ")|(" utterance_re ")|(" base_re ")|(" lmscale_re ")|(" lmname_re ")|(" wdpenalty_re ")"
#define nodecount_re "N=(?P<nodescount>\\d+)"
#define linkcount_re "L=(?P<linkscount>\\d+)"
#define counts_re "(" nodecount_re ")|(" linkcount_re ")|(" nodecount_re "[[:blank:]]+" linkcount_re ")|(" linkcount_re "[[:blank:]]+" nodecount_re ")"
#define attribute_re "[[:alpha:]]+=[[:alnum:].-_<>]+"
#define node_re "I=\\d+(" "[[:blank:]]+" attribute_re ")*"
#define link_re "J=\\d+(" "[[:blank:]]+" attribute_re ")*"

GRegex * slf_header_re = NULL;
GRegex * slf_node_re = NULL;
GRegex * slf_link_re = NULL;
GRegex * slf_version_re = NULL;
GRegex * slf_utterance_re = NULL;
GRegex * slf_lmscale_re = NULL;
GRegex * slf_lmname_re = NULL;
GRegex * slf_base_re = NULL;
GRegex * slf_wdpenalty_re = NULL;
GRegex * slf_nodecount_re = NULL;
GRegex * slf_linkcount_re = NULL;
GRegex * slf_count_re = NULL;
GRegex * slf_attribute_re = NULL;

slf_t * slf_load(const char * filename) {
    debug("starting");

    slf_header_re = g_regex_new(header_re,G_REGEX_CASELESS|G_REGEX_ANCHORED|G_REGEX_OPTIMIZE,0,NULL);
    slf_node_re = g_regex_new(node_re,G_REGEX_CASELESS|G_REGEX_ANCHORED|G_REGEX_OPTIMIZE,0,NULL);
    slf_link_re = g_regex_new(link_re,G_REGEX_CASELESS|G_REGEX_ANCHORED|G_REGEX_OPTIMIZE,0,NULL);
    slf_version_re = g_regex_new(link_re,G_REGEX_CASELESS|G_REGEX_ANCHORED|G_REGEX_OPTIMIZE,0,NULL);
    slf_utterance_re = g_regex_new(link_re,G_REGEX_CASELESS|G_REGEX_ANCHORED|G_REGEX_OPTIMIZE,0,NULL);
    slf_lmscale_re = g_regex_new(link_re,G_REGEX_CASELESS|G_REGEX_ANCHORED|G_REGEX_OPTIMIZE,0,NULL);
    slf_lmname_re = g_regex_new(lmname_re,G_REGEX_CASELESS|G_REGEX_ANCHORED|G_REGEX_OPTIMIZE,0,NULL);
    slf_base_re = g_regex_new(base_re,G_REGEX_CASELESS|G_REGEX_ANCHORED|G_REGEX_OPTIMIZE,0,NULL);
    slf_wdpenalty_re = g_regex_new(wdpenalty_re,G_REGEX_CASELESS|G_REGEX_ANCHORED|G_REGEX_OPTIMIZE,0,NULL);
    slf_nodecount_re = g_regex_new(nodecount_re,G_REGEX_CASELESS|G_REGEX_ANCHORED|G_REGEX_OPTIMIZE,0,NULL);
    slf_linkcount_re = g_regex_new(linkcount_re,G_REGEX_CASELESS|G_REGEX_ANCHORED|G_REGEX_OPTIMIZE,0,NULL);
    slf_count_re = g_regex_new(counts_re,G_REGEX_CASELESS|G_REGEX_ANCHORED|G_REGEX_OPTIMIZE|G_REGEX_DUPNAMES,0,NULL);
    slf_attribute_re = g_regex_new(attribute_re,G_REGEX_CASELESS|G_REGEX_ANCHORED|G_REGEX_OPTIMIZE,0,NULL);

    myFile f = myfopen(filename,"r");
    if(f.error) {
        myfclose(f);
        return NULL;
    }

    slf_t * lattice = slf_new();

    char * line = NULL;
    size_t linelen = 0;
    int linenb = 0;

    while(myclevergetline(&line,&linelen,f)>0) {
        linenb++;
        g_strstrip(line);
        switch(line_type(line)) {
            case SLF_HEADER:
                parse_header(lattice,line);
                break;
            case SLF_COUNT:
                parse_count(lattice,line);
                break;
            case SLF_NODE:
                parse_node(lattice,line);
                break;
            case SLF_LINK:
                parse_link(lattice,line);
                break;
            default:
                myerror("failed to recognise line '%s' (line #%i?) in file '%s'",line,linenb,filename);
                return NULL;
        }
    }

    if(line)
        free(line);
    myfclose(f);

    slf_build_connections(lattice);

    g_regex_unref(slf_header_re);
    g_regex_unref(slf_node_re);
    g_regex_unref(slf_link_re);
    g_regex_unref(slf_version_re);
    g_regex_unref(slf_utterance_re);
    g_regex_unref(slf_lmscale_re);
    g_regex_unref(slf_base_re);
    g_regex_unref(slf_wdpenalty_re);
    g_regex_unref(slf_nodecount_re);
    g_regex_unref(slf_linkcount_re);
    g_regex_unref(slf_count_re);
    g_regex_unref(slf_attribute_re);

    return lattice;

}

int line_type(char * line) {
    debug("starting");

    if(g_regex_match(slf_header_re,line,0,NULL)) {
        return SLF_HEADER;
    }
    if(g_regex_match(slf_count_re,line,0,NULL)) {
        return SLF_COUNT;
    }
    if(g_regex_match(slf_node_re,line,0,NULL)) {
        return SLF_NODE;
    }
    if(g_regex_match(slf_link_re,line,0,NULL)) {
        return SLF_LINK;
    }

    return -1;
}


void parse_count(slf_t * lat, char * line) {
    debug("starting");

    GMatchInfo *match;
    g_regex_match(slf_count_re,line,0,&match);
    char * ncount = g_match_info_fetch_named(match,"nodescount");
    char * lcount = g_match_info_fetch_named(match,"linkscount");
    if(ncount!=NULL && strlen(ncount)>0)
        lat->nodes_count = atoi(ncount);
    if(lcount!=NULL && strlen(lcount)>0)
        lat->links_count = atoi(lcount);
}

void parse_header(slf_t * lat, char * line) {
    debug("starting");

    GMatchInfo *match;

    if(g_regex_match(slf_version_re,line,0,&match)) {
        lat->version = g_match_info_fetch_named(match,"version");
        g_match_info_free(match);
        return;
    }
    if(g_regex_match(slf_utterance_re,line,0,&match)) {
        lat->utterance = g_match_info_fetch_named(match,"utterance");
        g_match_info_free(match);
        return;
    }
    if(g_regex_match(slf_lmname_re,line,0,&match)) {
        lat->lmname = g_match_info_fetch_named(match,"lmname");
        g_match_info_free(match);
        return;
    }
    if(g_regex_match(slf_base_re,line,0,&match)) {
        lat->base = atof(g_match_info_fetch_named(match,"base"));
        g_match_info_free(match);
        return;
    }
    if(g_regex_match(slf_lmscale_re,line,0,&match)) {
        lat->lmscale = atof(g_match_info_fetch_named(match,"lmscale"));
        g_match_info_free(match);
        return;
    }
    if(g_regex_match(slf_wdpenalty_re,line,0,&match)) {
        lat->wdpenalty = atof(g_match_info_fetch_named(match,"wdpenalty"));
        g_match_info_free(match);
        return;
    }

}

void parse_node(slf_t * lat, char * line) {
    debug("starting");

    int i = 0;
    gchar ** tokens = g_strsplit_set(line," \t",0);
    int32 node_id = -1;
    double t = -1;
    char * word = NULL;

    while(tokens[i]!=NULL) {
        if(strlen(tokens[i])>0) {
            gchar ** token = g_strsplit(tokens[i],"=",0);
            if(!strcmp(token[0],"I"))
                node_id = (int32)atoi(token[1]);
            else if(!strcmp(token[0],"t"))
                t = atof(token[1]);
            else if(!strcmp(token[0],"W"))
                word = g_strdup(token[1]);
            g_strfreev(token);
        }
        i += 1;
    }

    slf_add_node(lat,node_id,t,word);

    g_strfreev(tokens);
}

void parse_link(slf_t * lat, char * line) {
    debug("starting");

    int i = 0;
    gchar ** tokens = g_strsplit_set(line," \t",0);

    int32 link_id = -1;
    int32 from = -1;
    int32 to = -1;
    char * word = NULL;
    double a = 2;
    double l = 2;
    int v = -1;

    while(tokens[i]!=NULL) {
        if(strlen(tokens[i])>0) {
            gchar ** token = g_strsplit(tokens[i],"=",0);
            if(!strcmp(token[0],"J"))
                link_id = (int32)atoi(token[1]);
            else if(!strcmp(token[0],"S"))
                from = (int32)atoi(token[1]);
            else if(!strcmp(token[0],"E"))
                to = (int32)atoi(token[1]);
            else if(!strcmp(token[0],"v"))
                v = atoi(token[1]);
            else if(!strcmp(token[0],"a"))
                a = atof(token[1]);
            else if(!strcmp(token[0],"l"))
                l = atof(token[1]);
            else if(!strcmp(token[0],"W"))
                word = g_strdup(token[1]);
            g_strfreev(token);
        }
        i += 1;
    }

    slf_add_link(lat,link_id,from,to,word,a,l,v);

    g_strfreev(tokens);
}

void slf_add_link(slf_t * lat, int32 id, int32 from, int32 to, char * word, double a, double l, int v) {
    debug("starting");

    slf_link_t * new_link = slf_link_new(id,from,to,word,a,l,v);
    slf_links_list_add_link(&(lat->links),new_link);
}

void slf_links_list_add_link(slf_links_list_t ** list, slf_link_t * link) {
    debug("starting");

    slf_links_list_t * new = slf_links_list_new_link();
    new->link = link;
    new->tail = *list;
    *list = new;
}

slf_links_list_t * slf_links_list_get_last(slf_links_list_t * list) {
    debug("starting");

    slf_links_list_t * last = list;
    while(last->link!=NULL)
        last = last->tail;
    return last;
}

slf_links_list_t * slf_links_list_new_link() {
    debug("starting");

    slf_links_list_t * new = malloc(sizeof(slf_links_list_t));
//     *new = {0};
    new->link = NULL;
    new->tail = NULL;
    return new;
}

slf_link_t *  slf_link_new(int32 id, int32 from, int32 to, char * word, double a, double l, int v) {
    debug("starting");

    slf_link_t * link = malloc(sizeof(slf_link_t));
    link->id = id;
    link->from = from;
    link->to = to;
    link->word = word;
    link->acoustic = a;
    link->lm = l;
    link->v = v;
    return link;
}

void slf_nodes_list_add_node(slf_nodes_list_t ** list, slf_node_t * node) {
    debug("starting");

    slf_nodes_list_t * new_node = slf_nodes_list_new_node();
    new_node->node = node;
    new_node->tail = *list;
    *list = new_node;
}

void slf_add_node(slf_t * lat, int32 id, double time, char * word) {
    debug("starting");

    slf_node_t * new_node = slf_node_new(id,time,word);
    slf_nodes_list_add_node(&(lat->nodes),new_node);
}

slf_nodes_list_t * slf_nodes_list_get_last(slf_nodes_list_t * list) {
    debug("starting");

    slf_nodes_list_t * last = list;
    while(last->node!=NULL)
        last = last->tail;
    return last;
}

slf_nodes_list_t * slf_nodes_list_new_node() {
    debug("starting");

    slf_nodes_list_t * new = malloc(sizeof(slf_nodes_list_t));
    new->node = NULL;
    new->tail = NULL;
    return new;
}

slf_node_t * slf_get_node_by_id(slf_nodes_list_t * nodes_list,int32 id) {
    debug("starting");

    slf_nodes_list_t * node = nodes_list;
    while(node!=NULL && node->node!=NULL && node->node->id!=id)
        node = node->tail;
    return (node!=NULL && node->node!=NULL && node->node->id==id)?node->node:NULL;
}

slf_node_t * slf_node_new(int32 id, double time, char * word) {
    debug("starting");

    slf_node_t * node = malloc(sizeof(slf_node_t));
    node->id = id;
    node->word = word;
    node->t = time;
    node->entries = NULL;
    node->exits = NULL;
    return node;
}

/* returns and remove the first element */
slf_node_t * slf_nodes_list_pop(slf_nodes_list_t ** queue) {
    debug("starting");

    slf_nodes_list_t * first = *queue;
    slf_node_t * node = first->node;
    *queue = first->tail;
    free(first);
    return node;

}

slf_t * slf_new() {
    debug("starting");

    slf_t * lat = malloc(sizeof(slf_t));
//     *lat = {0};
    *lat = (slf_t) {0};
    return lat;
}

void slf_build_connections(slf_t * lat) {
    debug("starting");

    slf_nodes_list_t * node_list = lat->nodes;

    while(node_list!=NULL && node_list->node!=NULL) {
        slf_links_list_t * link_list = lat->links;
        while(link_list!=NULL && link_list->link!=NULL) {
            if(link_list->link->to==node_list->node->id) {
                slf_links_list_add_link(&(node_list->node->entries),link_list->link);
            }
            if(link_list->link->from==node_list->node->id) {
                slf_links_list_add_link(&(node_list->node->exits),link_list->link);
            }
            link_list = link_list->tail;
        }
        node_list = node_list->tail;
    }

}

/*
plf_t * slf_to_plf(ps_lattice_t * lattice) {
    mywarning("slf_to_plf is still a stub");
    return -1;
}*/

slf_node_t ** slf_topological_sort(slf_t * lat) {
    debug("starting");

    /* sorted nodes */
    slf_node_t ** sorted_stack = malloc(lat->nodes_count*sizeof(slf_node_t*));
    int sorted_ptr = 0;
    /* nodes with no incoming edges */
    slf_nodes_list_t * prime_queue = get_prime(lat);
    debug("null initial prime queue? %s",(prime_queue!=NULL && prime_queue->node!=NULL)?"no":"yes");
    /* 1 if link has been removed, 0 o/w */
    char * removed = calloc((size_t)lat->links_count,sizeof(char));

    while(prime_queue!=NULL && prime_queue->node!=NULL) {
        slf_node_t * node = slf_nodes_list_pop(&prime_queue);
        visit_exits(node,lat->nodes,&prime_queue,removed);
        sorted_stack[sorted_ptr++] = node;
        debug("prime node %i got stacked.",node->id);
    }

    if(!check_result(removed,lat->links_count))
        mywarning("Input lattice is not a DAG. Get lost.");

    slf_nodes_list_destroy_list_only(prime_queue);
    free(removed);

    #ifdef DEBUG
    fprintf(stderr,"sorted graph:\n");
    int i;
    for(i=0;i<lat->nodes_count;i++) {
        fprintf(stderr,"  %i\n",sorted_stack[i]->id);
        slf_links_list_t * link = sorted_stack[i]->exits;
        while(link!=NULL && link->link!=NULL) {
            fprintf(stderr,"    %i: %i -> %i: %s\n",link->link->id,link->link->from,link->link->to,link->link->word);
            link = link->tail;
        }
    }
    #endif

    return sorted_stack;

}

void visit_exits(slf_node_t * node, slf_nodes_list_t * nodes_list, slf_nodes_list_t ** prime_queue, char * removed) {
    debug("starting");

    slf_links_list_t * exits_list = node->exits;

    while(exits_list!=NULL && exits_list->link!=NULL) {
        removed[exits_list->link->id] = 1;
        slf_node_t * to = slf_get_node_by_id(nodes_list,exits_list->link->to);
        if(is_prime(to,removed))
            slf_nodes_list_add_node(prime_queue,to);
        exits_list = exits_list->tail;
    }

}

slf_nodes_list_t * get_prime(slf_t * lat) {
    debug("starting");

    slf_nodes_list_t * primes = slf_nodes_list_new_node();
//     slf_nodes_list_t * last = primes;
    slf_nodes_list_t * nodes = lat->nodes;

    while(nodes!=NULL && nodes->node!=NULL) {
        if(nodes->node->entries==NULL || nodes->node->entries->link==NULL) {
            slf_nodes_list_add_node(&primes,nodes->node);
            debug("node %i is prime",nodes->node->id);
        }
        nodes = nodes->tail;
    }

    return primes;

}

int check_result(char * removed, int n) {
    debug("starting");

    int i;
    for(i=0;i<n;i++)
        if(!removed[i])
            return 0;

        return 1;

}

void slf_nodes_list_destroy_list_only(slf_nodes_list_t * list) {
    debug("starting");

    slf_nodes_list_t * node_list = list;
    while(node_list!=NULL) {
        slf_nodes_list_t * next = node_list->tail;
        free(node_list);
        node_list = next;
    }
}

int is_prime(slf_node_t * to, char * removed) {
    debug("starting");

    slf_links_list_t * entries = to->entries;
    while(entries!=NULL && entries->link!=NULL) {
        if(!removed[entries->link->id])
            return 0;
        entries = entries->tail;
    }

    return 1;
}

int slf_links_list_len(slf_links_list_t * list) {
    int len = 0;
    while(list!=NULL && list->link!=NULL) {
        len += 1;
        list = list->tail;
    }
    return len;
}

int slf_nodes_list_len(slf_nodes_list_t * list) {
    int len = 0;
    while(list!=NULL && list->node!=NULL) {
        len += 1;
        list = list->tail;
    }
    return len;
}
