#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#
# Copyright (C) 2011, Sylvain Raybaud <sylvain.raybaud@loria.fr>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Sylvain Raybaud nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY Sylvain Raybaud <sylvain.raybaud@loria.fr> ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL Sylvain Raybaud BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#

from optparse import OptionParser
from mytools import open_plain_or_gz,warn,error,verbose_msg,openfiles,closefiles,getline
import sys

def load(bdlex_fname,cmu_fname):
    
    P_bdlex = set()
    P_cmu = set()
    
    bdlex = {}
    cmu = {}
    
    bdlex_f = open_plain_or_gz(bdlex_fname)
    l = bdlex_f.readline()
    while l!="":
        
        toks = l.split()
        w = toks[0]
        ph = toks[2:]
        
        P_bdlex = P_bdlex.union(ph)
        
        if w in bdlex:
            bdlex[w].append(ph)
        else:
            bdlex[w] = [ph]
        
        l = bdlex_f.readline()
    
    bdlex_f.close()
    
    cmu_f = open_plain_or_gz(cmu_fname)
    l = cmu_f.readline()
    while l!="":
        
        toks = l.split()
        w = toks[0]
        ph = toks[1:]
        
        P_cmu = P_cmu.union(ph)
        
        if w in cmu:
            cmu[w].append(ph)
        else:
            cmu[w] = [ph]
        
        l = cmu_f.readline()
    
    cmu_f.close()
    
    return bdlex,cmu,P_bdlex,P_cmu

def create_conversion_table(bdlex,cmu,verbose=False):
    
    T = {}
    
    mismatch = 0
    n_1ph = 0
    
    for w in cmu:
        if len(cmu[w])==1 and w in bdlex and len(bdlex[w])==1:
            
            n_1ph += 1
            
            ph_cmu = cmu[w][0]
            ph_bdlex = bdlex[w][0]
            
            if len(ph_cmu)==len(ph_bdlex):
                for i,p in enumerate(ph_cmu):
                    if not ph_bdlex[i] in T:
                        T[ph_bdlex[i]] = p
                    else:
                        if verbose:
                            warn("*** MISMATCH *** in word '%s': '%s' aligned with '%s' but previously aligned with '%s'" % (w,ph_bdlex[i],p,T[ph_bdlex[i]]))
                        mismatch += 1
            else:
                mismatch += 1
                if verbose:
                    warn("*** MISMATCH *** pronunciation lengths for word '%s' differ in bdlex and cmu (and no variant)")
    
    verbose_msg(verbose,"> Conversion table created based on %i words for %i bdlex phonemes" % (n_1ph,len(T)))
    
    return T,mismatch

def convert(bdlex,T):
    
    res = {}
    
    for w in bdlex:
        res[w] = []
        for p in bdlex[w]:
            new_p = []
            for ph in p:
                new_p.append(T[ph])
            res[w].append(new_p)
    
    return res

def write(res):
    
    for w in res:
        print("%s %s" % (w," ".join(res[w][0])))
        for i,p in enumerate(res[w][1:]):
            print("%s(%i) %s" % (w,i+2," ".join(res[w][0])))


if __name__=='__main__':
    parser = OptionParser(usage="Usage: %prog [options] bdlex_dict cmu_dict > res_cmu_dict")
    parser.add_option('-v','--verbose',default=False,action="store_true",dest="verbose",help="verbosity [%default]")
    opts,args = parser.parse_args()
    if len(args)!=2:
        parser.print_help()
        error("provide bdlex and cmu dictionaries.",1)
    
    bdlex_fname = args[0]
    cmu_fname = args[1]
    
    bdlex,cmu,P_bdlex,P_cmu = load(bdlex_fname,cmu_fname)
    
    if len(bdlex)!=len(cmu):
        warn("BDLEX and CMU do not have the same number of phonemes.")
    
    T,mismatch = create_conversion_table(bdlex,cmu,opts.verbose)
    
    if len(T)!=len(P_bdlex):
        warn("Conversion table does not cover all phonemes.")
    
    if mismatch>0:
        warn("%i mismatches have been detected." % mismatch)
    
    res = convert(bdlex,T)
    
    write(res)
